/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.datatypes;

import ij.ImagePlus;
import ij.gui.NewImage;
import ij.process.ImageProcessor;
import java.awt.Color;
import java.awt.Polygon;
import java.awt.geom.*;
import java.util.*;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImageByte;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.exceptions.*;
import de.unihalle.informatik.MiToBo.io.files.ImageSave;
import de.unihalle.informatik.MiToBo.segmentation.regions.labeling.*;

/**
 * Class to create 2D contour objects with a vector of points which belongs to
 * the contour. Every contour can include a couple of inner contours.
 * 
 * @modified Sep 27, 2010
 * 
 * @author misiak
 * 
 */

public class MTBContour2D {

  /**
   * The pixels which belong to the contour.
   */
  private Vector<Point2D.Double> points;

  /**
   * Possible inner contours of the contour object.
   */
  private Vector<MTBContour2D> inner;

  /**
   * The length of the contour, calculated from the number of included points.
   */
  private int length;

  /**
   * Standard constructor. Creates an empty 2D contour object.
   */
  public MTBContour2D() {
    this.points = new Vector<Point2D.Double>();
    this.inner = new Vector<MTBContour2D>();
    this.length = 0;
  }

  /**
   * Constructor to create an 2D contour object from a 2D point vector.
   * 
   * @param points
   *          vector with 2D points
   */
  public MTBContour2D(Vector<Point2D.Double> points) {
    this.points = points;
    this.inner = new Vector<MTBContour2D>();
    this.length = points.size();
  }

  /**
   * Override java.lang.Object.clone() to create and return a copy of this
   * object.
   */
  @Override
  public MTBContour2D clone() {
    MTBContour2D tmpContour = new MTBContour2D(this.points);
    tmpContour.setInner(this.inner);
    return tmpContour;
  }

  /**
   * Get a Contour2D copy of this object.
   * 
   * @return Copy of this Contour2D object.
   */
  public MTBContour2D getContour() {
    return this.clone();
  }

  /**
   * Get all points belonging to the contour object.
   * 
   * @return Vector with all 2D points of the contour.
   */
  public Vector<Point2D.Double> getPoints() {
    return points;
  }

  /**
   * Get a specific 2D point belonging to the contour object.
   * 
   * @return 2D point at specific index.
   */
  public Point2D.Double getPointAt(int index) {
    return points.elementAt(index);
  }

  /**
   * Get the length of the contour object.
   * 
   * @return The length of the contour (number of points).
   */
  public int getLength() {
    return length;
  }

  /**
   * Add a 2D point to the existing contour.
   * 
   * @param x
   *          x-coordinate of the new point
   * @param y
   *          y-coordinate of the new point
   */
  public void addPixel(int x, int y) {
    length++;
    points.addElement(new Point2D.Double(x, y));
  }

  /**
   * Set the inner contours of the contour object.
   * 
   * @param C
   *          vector with the inner contours belonging to the contour object
   */
  public void setInner(Vector<MTBContour2D> C) {
    inner = C;
  }

  /**
   * Add a inner contour to the existing contour object.
   * 
   * @param C
   *          the new 2D inner contour
   */
  public void addInner(MTBContour2D C) {
    inner.addElement(C);
  }

  /**
   * Get a specific inner contour from the contour object.
   * 
   * @param index
   *          specific index of the inner contour to get from the inner contour
   *          vector
   * @return Inner contour at the specific index as 2D contour object.
   */
  public MTBContour2D getInner(int index) {
    return inner.elementAt(index);
  }

  /**
   * Get all inner contours from the contour object.
   * 
   * @return All inner contours as vector 2D contour objects.
   */
  public Vector<MTBContour2D> getAllInner() {
    return this.inner;
  }

  /**
   * Get the number of inner contours including in the contour object.
   * 
   * @return The number of inner contours.
   */
  public int countInner() {
    return inner.size();
  }

  /**
   * Method to get the included region in a 2D contour as Region2D object.
   * 
   * @param imageSizeX
   *          width of image that includes the region
   * @param imageSizeY
   *          height of image that includes the region
   * @return 2D region included in the 2D contour.
   * @throws MTBOperatorException
   */
  // ToDo use contains method from Polygon2D if method was tuned up
  public MTBRegion2D getRegion(int imageSizeX, int imageSizeY)
      throws MTBOperatorException, MTBProcessingDAGException {
    /*
     * Calculate whole region which is included in the outer contour.
     */
    int[] xps = new int[this.points.size()];
    int[] yps = new int[this.points.size()];
    int n = 0;
    for (Point2D.Double p : this.points) {
      xps[n] = (int) (p.x + 0.5);
      yps[n] = (int) (p.y + 0.5);
      n++;
    }
    // create awt polygon
    Polygon awtPoly = new Polygon(xps, yps, this.points.size());
    // background black, polygon white
    ImagePlus img = NewImage.createByteImage("", imageSizeX, imageSizeY, 1,
        NewImage.FILL_BLACK);
    ImageProcessor ip = img.getProcessor();
    ip.setColor(Color.WHITE);
    // fill polygon with Java-AWT filling algorithm
    ip.fillPolygon(awtPoly);
    // create MTBImage object as result image
    MTBImageByte fullRegion = (MTBImageByte) MTBImage.createMTBImage(img);
    for (int i = 0; i < this.points.size(); i++) {
      Point2D.Double p = this.points.elementAt(i);
      fullRegion.putValueInt((int) Math.round(p.x), (int) Math.round(p.y), 255);
    }
    /*
     * If an inner contour exists, get all the pixels in this regions (the hole)
     * and subtract these pixels from the region of the outer contour to get the
     * hole back.
     */
    for (int i = 0; i < this.inner.size(); i++) {
      MTBContour2D tmpCont = this.inner.elementAt(i);
      // create Java-AWT polygon from current inner contour pixels
      int[] tmpXPS = new int[tmpCont.length];
      int[] tmpYPS = new int[tmpCont.length];
      int tmpN = 0;
      for (Point2D.Double p : tmpCont.points) {
        tmpXPS[tmpN] = (int) (p.x + 0.5);
        tmpYPS[tmpN] = (int) (p.y + 0.5);
        tmpN++;
      }
      Polygon tmpPoly = new Polygon(tmpXPS, tmpYPS, tmpCont.length);
      // background white, polygon black
      ImagePlus img2 = NewImage.createByteImage("", imageSizeX, imageSizeY, 1,
          NewImage.FILL_BLACK);
      ImageProcessor ip2 = img2.getProcessor();
      ip2.setColor(Color.WHITE);
      ip2.fillPolygon(tmpPoly);
      for (int y = 0; y < imageSizeY; ++y) {
        for (int x = 0; x < imageSizeX; ++x) {
          if (ip2.getPixel(x, y) == 255)
            fullRegion.putValueInt(x, y, 0);
        }
      }
      for (int j = 0; j < tmpCont.length; j++) {
        Point2D.Double p = tmpCont.points.elementAt(j);
        fullRegion.putValueInt((int) Math.round(p.x), (int) Math.round(p.y),
            255);
      }
    }
    /*
     * Create Region2D object from binary MTBImage. If some inner contours
     * exist, more then one region is returned. The outer Region is the first
     * one in the vector, than the inner regions follow.
     */
    LabelComponentsSequential LCS = new LabelComponentsSequential(fullRegion,
        true);
    LCS.runOp(null);
    MTBRegion2DSet resultRegion = LCS.getResultingRegions();
    System.out.println(resultRegion.size());
    return resultRegion.elementAt(0);
  }

  /**
   * Method to save a contour with all inner contours on a binary image with
   * specific width and height.
   * 
   * @param file
   *          path where the file should be saved, can be null
   * @param width
   *          width of the binary image
   * @param height
   *          height of the binary image
   * @return 2D inner- and outer-contours in a binary image of type
   *         MTBImageByte.
   * 
   * @author misiak
   * @throws MTBOperatorException
   * @throws MTBProcessingDAGException
   */
  public MTBImageByte toMTBImageByte(String file, int width, int height)
      throws MTBOperatorException, MTBProcessingDAGException {
    MTBImageByte imageToDraw = (MTBImageByte) MTBImage.createMTBImage(width,
        height, 1, 1, 1, MTBImageType.MTB_BYTE);
    for (int i = 0; i < points.size(); ++i) {

      Point2D.Double p = points.elementAt(i);
      imageToDraw.putValueDouble((int) Math.round(p.getX()), (int) Math.round(p
          .getY()), 255);
    }
    int end = inner.size();
    for (int j = 0; j < end; j++) {
      MTBContour2D c = inner.elementAt(j);
      Vector<Point2D.Double> in_points = c.getPoints();
      for (int k = 0; k < in_points.size(); k++) {
        Point2D.Double p = in_points.elementAt(k);
        imageToDraw.putValueDouble((int) Math.round(p.getX()), (int) Math
            .round(p.getY()), 255);
      }
    }
    if (file != null) {
      ImageSave IS = new ImageSave(file, (MTBImage) imageToDraw, true, false);
      IS.runOp(null);
    }
    return imageToDraw;
  }

  /**
   * Method to save a contour with all inner contours in a given MTBImage image.
   * 
   * @param file
   *          path where the file should be saved, can be null
   * @param image
   *          image where the contours should be drawn in
   * @return 2D inner- and outer-contours in a binary image of type
   *         MTBImageByte.
   * 
   * @author misiak
   * @throws MTBOperatorException
   * @throws MTBProcessingDAGException
   */
  public MTBImage toMTBImage(String file, MTBImage image)
      throws MTBOperatorException, MTBProcessingDAGException {
    MTBImage imageToDraw = image.duplicate();
    for (int i = 0; i < points.size(); ++i) {
      Point2D.Double p = points.elementAt(i);
      imageToDraw.putValueDouble((int) Math.round(p.getX()), (int) Math.round(p
          .getY()), image.getTypeMax());
    }
    int end = inner.size();
    for (int j = 0; j < end; j++) {
      MTBContour2D c = inner.elementAt(j);
      Vector<Point2D.Double> in_points = c.getPoints();
      for (int k = 0; k < in_points.size(); k++) {
        Point2D.Double p = in_points.elementAt(k);
        imageToDraw.putValueDouble((int) Math.round(p.getX()), (int) Math
            .round(p.getY()), image.getTypeMax());
      }
    }
    if (file != null) {
      ImageSave IS = new ImageSave(file, imageToDraw, true, false);
      IS.runOp(null);
    }
    return imageToDraw;
  }
}
