/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.datatypes;

import java.util.Vector;

import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.exceptions.MTBProcessingDAGException;
import de.unihalle.informatik.MiToBo.operator.MTBData;
import de.unihalle.informatik.MiToBo.operator.MTBOpArgumentDescriptor;
import de.unihalle.informatik.MiToBo.operator.MTBOperator;

/**
 * A set of 2D regions living in a common domain which is a rectangular subset
 * of R x R. Each regions is of type Region2D
 * 
 * @modified Oct 1, 2010
 * 
 * @author Stefan Posch
 */

public class MTBRegion2DSet extends MTBData {

		@SuppressWarnings("unused")
		private boolean debug = false;

		// subset of R x R
		/**
		 * Minimal x coordinate of the domain of this region set.
		 */
		private double xMin;

		/**
		 * Minimal y coordinate of the domain of this region set.
		 */
		private double yMin;

		/**
		 * Maximal x coordinate of the domain of this region set.
		 */
		private double xMax;

		/**
		 * Maximal y coordinate of the domain of this region set.
		 */
		private double yMax;

		/**
		 * The set of regions represented as a Vector.
		 */
		Vector<MTBRegion2D> regionSet;
		
		/**
		 * Construct an empty set of regions with given extent of domain
		 */
		public MTBRegion2DSet(double _xMin, double _yMin, double _xMax, double _yMax) {
				this.xMin = _xMin;
				this.yMin = _yMin;
				this.xMax = _xMax;
				this.yMax = _yMax;

				this.regionSet = new Vector<MTBRegion2D>();

				setProperty("xMin", new Double(this.xMin));
				setProperty("yMin", new Double(this.yMin));
				setProperty("xMax", new Double(this.xMax));
				setProperty("yMax", new Double(this.yMax));
		}

		/**
		 * Construct an empty set of regions with given extent of domain
		 */
		public MTBRegion2DSet(Vector<MTBRegion2D> regions, double _xMin,
		    double _yMin, double _xMax, double _yMax) {
				this.xMin = _xMin;
				this.yMin = _yMin;
				this.xMax = _xMax;
				this.yMax = _yMax;

				this.regionSet = new Vector<MTBRegion2D>();
				for (int i = 0; i < regions.size(); i++)
						add(regions.elementAt(i));

				setProperty("xMin", new Double(this.xMin));
				setProperty("yMin", new Double(this.yMin));
				setProperty("xMax", new Double(this.xMax));
				setProperty("yMax", new Double(this.yMax));
		}

		/**
		 * Get the number of regions of this region set
		 * 
		 * @return number of regions
		 */
		public int size() {
				return this.regionSet.size();
		}

		public double getXmin() {
				return this.xMin;
		}

		public double getXmax() {
				return this.xMax;
		}

		public double getYmin() {
				return this.yMin;
		}

		public double getYmax() {
				return this.yMax;
		}

		/**
		 * Remove all regions from this set
		 */
		public void clear() {
				this.regionSet.clear();
		}

		/**
		 * Get a region by index
		 * 
		 * @return i-th region
		 */
		public MTBRegion2D get(int i) {
				return this.regionSet.get(i);
		}

		/**
		 * Get a region by index
		 * 
		 * @return i-th region
		 */
		public MTBRegion2D elementAt(int i) {
				if (i < 0 || i >= size())
						throw new ArrayIndexOutOfBoundsException();

				return this.regionSet.elementAt(i);
		}

		/**
		 * Append a region (at the end) to the set of regions.
		 * 
		 * @param region
		 *          region to add
		 */
		public boolean add(MTBRegion2D region) {
				return this.regionSet.add(region);
		}

		/**
		 * Set the i-th region from the set
		 * 
		 * @param i
		 *          index of region to remove
		 */
		public void setElementAt(MTBRegion2D region, int i) {
				this.regionSet.setElementAt(region, i);
		}

		/**
		 * Delete the i-th region from the set
		 * 
		 * @param i
		 *          index of region to remove
		 */
		public void removeElementAt(int i) {
				this.regionSet.removeElementAt(i);
		}

		/**
		 * Calculates the average size of the set.
		 * 
		 * @return average size of all regions, -1 if vector is empty
		 */
		public double calcAverageSize() {

				if (this.regionSet.size() == 0)
						return -1;

				double avgSize = 0;
				for (int i = 0; i < this.regionSet.size(); ++i) {
						MTBRegion2D r = this.regionSet.elementAt(i);
						avgSize = avgSize + r.getArea();
				}
				return avgSize / this.regionSet.size();
		}

		public MTBRegion2DSet selectLargeRegions(int minSize)
		    throws MTBOperatorException, MTBProcessingDAGException {
				RegionSelector rs = new RegionSelector(this, minSize);
				rs.runOp(null);
				return (MTBRegion2DSet) rs.getOutput("outReg");
		}

//		public MTBImageShort makeLabelImageGray(int imgW, int imgH)
//		    throws MTBOperatorException {
//				ImageMaker im = new ImageMaker(this, imgW, imgH);
//				return (MTBImageShort) im.getOutput("outImg");
//		}		
		
		private class RegionSelector extends MTBOperator {

				protected void collectArguments()  throws MTBOperatorException {
						super.collectArguments();

						MTBOpArgumentDescriptor[] inputDescriptors = new MTBOpArgumentDescriptor[] { new MTBOpArgumentDescriptor(
						    "inReg", MTBRegion2DSet.class, "Input Regions", true, null) };

						MTBOpArgumentDescriptor[] outputDescriptors = new MTBOpArgumentDescriptor[] { new MTBOpArgumentDescriptor(
						    "outReg", MTBRegion2DSet.class, "Output Region", true, null) };

						MTBOpArgumentDescriptor[] parameterDescriptors = new MTBOpArgumentDescriptor[] { new MTBOpArgumentDescriptor(
						    "minSize", java.lang.Integer.class, "type", false, null) };

						addDescriptors(parameterDescriptors, inputDescriptors,
						    outputDescriptors, null);
				}

				@SuppressWarnings("unused")
				public RegionSelector() throws MTBOperatorException {
						// nothing to do here
				}

				public RegionSelector(MTBRegion2DSet input, int msize)
				    throws MTBOperatorException {
						this.setInput("inReg", input);
						this.setParameter("minSize", new Integer(msize));
				}

				@Override
				protected void operate() throws MTBOperatorException {

						MTBRegion2DSet inregs = (MTBRegion2DSet) this.getInput("inReg");
						MTBRegion2DSet outregs = new MTBRegion2DSet(inregs.getXmin(), inregs
						    .getYmin(), inregs.getXmax(), inregs.getYmax());
						int minSize = ((Integer) this.getParameter("minSize")).intValue();
						for (int i = 0; i < inregs.size(); ++i) {
								MTBRegion2D reg = inregs.elementAt(i);
								if (reg.getArea() >= minSize)
										outregs.add(reg);
						}
						this.setOutput("outReg", outregs);
				}
		}
		
		
}
