/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.datatypes;

import java.io.*;
import java.util.*;

import de.unihalle.informatik.MiToBo.operator.MTBData;

/**
 * Class to create a 3D region object. The center of mass and the volume of the
 * 3D region can be calculated as well as other 3D region properties.
 * 
 * @modified Oct 1, 2010
 * 
 * @author glass
 */
public class MTBRegion3D extends MTBData {
		/**
		 * Points belonging to the region
		 */
		private Vector<MTBPoint3D> points;
		/**
		 * The volume of the region (number of points included).
		 */
		private int volume;
		/**
		 * The centroid of the region (not normalised by volume!)
		 */
		private MTBPoint3D centroid;
		/**
		 * ID of the region (uniqueness is not guaranteed nor checked!)
		 */
		private int id;

		/**
		 * Construct a new MTBRegion3D object.
		 */
		public MTBRegion3D() {
				this.points = new Vector<MTBPoint3D>();
				this.volume = 0;
				this.centroid = new MTBPoint3D();
		}

		/**
		 * Construct a new MTBRegion3D object from the given points.
		 * 
		 * @param p
		 *          vector of region points
		 */
		public MTBRegion3D(Vector<MTBPoint3D> p) {
				this.points = p;
				this.volume = p.size();
				this.centroid = calcCentre(p);
		}

		/**
		 * Set the ID of the region. Uniqueness is not guaranteed nor checked!
		 * 
		 * @param id
		 *          region ID
		 */
		public void setID(int id) {
				this.id = id;
		}

		/**
		 * Return the ID of the region. Uniqueness is publicnot guaranteed nor
		 * checked!
		 * 
		 * @return Region ID
		 */
		public int getID() {
				return this.id;
		}

		/**
		 * Return the points which are included in the 3D region.
		 * 
		 * @return Region points.
		 */
		public Vector<MTBPoint3D> getPoints() {
				return points;
		}

		/**
		 * Return the volume of the 3D region (number of points within the region).
		 * 
		 * @return Region volume.
		 */
		public double getVolume() {
				return volume;
		}

		/**
		 * Return the center of the region as 3D point.
		 * 
		 * @return Center of region.
		 */
		public MTBPoint3D getCentre() {
				MTBPoint3D c = new MTBPoint3D(centroid);
				c.setX(c.getX() / this.volume);
				c.setY(c.getY() / this.volume);
				c.setZ(c.getZ() / this.volume);
				return c;
		}

		/**
		 * Return the x-coordinate of the center of the region.
		 * 
		 * @return Coordinate of the center in x-dimension.
		 */
		public double getCentreX() {
				return (centroid.getX() / volume);
		}

		/**
		 * Return the y-coordinate of the center of the region.
		 * 
		 * @return Coordinate of the center in y-dimension.
		 */
		public double getCentreY() {
				return (centroid.getY() / volume);
		}

		/**
		 * Return the z-coordinate of the center of the region.
		 * 
		 * @return Coordinate of the center in z-dimension.
		 */
		public double getCentreZ() {
				return (centroid.getZ() / volume);
		}

		/**
		 * Append a 3D point to the region. The point is added at the end of the point
		 * list.
		 * 
		 * @param p
		 *          point to add
		 */
		public void addPoint(MTBPoint3D p) {
				points.addElement(p);
				volume++;
				centroid.translate(p.getX(), p.getY(), p.getZ());
		}

		/**
		 * Append a point to the region from the given coordinates. The point is added
		 * at the end of the point list.
		 * 
		 * @param px
		 *          coordinate in x-dimension
		 * @param py
		 *          coordinate in y-dimension
		 * @param pz
		 *          coordinate in z-dimension
		 */
		public void addPoint(double px, double py, double pz) {
				points.addElement(new MTBPoint3D(px, py, pz));
				volume++;
				centroid.translate(px, py, pz);
		}

		/**
		 * Calculate the centroid (not yet normalized by volume) of a given vector of
		 * points.
		 * 
		 * @param p
		 *          vector of Point3D
		 * @return The centroid (not yet normalised by volume).
		 */
		public MTBPoint3D calcCentre(Vector<MTBPoint3D> p) {
				MTBPoint3D c = new MTBPoint3D();
				for (int n = 0; n < p.size(); n++) {
						c.translate(p.elementAt(n).getX(), p.elementAt(n).getY(), p.elementAt(n)
						    .getZ());
				}
				return c;
		}

		/**
		 * Calculate the expansion in z-direction of the region.
		 * 
		 * @return Expansion of the region in z-direction.
		 */
		public int getSizeZ() {
				int max = (int) points.elementAt(0).getZ();
				int min = (int) points.elementAt(0).getZ();
				for (int i = 0; i < points.size(); i++) {
						int z = (int) points.elementAt(i).getZ();
						if (z > max) {
								max = z;
						} else if (z < min) {
								min = z;
						}
				}
				return max - min + 1;
		}

		/**
		 * Calculate the expansion in y-direction of the region.
		 * 
		 * @return Expansion of the region in y-direction.
		 */
		public int getSizeY() {
				int max = (int) points.elementAt(0).getY();
				int min = (int) points.elementAt(0).getY();
				for (int i = 0; i < points.size(); i++) {
						int y = (int) points.elementAt(i).getY();
						if (y > max) {
								max = y;
						} else if (y < min) {
								min = y;
						}
				}
				return max - min + 1;
		}

		/**
		 * Calculate the expansion in x-direction of the region.
		 * 
		 * @return Expansion of the region in x-direction.
		 */
		public int getSizeX() {
				int max = (int) points.elementAt(0).getX();
				int min = (int) points.elementAt(0).getX();
				for (int i = 0; i < points.size(); i++) {
						int x = (int) points.elementAt(i).getX();
						if (x > max) {
								max = x;
						} else if (x < min) {
								min = x;
						}
				}
				return max - min + 1;
		}

		/**
		 * Write the coordinates of the included 3D points into a text file.
		 * 
		 * @param fileName
		 *          File name to store the region on disk.
		 */
		public void regionToFile(String fileName) {
				try {
						BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));

						for (int i = 0; i < this.points.size(); i++) {
								bw.write("[" + this.points.elementAt(i).getX() + ", "
								    + this.points.elementAt(i).getY() + ", "
								    + this.points.elementAt(i).getZ() + "], ");

								if (i % 5 == 0) // line break after 5 points
								{
										bw.newLine();
								}
						}
						bw.close();
				} catch (FileNotFoundException e) {
						e.printStackTrace();
				} catch (IOException e) {
						e.printStackTrace();
				}
		}
}
