/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.datatypes.images;

import ij.ImagePlus;
import ij.ImageStack;
import ij.WindowManager;
import ij.gui.ImageWindow;
import ij.gui.StackWindow;
import ij.measure.Calibration;
import ij.plugin.Zoom;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.exceptions.MTBProcessingDAGException;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException.OperatorExceptionType;
import de.unihalle.informatik.MiToBo.operator.MTBData;
import de.unihalle.informatik.MiToBo.operator.MTBOpArgumentDescriptor;
import de.unihalle.informatik.MiToBo.operator.MTBOperator;

/**
 * Abstract wrapper class for easy access to hyperstacks. Indices range is
 * different from ImageJ. Here, indices in each dimension range from 0 to
 * (dimSize - 1), while ImageJ stack indices range from 1 to dimSize. Subclasses
 * implement the wrapper for different types.
 * 
 * @author gress
 * 
 */
public abstract class MTBImage extends MTBData implements MTBImageManipulator, KeyListener {

  // ----- member fields

  /** the underlying ImagePlus object */
  protected ImagePlus m_img;

  /** reference to the ImageStack object of the underlying ImagePlus object */
  protected ImageStack m_imgStack;

  /** size of x-dimension */
  protected int m_sizeX;

  /** size of y-dimension */
  protected int m_sizeY;

  /** size of z-dimension */
  protected int m_sizeZ;

  /** size of t-dimension */
  protected int m_sizeT;

  /** size of c-dimension */
  protected int m_sizeC;

  /** stacksize */
  protected int m_sizeStack;

  /** MiToBo image type */
  protected MTBImageType m_type;

  /**
   * index of the actual slice, this has nothing to do with the ij.ImageStack's
   * current slice
   */
  protected int m_actSliceIdx;

  /** actual t coordinate, for 3D only access functions */
  protected int m_actT;

  /** actual c coordinate, for 3D only access functions */
  protected int m_actC;

  /** image title */
  protected String m_title;

  /** physical voxel size (stepsize) in x-dimension */
  protected double m_stepX;

  /** physical voxel size (stepsize) in y-dimension */
  protected double m_stepY;

  /** physical voxel size (stepsize) in x-dimension */
  protected double m_stepZ;

  /** stepsize in t-dimension */
  protected double m_stepT;

  /** spatial x-unit */
  protected String m_unitX;

  /** spatial y-unit */
  protected String m_unitY;

  /** spatial z-unit */
  protected String m_unitZ;

  /** time unit */
  protected String m_unitT;

  
  /** MiToBo image types */
  public enum MTBImageType {
	  MTB_BYTE, MTB_SHORT, MTB_INT, MTB_FLOAT, MTB_DOUBLE, MTB_RGB
  }
  
  /** methods to create new image from an existing one. used in MTBImageFactory */
  protected enum FactoryMethod {
	  DUPLICATE, CONVERT, SLICE, IMAGE_PART
  }


  // ----- Constructors

  /**
   * Constructor. Initializes the sizes corresponding to the underlying
   * ImagePlus. Actual slice index, actual z-stack coordinates are all
   * initialized to 0. If a Calibration object is present in the ImagePlus,
   * physical pixel sizes and units are assigned.
   * 
   * @param wrapped
   *          ImagePlus object
   */
  protected MTBImage(ImagePlus img) {

    // reference to ImagePlus object
    m_img = img;
    m_imgStack = m_img.getStack();

    // dimension sizes
    m_sizeX = m_img.getWidth();
    m_sizeY = m_img.getHeight();
    m_sizeZ = m_img.getNSlices();
    m_sizeT = m_img.getNFrames();
    m_sizeC = m_img.getNChannels();
    
    this.setProperty("SizeX", m_sizeX);
    this.setProperty("SizeY", m_sizeY);
    this.setProperty("SizeZ", m_sizeZ);
    this.setProperty("SizeT", m_sizeT);
    this.setProperty("SizeC", m_sizeC);
    

    m_sizeStack = m_img.getStackSize();

    m_actSliceIdx = 0;

    m_actT = 0;
    m_actC = 0;

   // m_type = -1;

    m_title = img.getTitle();

    Calibration cal = img.getCalibration();
    if (cal != null) {
      this.setStepsizeX(cal.pixelWidth);
      this.setStepsizeY(cal.pixelHeight);
      this.setStepsizeZ(cal.pixelDepth);
      this.setStepsizeT(cal.frameInterval);

      this.setUnitX(cal.getXUnit());
      this.setUnitY(cal.getYUnit());
      this.setUnitZ(cal.getZUnit());
      this.setUnitT(cal.getTimeUnit());
    } else {
      this.setStepsizeX(1.0);
      this.setStepsizeY(1.0);
      this.setStepsizeZ(1.0);
      this.setStepsizeT(1.0);

      this.setUnitX("");
      this.setUnitY("");
      this.setUnitZ("");
      this.setUnitT("");
    }
    
    if (m_img.getWindow() != null)
    	if (m_img.getWindow().getKeyListeners().length == 0)
    		m_img.getWindow().addKeyListener(this);
  }

  /**
   * Constructor for initializing an empty image (dimension sizes = -1). Actual
   * slice index, actual z-stack coordinates are all initialized to 0. Needed
   * for non-ImageJ types, which cannot be represented by an ImagePlus (e.g.
   * Double)
   * 
   * @param wrapped
   *          ImagePlus object
   */
  protected MTBImage() {

    // reference to ImagePlus object
    m_img = null;
    m_imgStack = null;

    // dimension sizes
    m_sizeX = -1;
    m_sizeY = -1;
    m_sizeZ = -1;
    m_sizeT = -1;
    m_sizeC = -1;

    m_sizeStack = -1;

    m_actSliceIdx = 0;

    m_actT = 0;
    m_actC = 0;

   // m_type = -1;

    m_title = "";

    this.setStepsizeX(1.0);
    this.setStepsizeY(1.0);
    this.setStepsizeZ(1.0);
    this.setStepsizeT(1.0);

    this.setUnitX("");
    this.setUnitY("");
    this.setUnitZ("");
    this.setUnitT("");
  }

  // ----- Data access functions
  /**
   * Checks if the given type identifier is valid.
   * 
   * @param type
   *          Integer specifying a MTBImage type.
   * @return True, if type integer is valid.
   */
  @Deprecated
  public static boolean isValidType(MTBImageType type) {
    return (type == MTBImageType.MTB_BYTE) || (type == MTBImageType.MTB_SHORT) || (type == MTBImageType.MTB_INT)
        || (type == MTBImageType.MTB_FLOAT) || (type == MTBImageType.MTB_DOUBLE) || (type == MTBImageType.MTB_RGB);
  }

  /**
   * Returns type identifier for given id.
   * 
   * @param type
   *          Numerical identifier for image.
   * @return Identifier string, null is non-existent.
   */
  @Deprecated
  public static String getTypeName(int type) {
    switch (type) {
      case 0:
        return "MTB_BYTE";
      case 1:
        return "MTB_SHORT";
      case 3:
        return "MTB_INT";
      case 4:
        return "MTB_FLOAT";
      case 5:
        return "MTB_DOUBLE";
      case 6:
        return "MTB_RGB";
    }
    return null;
  }

  /**
   * Test if this object and img have the same size in x-, y-, z-, t- and
   * c-dimension
   * 
   * @param img
   *          image to compare with
   * @return true if this object and img have the same size, else false
   */
  public boolean equalSize(MTBImage img) {

    return (this.m_sizeX == img.m_sizeX && this.m_sizeY == img.m_sizeY
        && this.m_sizeZ == img.m_sizeZ && this.m_sizeT == img.m_sizeT && this.m_sizeC == img.m_sizeC);
  }

  /**
   * Test if this object and img are of same type, have same size and same
   * values
   * 
   * @param img
   *          image to compare with
   * @return true if same type, size and values
   */
  public boolean equals(MTBImage img) {

    if (this.m_type != img.m_type) {
      return false;
    }

    if (!this.equalSize(img)) {
      return false;
    }

    boolean equal = true;
    int tIdx = this.m_actSliceIdx;
    int iIdx = img.m_actSliceIdx;
    for (int i = 0; i < this.m_sizeStack && equal; i++) {
      this.m_actSliceIdx = i;
      img.m_actSliceIdx = i;

      for (int y = 0; y < this.m_sizeY && equal; y++) {
        for (int x = 0; x < this.m_sizeX && equal; x++) {
          if (this.getValueDouble(x, y) != img.getValueDouble(x, y)) {
            equal = false;
          }
        }
      }
    }
    this.m_actSliceIdx = tIdx;
    img.m_actSliceIdx = iIdx;

    return equal;
  }

  /**
   * Get the physical size of a voxel (stepsize) in x-dimension
   * 
   * @return physical voxel size in x-dimension
   */
  public double getStepsizeX() {
    return m_stepX;
  }

  /**
   * Get the physical size of a voxel (stepsize) in y-dimension
   * 
   * @return physical voxel size in y-dimension
   */
  public double getStepsizeY() {
    return m_stepY;
  }

  /**
   * Get the physical size of a voxel (stepsize) in z-dimension
   * 
   * @return physical voxel size in z-dimension
   */
  public double getStepsizeZ() {
    return m_stepZ;
  }

  /**
   * Get the stepsize in t-dimension (timestep)
   * 
   * @return time stepsize
   */
  public double getStepsizeT() {
    return m_stepT;
  }

  /**
   * Set the physical size of a voxel (stepsize) in x-dimension
   * 
   * @param stepsize
   */
  public void setStepsizeX(double stepsize) {
    m_stepX = stepsize;
    this.setProperty("StepsizeX", stepsize);
  }

  /**
   * Set the physical size of a voxel (stepsize) in y-dimension
   * 
   * @param stepsize
   */
  public void setStepsizeY(double stepsize) {
    m_stepY = stepsize;
    this.setProperty("StepsizeY", stepsize);
  }

  /**
   * Set the physical size of a voxel (stepsize) in z-dimension
   * 
   * @param stepsize
   */
  public void setStepsizeZ(double stepsize) {
    m_stepZ = stepsize;
    this.setProperty("StepsizeZ", stepsize);
  }

  /**
   * Set the stepsize in t-dimension (timestep)
   * 
   * @param stepsize
   */
  public void setStepsizeT(double stepsize) {
    m_stepT = stepsize;
    this.setProperty("StepsizeT", stepsize);
  }

  /**
   * Get the unit of the x-dimension
   * 
   * @return String of x-dimension's unit
   */
  public String getUnitX() {
    return m_unitX;
  }

  /**
   * Get the unit of the y-dimension
   * 
   * @return String of y-dimension's unit
   */
  public String getUnitY() {
    return m_unitY;
  }

  /**
   * Get the unit of the z-dimension
   * 
   * @return String of z-dimension's unit
   */
  public String getUnitZ() {
    return m_unitZ;
  }

  /**
   * Get the unit of the t-dimension
   * 
   * @return String of t-dimension's unit
   */
  public String getUnitT() {
    return m_unitT;
  }

  /**
   * Set the unit of the x-dimension
   * 
   * @param unit
   *          String of x-dimension unit
   */
  public void setUnitX(String unit) {
    m_unitX = unit;
    this.setProperty("UnitX", unit);
  }

  /**
   * Set the unit of the y-dimension
   * 
   * @param unit
   *          String of y-dimension unit
   */
  public void setUnitY(String unit) {
    m_unitY = unit;
    this.setProperty("UnitY", unit);
  }

  /**
   * Set the unit of the z-dimension
   * 
   * @param unit
   *          String of z-dimension unit
   */
  public void setUnitZ(String unit) {
    m_unitZ = unit;
    this.setProperty("UnitZ", unit);
  }

  /**
   * Set the unit of the t-dimension
   * 
   * @param unit
   *          String of t-dimension unit
   */
  public void setUnitT(String unit) {
    m_unitT = unit;
    this.setProperty("UnitT", unit);
  }
  
  /**
   * Copy physical properties like stepsizes and units from sourceImg to this. 
   * @param sourceImg
   */
  public void copyPhysicalProperties(MTBImage sourceImg) {
	  this.setStepsizeX(sourceImg.m_stepX);
	  this.setStepsizeY(sourceImg.m_stepY);
	  this.setStepsizeZ(sourceImg.m_stepZ);
	  this.setStepsizeT(sourceImg.m_stepT);
	  this.setUnitX(sourceImg.m_unitX);
	  this.setUnitY(sourceImg.m_unitY);
	  this.setUnitZ(sourceImg.m_unitZ);
	  this.setUnitT(sourceImg.m_unitT); 
  }

  /**
   * Update physical properties (stepsizes, units) of the image (member
   * variables) from its properties hashtable
   */
  protected void updatePhysProperties_PropToImg() {

    this.m_stepX = Double.parseDouble(this.getProperty("StepsizeX"));
    this.m_stepY = Double.parseDouble(this.getProperty("StepsizeY"));
    this.m_stepZ = Double.parseDouble(this.getProperty("StepsizeZ"));
    this.m_stepT = Double.parseDouble(this.getProperty("StepsizeT"));

    this.m_unitX = this.getProperty("UnitX");
    this.m_unitY = this.getProperty("UnitY");
    this.m_unitZ = this.getProperty("UnitZ");
    this.m_unitT = this.getProperty("UnitT");

  }

  /**
   * Update physical properties (stepsizes, units) in the properties hashtable
   * from its member variables
   */
  protected void updatePhysProperties_ImgToProp() {

    this.setProperty("StepsizeX", m_stepX);
    this.setProperty("StepsizeY", m_stepY);
    this.setProperty("StepsizeZ", m_stepZ);
    this.setProperty("StepsizeT", m_stepT);

    this.setProperty("UnitX", m_unitX);
    this.setProperty("UnitY", m_unitY);
    this.setProperty("UnitZ", m_unitZ);
    this.setProperty("UnitT", m_unitT);
  }

  /**
   * Update the image size in the properties hashtable from the image's size
   */
  protected void updateImageSize_ImgToProp() {
    this.setProperty("SizeX", m_sizeX);
    this.setProperty("SizeY", m_sizeY);
    this.setProperty("SizeZ", m_sizeZ);
    this.setProperty("SizeT", m_sizeT);
    this.setProperty("SizeC", m_sizeT);
  }

  /**
   * Get the image's title
   */
  public String getTitle() {
    return m_title;
  }

  /**
   * Set the image's title
   * 
   * @param title
   */
  public void setTitle(String title) {
    m_title = title;

    if (m_img != null) {
      m_img.setTitle(title);
    }
  }

  /**
   * Get size of x-dimension
   * 
   * @return size of x-dimension
   */
  public int getSizeX() {
    return m_sizeX;
  }

  /**
   * Get size of y-dimension
   * 
   * @return size of y-dimension
   */
  public int getSizeY() {
    return m_sizeY;
  }

  /**
   * Get size of z-dimension
   * 
   * @return size of z-dimension
   */
  public int getSizeZ() {
    return m_sizeZ;
  }

  /**
   * Get size of t(ime)-dimension
   * 
   * @return size of t(ime)-dimension
   */
  public int getSizeT() {
    return m_sizeT;
  }

  /**
   * Get size of c(hannel)-dimension
   * 
   * @return size of c(hannel)-dimension
   */
  public int getSizeC() {
    return m_sizeC;
  }

  /**
   * Get MiToBo image type
   * 
   * @return image type (MTBImage.MTB_BYTE, MTBImage.MTB_SHORT,
   *         MTBImage.MTB_INT, MTBImage.MTB_FLOAT, MTBImage.MTB_DOUBLE)
   */
  public MTBImageType getType() {
    return m_type;
  }

  /**
   * Get size of the underlying (ImageJ) stack which corresponds to the number
   * of slices
   * 
   * @return size of c(hannel)-dimension
   */
  public int getSizeStack() {
    return m_sizeStack;
  }

  /**
   * Return the maximum value that the current type can handle
   * 
   * @return
   */
  public double getTypeMax() {
	  
    if (this.m_type == MTBImageType.MTB_BYTE)
    	return ((double) Byte.MAX_VALUE) * 2.0 + 1.0;
    else if (this.m_type == MTBImageType.MTB_SHORT)
        return ((double) Short.MAX_VALUE) * 2.0 + 1.0;
    else if (this.m_type == MTBImageType.MTB_INT)
        return (double) Integer.MAX_VALUE;
    else if (this.m_type == MTBImageType.MTB_FLOAT)
        return (double) Float.MAX_VALUE;
    else if (this.m_type == MTBImageType.MTB_DOUBLE)
        return Double.MAX_VALUE;
    else if (this.m_type == MTBImageType.MTB_RGB)
        return ((double) Byte.MAX_VALUE) * 2.0 + 1.0;
    else
    	return -1.0;
    
  }

  /**
   * Return the minimum value that the current type can handle
   * 
   * @return
   */
  public double getTypeMin() {
	  if (this.m_type == MTBImageType.MTB_BYTE)
        return 0.0;
	  else if (this.m_type == MTBImageType.MTB_SHORT)
        return 0.0;
	  else if (this.m_type == MTBImageType.MTB_INT)
        return (double) Integer.MIN_VALUE;
	  else if (this.m_type == MTBImageType.MTB_FLOAT)
        return (double) Float.MIN_VALUE;
	  else if (this.m_type == MTBImageType.MTB_DOUBLE)
        return Double.MIN_VALUE;
	  else if (this.m_type == MTBImageType.MTB_RGB)
        return 0.0;
	  else
        return -1.0;
  }

  /**
   * Get the actual slice coordinates (z,t,c) (for functions that work on a
   * slice only, this has nothing to do with the ij.ImageStack's current slice)
   * 
   * @return (z,t,c)-coordinate of the actual slice. [0]= z-coordinate, [1]=
   *         t-coordinate, [2]= c-coordinate
   */
  public int[] getActualSliceCoords() {
    int[] coords = new int[3];

    coords[0] = ((int) m_actSliceIdx / m_sizeC) % m_sizeZ;
    coords[1] = m_actSliceIdx / (m_sizeC * m_sizeZ);
    coords[2] = m_actSliceIdx % m_sizeC;

    return coords;
  }

  /**
   * Get the actual slice index (for functions that work on a slice only, this
   * has nothing to do with the ij.ImageStack's current slice)
   * 
   * @return (z,t,c)-coordinate of the actual slice. [0]= z-coordinate, [1]=
   *         t-coordinate, [2]= c-coordinate
   */
  public int getActualSliceIndex() {

    return m_actSliceIdx;
  }

  /**
   * Set the actual slice for functions that work on a slice only (this has
   * nothing to do with the ij.ImageStack's current slice)
   * 
   * @param z
   * @param t
   * @param c
   */
  public void setActualSliceCoords(int z, int t, int c) {
    m_actSliceIdx = t * m_sizeC * m_sizeZ + z * m_sizeC + c;
  }

  /**
   * Set the actual slice for functions that work on a slice only (this has
   * nothing to do with the ij.ImageStack's current slice) Stack indices range
   * from 0 to N-1 (unlike ImageJ, where stack indices range from 1 to N)
   * 
   * @param stackindex
   *          index of the slice in the underlying (ImageJ) stack
   */
  public void setActualSliceIndex(int stackindex) {
    m_actSliceIdx = stackindex;
  }

  /**
   * Get the actual z-stack coordinates. [0]=t-coordinate, [1]=c-coordinate.
   * They are independent of the actual slice index.
   * 
   * @return array with the actual t and c coordinates ([0]= t-coordinate, [1]=
   *         c-coordinate)
   */
  public int[] getActualZStackCoords() {
    int[] actTC = { m_actT, m_actC };
    return actTC;
  }

  /**
   * Set the actual z-stack coordinates for functions that work on (spatial) 3D
   * only. The coordinates are not checked for their validity. They are
   * independent of the actual slice index. E.g. after setting z-stack
   * coordinates to 't' and 'c', the functions getValueInt(x,y,z) is equivalent
   * to getValueInt(x,y,z,'t','c').
   * 
   * @param t
   *          actual t-coordinate
   * @param c
   *          actual c-coordinate
   */
  public void setActualZStackCoords(int t, int c) {
    m_actT = t;
    m_actC = c;
  }

  /**
   * Display the image as an ImagePlus. Keep in mind that the MiToBo image types
   * MTB_INT and MTB_DOUBLE have no equivalent ImageJ types. Thus this function
   * creates an ImagePlus object of a reduced precision type (short/float) for
   * display. These ImagePlus objects do NOT share data and thus are uncoupled
   * of the MTBImage. You can find the MTBImage in the ImagePlus' properties
   * with the key "MTBImage"
   */
  public void show() {

    this.updateImagePlus();
    // reset the adjustment
    m_img.getProcessor().resetMinAndMax();
    // show the image
    m_img.show();

    ImageWindow w = m_img.getWindow();
    if (w instanceof StackWindow) {
      int actslice = m_img.getCurrentSlice() - 1;
      ((StackWindow) w).setPosition((actslice % m_sizeC) + 1,
          ((int) ((actslice / m_sizeC)) % m_sizeZ) + 1,
          (actslice / (m_sizeC * m_sizeZ)) + 1);
      ((StackWindow) w).updateSliceSelector();
    }
    
 //   KeyListener[] kl = w.getKeyListeners();
 //   boolean containsThis = false;
  //  for (int i = 0; i < kl.length && (containsThis == false); i++) {
 //   	if (kl[i] == this)
 //   		containsThis = true;
 //   }
    
    if (w.getKeyListeners().length == 0)
    	w.addKeyListener(this);
  }

  /**
   * Update and repaint the image window if the image is displayed by ImageJ
   */
  public void updateAndRepaintWindow() {
    this.updateImagePlus();
    m_img.updateAndRepaintWindow();
    
//    ImageWindow w = m_img.getWindow();
//    if (w != null) {
//    	System.out.println("# of key listeners: " + w.getKeyListeners().length);
//    	if (w.getKeyListeners().length == 0)
//    		w.addKeyListener(this);
//    }
  }

  /**
   * Close the displayed ImagePlus
   */
  public void close() {
    if (m_img != null) {
      m_img.close();
    }
  }
  
  /**
   * Create an image of given type from this image's values. Information is lost
   * if a type of less precision is used!
   * This method uses the internal MTBImageFactory, which extends MTBOperator,
   * thus the new image will share the history. If you need to specify the calling operator, use getImagePart(MTBOperator, ...)
   * 
   * @param type
   *          new image's type (see static final fields for types)
   * @param scaleDown
   *          If true, the data is scaled down to fit in the range of the new
   *          image type values, if the new image type is of less precision. If
   *          false, the values are simply casted.
   * @return new image of given type or null if the operation was not successful
   */
  public MTBImage convertType(MTBImageType type, boolean scaleDown) {
	  return this.convertType(null, type, scaleDown);
  }
  
  /**
   * Create an image of given type from this image's values. Information is lost
   * if a type of less precision is used!
   * This method uses the internal MTBImageFactory, which extends MTBOperator,
   * thus the new image will share the history.
   * 
   * @param callingOperator the MTBOperator, which calls this method
   * @param type
   *          new image's type (see static final fields for types)
   * @param scaleDown
   *          If true, the data is scaled down to fit in the range of the new
   *          image type values, if the new image type is of less precision. If
   *          false, the values are simply casted.
   * @return new image of given type or null if the operation was not successful
   */
  public MTBImage convertType(MTBOperator callingOperator, MTBImageType type, boolean scaleDown) {

	  MTBImageFactory factory = null;
	  MTBImage newImg = null;
	  
	  try {
		  
		factory = new MTBImageFactory(this, type, scaleDown);
		
		factory.runOp(callingOperator);
		
		newImg = factory.getResultImg();
		
	  } catch (MTBOperatorException e) {
		e.printStackTrace();
	  } catch (MTBProcessingDAGException e) {
		e.printStackTrace();
	  }
	  
	  return newImg; 
  }

  
  /**
   * Get a copy of a part of this image as new MTBImage. This method uses the internal MTBImageFactory, which extends MTBOperator,
   * thus the new image will share the history. If you need to specify the calling operator, use getImagePart(MTBOperator, ...)
   * 
   * @param x
   *          x-coordinate where the first value is copied from
   * @param y
   *          y-coordinate where the first value is copied from
   * @param z
   *          z-coordinate where the first value is copied from
   * @param t
   *          t-coordinate where the first value is copied from
   * @param c
   *          c-coordinate where the first value is copied from
   * @param sizeX
   *          size of the copied part in x-dimension
   * @param sizeY
   *          size of the copied part in y-dimension
   * @param sizeZ
   *          size of the copied part in z-dimension
   * @param sizeT
   *          size of the copied part in t-dimension
   * @param sizeC
   *          size of the copied part in c-dimension
   * @return new MTBImage with values equal to the specified part or null if the operation was not successful
   * @throws IllegalArgumentException
   *           if image boundaries are exceeded in any way
   */
  public MTBImage getImagePart(int x, int y, int z, int t, int c, int sizeX,
		  							int sizeY, int sizeZ, int sizeT, int sizeC)
      							throws IllegalArgumentException {

	  return this.getImagePart(null, x, y, z, t, c, sizeX, sizeY, sizeZ, sizeT, sizeC);
  }
  
  
  /**
   * Get a copy of a part of this image as new MTBImage. This method uses the internal MTBImageFactory, which extends MTBOperator,
   * thus the new image will share the history.
   * 
   * @param callingOperator the MTBOperator, which calls this method
   * @param x
   *          x-coordinate where the first value is copied from
   * @param y
   *          y-coordinate where the first value is copied from
   * @param z
   *          z-coordinate where the first value is copied from
   * @param t
   *          t-coordinate where the first value is copied from
   * @param c
   *          c-coordinate where the first value is copied from
   * @param sizeX
   *          size of the copied part in x-dimension
   * @param sizeY
   *          size of the copied part in y-dimension
   * @param sizeZ
   *          size of the copied part in z-dimension
   * @param sizeT
   *          size of the copied part in t-dimension
   * @param sizeC
   *          size of the copied part in c-dimension
   * @return new MTBImage with values equal to the specified part or null if the operation was not successful
   * @throws IllegalArgumentException
   *           if image boundaries are exceeded in any way
   */
  public MTBImage getImagePart(MTBOperator callingOperator, int x, int y, int z, int t, int c, int sizeX,
		  							int sizeY, int sizeZ, int sizeT, int sizeC)
      							throws IllegalArgumentException {

    if (x + sizeX > this.m_sizeX || y + sizeY > this.m_sizeY
        || z + sizeZ > this.m_sizeZ || t + sizeT > this.m_sizeT
        || c + sizeC > this.m_sizeC) {
      throw new IllegalArgumentException(
          "MTBImage.getImagePart(..): Specified image part exceeds the image boundaries.");
    } else if (x < 0 || x >= this.m_sizeX || y < 0 || y >= this.m_sizeY
        || z < 0 || z >= this.m_sizeZ || t < 0 || t >= this.m_sizeT || c < 0
        || c >= this.m_sizeC) {
      throw new IllegalArgumentException(
          "MTBImage.getImagePart(..): Specified coordinate exceeds the image boundaries.");
    } else {

  	  MTBImageFactory factory = null;
	  MTBImage newImg = null;
	  
	  try {
		  
		factory = new MTBImageFactory(this, x, y, z, t, c, sizeX, sizeY, sizeZ, sizeT, sizeC);
		
		factory.runOp(callingOperator);
		
		newImg = factory.getResultImg();
		
	  } catch (MTBOperatorException e) {
		e.printStackTrace();
	  } catch (MTBProcessingDAGException e) {
		e.printStackTrace();
	  }

      return newImg;
    }
  }

  /**
   * Set a part of this image to the values of the source
   * 
   * @param src
   *          source MTBImage (or MTBImageWindow)
   * @param x
   *          x-coordinate where the first value is copied to
   * @param y
   *          y-coordinate where the first value is copied to
   * @param z
   *          z-coordinate where the first value is copied to
   * @param t
   *          t-coordinate where the first value is copied to
   * @param c
   *          c-coordinate where the first value is copied to
   * @throws IllegalArgumentException
   *           if image boundaries are exceeded in any way
   */
  public void setImagePart(MTBImageManipulator src, int x, int y, int z, int t,
      int c) throws IllegalArgumentException {
    if (x + src.getSizeX() > this.m_sizeX || y + src.getSizeY() > this.m_sizeY
        || z + src.getSizeZ() > this.m_sizeZ
        || t + src.getSizeT() > this.m_sizeT
        || c + src.getSizeC() > this.m_sizeC) {
      throw new IllegalArgumentException(
          "MTBImage.setImagePart(..): Specified position with the size of source exceed target size");
    } else if (x < 0 || x >= this.m_sizeX || y < 0 || y >= this.m_sizeY
        || z < 0 || z >= this.m_sizeZ || t < 0 || t >= this.m_sizeT || c < 0
        || c >= this.m_sizeC) {
      throw new IllegalArgumentException(
          "MTBImage.setImagePart(..): Specified coordinate exceeds the image boundaries.");
    } else {
      int sx = src.getSizeX();
      int sy = src.getSizeY();
      int sz = src.getSizeZ();
      int st = src.getSizeT();
      int sc = src.getSizeC();

      for (int cc = 0; cc < sc; cc++) {
        for (int tt = 0; tt < st; tt++) {
          for (int zz = 0; zz < sz; zz++) {
            for (int yy = 0; yy < sy; yy++) {
              for (int xx = 0; xx < sx; xx++) {
                this.putValueDouble(x + xx, y + yy, z + zz, t + tt, c + cc, src
                    .getValueDouble(xx, yy, zz, tt, cc));
              }
            }
          }
        }
      }
    }
  }

  /**
   * Get a copy of the specified slice. This method uses the internal MTBImageFactory, which extends MTBOperator,
   * thus the new image will share the history. If you need to specify the calling operator, use getSlice(MTBOperator, ...)
   * 
   * @param z
   * @param t
   * @param c
   * @return copy of specified slice or null if the operation was not successful
   * @throws IllegalArgumentException if z, c or t do not match the image size
   */
  public MTBImage getSlice(int z, int t, int c) throws IllegalArgumentException {
	  return this.getSlice(null, z, t, c);
  }

  /**
   * Get a copy of the specified slice. This method uses the internal MTBImageFactory, which extends MTBOperator,
   * thus the new image will share the history.
   * 
   * @param callingOperator the MTBOperator, which calls this method
   * @param z
   * @param t
   * @param c
   * @return copy of specified slice or null if the operation was not successful
   * @throws IllegalArgumentException if z, c or t do not match the image size
   */
  public MTBImage getSlice(MTBOperator callingOperator, int z, int t, int c) throws IllegalArgumentException {
	  
	  if (z < 0 || z >= this.m_sizeZ || t < 0 || t >= this.m_sizeT || c < 0 || c >= this.m_sizeC)
		  throw new IllegalArgumentException("Invalid slice coordinate. (z,c,t)=("+z+","+t+","+c+") is outside the image domain.");
	  
  	  MTBImageFactory factory = null;
	  MTBImage newImg = null;
	  
	  try {
		  
		factory = new MTBImageFactory(this, z, t, c);
		
		factory.runOp(callingOperator);
		
		newImg = factory.getResultImg();
		
	  } catch (MTBOperatorException e) {
		e.printStackTrace();
	  } catch (MTBProcessingDAGException e) {
		e.printStackTrace();
	  }

      return newImg;
  }
  
  /**
   * Get a copy of the actual slice. This method uses the internal MTBImageFactory, which extends MTBOperator,
   * thus the new image will share the history. If you need to specify the calling operator, use getActualSlice(MTBOperator)
   * 
   * @return copy of actual slice or null if the operation was not successful
   * @throws IllegalArgumentException
   */
  public MTBImage getActualSlice() {
	  int[] sliceCoord = this.getActualSliceCoords();	  
	  MTBImage newImg = null;
	  
	  try {
		  
		  newImg = this.getSlice(null, sliceCoord[0], sliceCoord[1], sliceCoord[2]);
		  
	  } catch (IllegalArgumentException e) {
		  e.printStackTrace();
	  }
	  
	  return newImg;
  }
  
  /**
   * Get a copy of the actual slice. This method uses the internal MTBImageFactory, which extends MTBOperator,
   * thus the new image will share the history.
   * 
   * @param callingOperator the MTBOperator, which calls this method
   * @return copy of actual slice or null if the operation was not successful
   * @throws IllegalArgumentException
   */
  public MTBImage getActualSlice(MTBOperator callingOperator) {
	  int[] sliceCoord = this.getActualSliceCoords();
	  MTBImage newImg = null;
	  
	  try {
		  
		  newImg = this.getSlice(callingOperator, sliceCoord[0], sliceCoord[1], sliceCoord[2]);
		  
	  } catch (IllegalArgumentException e) {
		  e.printStackTrace();
	  }
	  
	  return newImg;
  }
  
  /**
   * Copy the value of source to the specified slice
   * 
   * @param src
   * @param z
   * @param t
   * @param c
   * @throws IllegalArgumentException
   */
  public void setSlice(MTBImageManipulator src, int z, int t, int c)
      throws IllegalArgumentException {
    try {
      this.setImagePart(src, 0, 0, z, t, c);
    } catch (IllegalArgumentException e) {
      throw e;
    }
  }

  /**
   * Copy the values of the source to the actual slice
   * 
   * @param src
   * @throws IllegalArgumentException
   */
  public void setActualSlice(MTBImageManipulator src)
      throws IllegalArgumentException {
    try {
      int[] sliceCoord = this.getActualSliceCoords();
      this.setImagePart(src, 0, 0, sliceCoord[0], sliceCoord[1], sliceCoord[2]);
    } catch (IllegalArgumentException e) {
      throw e;
    }
  }

  /**
   * Returns the String "IMG("title of the image")"
   */
  public String toString() {
    String s = "IMG(\"" + m_title + "\")";

    /*
     * for (int c = 0; c < m_sizeC; c++) {
     * 
     * s += "("; for (int t = 0; t < m_sizeT; t++) {
     * 
     * s += "("; for (int z = 0; z < m_sizeZ; z++) {
     * 
     * s += "("; for (int y = 0; y < m_sizeY; y++) {
     * 
     * s += "("; for (int x = 0; x < m_sizeX; x++) { if (x == 0) { s +=
     * this.getValueDouble(x, y, z, t, c); } else { s += ", " +
     * this.getValueDouble(x, y, z, t, c); } } s += ")";
     * 
     * } s += ")";
     * 
     * } s += ")";
     * 
     * } s += ")";
     * 
     * }
     */

    return s;
  }

  /**
   * Draws a 2D line into the actual slice of the image.
   * <p>
   * MTBImages are 5D, but here t- and c-dimensions are ignored. This function
   * implements the Bresenham algorithm. Code was 'stolen' from Wikipedia,
   * {@link http://de.wikipedia.org/wiki/Bresenham-Algorithmus}, and then
   * translated into Java (German comments where kept).
   * 
   * @param xstart
   *          x-coordinate of start point.
   * @param ystart
   *          y-coordinate of start point.
   * @param xend
   *          x-coordinate of end point.
   * @param yend
   *          y-coordinate of end point.
   * @param value
   *          Color/gray-scale value of the polygon.
   */
  public void drawLine2D(int xstart, int ystart, int xend, int yend, int value) {

    int x, y, t, dx, dy, incx, incy, pdx, pdy, ddx, ddy, es, el, err;

    /* Entfernung in beiden Dimensionen berechnen */
    dx = xend - xstart;
    dy = yend - ystart;

    /* Vorzeichen des Inkrements bestimmen */
    incx = (int) Math.signum(dx);
    incy = (int) Math.signum(dy);
    if (dx < 0)
      dx = -dx;
    if (dy < 0)
      dy = -dy;

    /* feststellen, welche Entfernung größer ist */
    if (dx > dy) {
      /* x ist schnelle Richtung */
      pdx = incx;
      pdy = 0; /* pd. ist Parallelschritt */
      ddx = incx;
      ddy = incy; /* dd. ist Diagonalschritt */
      es = dy;
      el = dx; /* Fehlerschritte schnell, langsam */
    } else {
      /* y ist schnelle Richtung */
      pdx = 0;
      pdy = incy; /* pd. ist Parallelschritt */
      ddx = incx;
      ddy = incy; /* dd. ist Diagonalschritt */
      es = dx;
      el = dy; /* Fehlerschritte schnell, langsam */
    }

    /* Initialisierungen vor Schleifenbeginn */
    x = xstart;
    y = ystart;
    err = el / 2;
    if (x >= 0 && x < this.getSizeX() && y >= 0 && y < this.getSizeY())
      this.putValueInt(x, y, value);

    /* Pixel berechnen */
    for (t = 0; t < el; ++t) /* t zaehlt die Pixel, el ist auch Anzahl */
    {
      /* Aktualisierung Fehlerterm */
      err -= es;
      if (err < 0) {
        /* Fehlerterm wieder positiv (>=0) machen */
        err += el;
        /* Schritt in langsame Richtung, Diagonalschritt */
        x += ddx;
        y += ddy;
      } else {
        /* Schritt in schnelle Richtung, Parallelschritt */
        x += pdx;
        y += pdy;
      }
      if (x >= 0 && x < this.getSizeX() && y >= 0 && y < this.getSizeY())
        this.putValueInt(x, y, value);
    }
  }
  
  /**
   * Get the slice label of the slice specified by the actual slice index
   * @return
   */
  public String getActualSliceLabel() {
	  if (m_img != null) {
		  return m_img.getStack().getSliceLabel(this.m_actSliceIdx + 1);
	  }
	  else {
		  return null;
	  }
  }
  
  /**
   * Get the slice label of the slice specified by (z,t,c)
   * @param z
   * @param t
   * @param c
   * @return
   */
  public String getSliceLabel(int z, int t, int c) {
	  int tmpIdx = this.m_actSliceIdx;
	  this.setActualSliceCoords(z, t, c);
	  
	  String s = this.getActualSliceLabel();
	  this.m_actSliceIdx = tmpIdx;
	  
	  return s;
  }
  
  /**
   * Set the slice label of the slice specified by the actual slice index
   * @param label
   */
  public void setActualSliceLabel(String label) {
	  if (m_img != null) {
		  m_img.getStack().setSliceLabel(label, this.m_actSliceIdx + 1);
	  } 
  }
  
  /**
   * Set the slice label of the slice specified by (z,t,c)
   * @param label
   * @param z
   * @param t
   * @param c
   */
  public void setSliceLabel(String label, int z, int t, int c) {
	  int tmpIdx = this.m_actSliceIdx;
	  this.setActualSliceCoords(z, t, c);	  
	  this.setActualSliceLabel(label);
	  this.m_actSliceIdx = tmpIdx;
  }

  /**
   * Copy slice labels from src to this
   * @param src
   */
  public void adoptSliceLabels(MTBImage src) throws IllegalArgumentException {
	  
	  if (this.m_sizeStack != src.m_sizeStack) {
		  throw new IllegalArgumentException("Images must have the same number of slices.");
	  }
	  else {
		  int tmpIdx = this.m_actSliceIdx;
		  int tmpSrcIdx = src.m_actSliceIdx;
		  
		  for (int i = 0; i < this.m_sizeStack; i++) {
			  this.m_actSliceIdx = i;
			  src.m_actSliceIdx = i;
			  this.setActualSliceLabel(src.getActualSliceLabel());
		  }
		  this.m_actSliceIdx = tmpIdx;
		  src.m_actSliceIdx = tmpSrcIdx;
	  }
  }
  
  /**
   * Get an ImagePlus object. For ImageJ-types (MTB_BYTE, MTB_SHORT, MTB_FLOAT)
   * with underlying ImagePlus object, a reference to this ImagePlus is
   * returned. For Non-ImageJ-types (MTB_INT, MTB_DOUBLE) an ImagePlus object is
   * created for displaying but with definite LOSS OF INFORMATION. The ImagePlus
   * object's properties contain a reference to the MTBImage under the key
   * "MTBImage", which can be retrieved by the following code:
   * 
   * {@code (MTBImage)(ImagePlus.getProperty("MTBImage")) };
   * 
   * @return ImagePlus object
   */
  abstract public ImagePlus getImagePlus();

  /**
   * Should be used to create or update ImagePlus data if the MTBImage is not
   * using an ImagePlus to store the data (MTB_INT, MTB_DOUBLE, MTB_RGB). For
   * the other types, changes are made to the underlying ImagePlus directly,
   * thus this method should do nothing.
   */
  abstract protected void updateImagePlus();

  /**
   * Get the value of the 5D image at coordinate (x,y,z,t,c) as an Integer
   * 
   * @param x
   *          x-coordinate ranging from 0 to (sizeX - 1)
   * @param y
   *          y-coordinate ranging from 0 to (sizeY - 1)
   * @param z
   *          z-coordinate ranging from 0 to (sizeZ - 1)
   * @param t
   *          t-coordinate ranging from 0 to (sizeT - 1)
   * @param c
   *          c-coordinate ranging from 0 to (sizeC - 1)
   * @return voxel value
   */
  abstract public int getValueInt(int x, int y, int z, int t, int c);

  /**
   * Get the value of the 5D image at coordinate (x,y,z,t,c) as a Double
   * 
   * @param x
   *          x-coordinate ranging from 0 to (sizeX - 1)
   * @param y
   *          y-coordinate ranging from 0 to (sizeY - 1)
   * @param z
   *          z-coordinate ranging from 0 to (sizeZ - 1)
   * @param t
   *          t-coordinate ranging from 0 to (sizeT - 1)
   * @param c
   *          c-coordinate ranging from 0 to (sizeC - 1)
   * @return voxel value
   */
  abstract public double getValueDouble(int x, int y, int z, int t, int c);

  /**
   * Set the value of the 5D image at coordinate (x,y,z,t,c) using an Integer
   * 
   * @param x
   *          x-coordinate ranging from 0 to (sizeX - 1)
   * @param y
   *          y-coordinate ranging from 0 to (sizeY - 1)
   * @param z
   *          z-coordinate ranging from 0 to (sizeZ - 1)
   * @param t
   *          t-coordinate ranging from 0 to (sizeT - 1)
   * @param c
   *          c-coordinate ranging from 0 to (sizeC - 1)
   * @param value
   *          to set the voxel to
   */
  abstract public void putValueInt(int x, int y, int z, int t, int c, int value);

  /**
   * Set the value of the 5D image at coordinate (x,y,z,t,c) using a Double
   * 
   * @param x
   *          x-coordinate ranging from 0 to (sizeX - 1)
   * @param y
   *          y-coordinate ranging from 0 to (sizeY - 1)
   * @param z
   *          z-coordinate ranging from 0 to (sizeZ - 1)
   * @param t
   *          t-coordinate ranging from 0 to (sizeT - 1)
   * @param c
   *          c-coordinate ranging from 0 to (sizeC - 1)
   * @param value
   *          to set the voxel to
   */
  abstract public void putValueDouble(int x, int y, int z, int t, int c,
      double value);

  /**
   * Get the value of the actual z-stack coordinate (x,y,z) as an Integer
   * 
   * @param x
   *          x-coordinate ranging from 0 to (sizeX - 1)
   * @param y
   *          y-coordinate ranging from 0 to (sizeY - 1)
   * @param z
   *          z-coordinate ranging from 0 to (sizeZ - 1)
   * @return voxel value
   */
  abstract public int getValueInt(int x, int y, int z);

  /**
   * Get the value of the actual slice at coordinate (x,y,z) as an Double
   * 
   * @param x
   *          x-coordinate ranging from 0 to (sizeX - 1)
   * @param y
   *          y-coordinate ranging from 0 to (sizeY - 1)
   * @param z
   *          z-coordinate ranging from 0 to (sizeZ - 1)
   * @return voxel value
   */
  abstract public double getValueDouble(int x, int y, int z);

  /**
   * Set the value of the actual slice at coordinate (x,y,z) using an Integer
   * 
   * @param x
   *          x-coordinate ranging from 0 to (sizeX - 1)
   * @param y
   *          y-coordinate ranging from 0 to (sizeY - 1)
   * @param z
   *          z-coordinate ranging from 0 to (sizeZ - 1)
   * @param value
   *          to set the voxel to
   */
  abstract public void putValueInt(int x, int y, int z, int value);

  /**
   * Set the value of the actual slice at coordinate (x,y,z) using a Double
   * 
   * @param x
   *          x-coordinate ranging from 0 to (sizeX - 1)
   * @param y
   *          y-coordinate ranging from 0 to (sizeY - 1)
   * @param z
   *          z-coordinate ranging from 0 to (sizeZ - 1)
   * @param value
   *          to set the voxel to
   */
  abstract public void putValueDouble(int x, int y, int z, double value);

  /**
   * Get the value of the actual slice at coordinate (x,y) as an Integer
   * 
   * @param x
   *          x-coordinate ranging from 0 to (sizeX - 1)
   * @param y
   *          y-coordinate ranging from 0 to (sizeY - 1)
   * @return voxel value
   */
  abstract public int getValueInt(int x, int y);

  /**
   * Get the value of the actual slice at coordinate (x,y) as an Double
   * 
   * @param x
   *          x-coordinate ranging from 0 to (sizeX - 1)
   * @param y
   *          y-coordinate ranging from 0 to (sizeY - 1)
   * @return voxel value
   */
  abstract public double getValueDouble(int x, int y);

  public void fillWhite() {
    for (int c = 0; c < this.getSizeC(); ++c)
      for (int t = 0; t < this.getSizeT(); ++t)
        for (int z = 0; z < this.getSizeZ(); ++z)
          for (int y = 0; y < this.getSizeY(); ++y)
            for (int x = 0; x < this.getSizeX(); ++x)
              this.putValueDouble(x, y, z, t, c, this.getTypeMax());
  }

  public void fillBlack() {
    for (int c = 0; c < this.getSizeC(); ++c)
      for (int t = 0; t < this.getSizeT(); ++t)
        for (int z = 0; z < this.getSizeZ(); ++z)
          for (int y = 0; y < this.getSizeY(); ++y)
            for (int x = 0; x < this.getSizeX(); ++x)
              this.putValueDouble(x, y, z, t, c, this.getTypeMin());
  }

  /**
   * Set the value of the actual slice at coordinate (x,y) using an Integer
   * 
   * @param x
   *          x-coordinate ranging from 0 to (sizeX - 1)
   * @param y
   *          y-coordinate ranging from 0 to (sizeY - 1)
   * @param value
   *          to set the voxel to
   */
  abstract public void putValueInt(int x, int y, int value);

  /**
   * Set the value of the actual slice at coordinate (x,y) using a Double
   * 
   * @param x
   *          x-coordinate ranging from 0 to (sizeX - 1)
   * @param y
   *          y-coordinate ranging from 0 to (sizeY - 1)
   * @param value
   *          to set the voxel to
   */
  abstract public void putValueDouble(int x, int y, double value);

  /**
   * Get minimum and maximum value of the image as int
   * 
   * @return min at int[0], max at int[1]
   */
  abstract public int[] getMinMaxInt();

  /**
   * Get minimum and maximum value of the image as double
   * 
   * @return min at double[0], max at double[1]
   */
  abstract public double[] getMinMaxDouble();

  /**
   * Duplicates the object. That means also a copy of the underlying ImagePlus
   * (or other data structure) is created. The method uses the internal MTBImageFactory, which extends MTBOperator, 
   * thus the duplicated image will keep its history. The reference to the calling operator is set to null, use duplicate(MTBOperator)
   * if you need to specify the calling operator.
   * 
   * @return MTBImage of the same type
   */
  public MTBImage duplicate() {
	  return this.duplicate(null);
  }
  
  /**
   * Duplicates the object. That means also a copy of the underlying ImagePlus
   * (or other data structure) is created. The method uses the internal MTBImageFactory, which extends MTBOperator, 
   * thus the duplicated image will keep its history.
   * 
   * @param callingOperator the MTBOperator, which calls this method
   * @return MTBImage of the same type
   */
  public MTBImage duplicate(MTBOperator callingOperator) {
	  
	  MTBImageFactory factory = null;
	  MTBImage newImg = null;
	  
	  try {
		  
		factory = new MTBImageFactory(this);
		
		factory.runOp(callingOperator);
		
		newImg = factory.getResultImg();
		
	  } catch (MTBOperatorException e) {
		e.printStackTrace();
	  } catch (MTBProcessingDAGException e) {
		e.printStackTrace();
	  }
	  
	  return newImg; 
  }
  
  /**
   * For a given String, return the corresponding MTBImageType. If the String doesn't match any type, null is returned
   * @param s String must be one of "MTB_BYTE", "MTB_SHORT", "MTB_INT", "MTB_FLOAT", "MTB_DOUBLE", "MTB_RGB"
   * @return
   */
  public static MTBImageType stringToType(String s) {
	  
	  MTBImageType type = null;
	  
	  if (s.equals("MTB_BYTE"))
		  type = MTBImageType.MTB_BYTE;
	  else if (s.equals("MTB_SHORT"))
		  type = MTBImageType.MTB_SHORT;
	  else if (s.equals("MTB_INT"))
		  type = MTBImageType.MTB_INT;			  
	  else if (s.equals("MTB_FLOAT"))
		  type = MTBImageType.MTB_FLOAT;
	  else if (s.equals("MTB_DOUBLE"))
		  type = MTBImageType.MTB_DOUBLE;
	  else if (s.equals("MTB_RGB"))
		  type = MTBImageType.MTB_RGB;
	  
	  return type;
  }
  
  /**
   * Return the corresponding MTBImageType for its index in the enum. If this index doesn't exist, null is returned
   * @param s String must be one of "MTB_BYTE", "MTB_SHORT", "MTB_INT", "MTB_FLOAT", "MTB_DOUBLE", "MTB_RGB"
   * @return
   */
  public static MTBImageType ordinalToType(int enumIdx) {
	  
	  MTBImageType type = null;
	  
	  if (enumIdx == 0)
		  type = MTBImageType.MTB_BYTE;
	  else if (enumIdx == 1)
		  type = MTBImageType.MTB_SHORT;
	  else if (enumIdx == 2)
		  type = MTBImageType.MTB_INT;			  
	  else if (enumIdx == 3)
		  type = MTBImageType.MTB_FLOAT;
	  else if (enumIdx == 4)
		  type = MTBImageType.MTB_DOUBLE;
	  else if (enumIdx == 5)
		  type = MTBImageType.MTB_RGB;
	  
	  return type;
  }
  
	@Override
	public void keyPressed(KeyEvent e) {
		char cmd = e.getKeyChar();

	//	Point loc = canvas.getCursorLoc();
	//	int x = canvas.screenX(loc.x);
	//	int y = canvas.screenY(loc.y);
		if (this.m_img != null) {
			if (this.m_img.getWindow() != null) {
			
				WindowManager.setCurrentWindow(this.m_img.getWindow());
				
				if (cmd == '+' || cmd == ']') {
		 			//canvas.zoomIn(x, y);
					//if (canvas.getMagnification()<=1.0) img.getImagePlus().repaintWindow();
					Zoom z = new Zoom();
					z.run("in");
				}
				else if (cmd == '-' || cmd == '/') {
					//canvas.zoomOut(x, y);
					//if (canvas.getMagnification()<1.0) img.getImagePlus().repaintWindow();
					Zoom z = new Zoom();
					z.run("out");
				}
			}
		}
		
	}


	@Override
	public void keyReleased(KeyEvent e) {
		// TODO Auto-generated method stub
		
	}


	@Override
	public void keyTyped(KeyEvent e) {
		// TODO Auto-generated method stub
		
	}
  
  
  /**
   * Factory method to create MTBImage objects. Checks the image type and
   * returns the appropriate subclass object (MTBImageByte, MTBImageShort, ...)
   * If the ImagePlus already contains a MTBImage in its properties, this
   * MTBImage is returned
   * 
   * @param img
   *          ImagePlus object to wrap
   * @return MTBImage subclass object
   * @throws IllegalArgumentException
   *           for unsupported types
   */
  public static MTBImage createMTBImage(ImagePlus img)
      throws IllegalArgumentException {

    if (img.getProperties() != null
        && img.getProperties().containsKey("MTBImage")) {

      return (MTBImage) img.getProperty("MTBImage");
    } else {

      if (img.getType() == ImagePlus.GRAY8) {
        return new MTBImageByte(img);
      } else if (img.getType() == ImagePlus.GRAY16) {
        return new MTBImageShort(img);
      } else if (img.getType() == ImagePlus.GRAY32) {
        return new MTBImageFloat(img);
      } else if (img.getType() == ImagePlus.COLOR_RGB) {
        return new MTBImageRGB(img);
      } else {
        throw new IllegalArgumentException(
            "MTBImage.createMTBImage(..): Image type " + img.getType()
                + " not supported.");
      }
    }
  }

  /**
   * Factory method to create MTBImage objects with value 0 at each coordinate.
   * 
   * @param sizeX
   *          size in x-dimension
   * @param sizeY
   *          size in y-dimension
   * @param sizeZ
   *          size in z-dimension
   * @param sizeT
   *          size in t-dimension
   * @param sizeC
   *          size in c-dimension
   * @param type
   *          image type (one of MTB_BYTE, MTB_SHORT, MTB_INT, MTB_FLOAT,
   *          MTB_DOUBLE)
   * @return MTBImage subclass object
   */
  public static MTBImage createMTBImage(int sizeX, int sizeY, int sizeZ,
      int sizeT, int sizeC, MTBImageType type) throws IllegalArgumentException {

    if (type == MTBImageType.MTB_BYTE) {
      return new MTBImageByte(sizeX, sizeY, sizeZ, sizeT, sizeC);
    } else if (type == MTBImageType.MTB_SHORT) {
      return new MTBImageShort(sizeX, sizeY, sizeZ, sizeT, sizeC);
    } else if (type == MTBImageType.MTB_INT) {
      return new MTBImageInt(sizeX, sizeY, sizeZ, sizeT, sizeC);
    } else if (type == MTBImageType.MTB_FLOAT) {
      return new MTBImageFloat(sizeX, sizeY, sizeZ, sizeT, sizeC);
    } else if (type == MTBImageType.MTB_DOUBLE) {
      return new MTBImageDouble(sizeX, sizeY, sizeZ, sizeT, sizeC);
    } else if (type == MTBImageType.MTB_RGB) {
      return new MTBImageRGB(sizeX, sizeY, sizeZ, sizeT, sizeC);
    } else {
      throw new IllegalArgumentException(
          "MTBImage.createMTBImage(.., int type): Unknown type: " + type);
    }
  }

  protected static MTBImage createMTBImageByte(int sizeX, int sizeY, int sizeZ,
      int sizeT, int sizeC, byte[][] data) throws IllegalArgumentException {
    MTBImage byteImg = null;
    try {
      byteImg = new MTBImageByte(data, sizeX, sizeY, sizeZ, sizeT, sizeC);
    } catch (IllegalArgumentException e) {
      throw e;
    }
    return byteImg;
  }
  
  
  
  /**
   * A class for creating MTBImages which implements the MTBOperator.
   * @author gress
   *
   */
  public class MTBImageFactory extends MTBOperator {
  	
  
	protected void collectArguments() throws MTBOperatorException {
        super.collectArguments();

		MTBOpArgumentDescriptor[] inputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("inImg", MTBImage.class, 
						"Input image", true, null)
		};
	
		MTBOpArgumentDescriptor[] outputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("resultImg", MTBImage.class, 
						"Result image", true, null)
		};
	
		MTBOpArgumentDescriptor[] parameterDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("factoryMethod", FactoryMethod.class,
						"Type of image creation method", true, null),
						
				new MTBOpArgumentDescriptor("targetImageType", MTBImageType.class,
						"Image type of new image if factory method is CONVERT", false, false),
				new MTBOpArgumentDescriptor("scaleDown", Boolean.class,
						"Scale image values if factory method is CONVERT", false, false),		
						
				new MTBOpArgumentDescriptor("x", Integer.class,
						"Starting coordinate in x-dimension for copying an image part if factory method is IMAGE_PART", false, 0.0),
				new MTBOpArgumentDescriptor("y", Integer.class,
						"Starting coordinate in y-dimension for copying an image part if factory method is IMAGE_PART", false, 0.0),
				new MTBOpArgumentDescriptor("z", Integer.class,
						"Starting coordinate in z-dimension for copying an image part if factory method is IMAGE_PART" +
						" or slice coordinate if factory method is SLICE", false, 0.0),
				new MTBOpArgumentDescriptor("t", Integer.class,
						"Starting coordinate in t-dimension for copying an image part if factory method is IMAGE_PART" +
						" or slice coordinate if factory method is SLICE", false, 0.0),
				new MTBOpArgumentDescriptor("c", Integer.class,
						"Starting coordinate in c-dimension for copying an image part if factory method is IMAGE_PART" +
						" or slice coordinate if factory method is SLICE", false, 0.0),
						
				new MTBOpArgumentDescriptor("sizeX", Integer.class,
						"Size in x-dimension for copying an image part if factory method is IMAGE_PART", false, 0.0),
				new MTBOpArgumentDescriptor("sizeY", Integer.class,
						"Size in y-dimension for copying an image part if factory method is IMAGE_PART", false, 0.0),
				new MTBOpArgumentDescriptor("sizeZ", Integer.class,
						"Size in z-dimension for copying an image part if factory method is IMAGE_PART", false, 0.0),
				new MTBOpArgumentDescriptor("sizeT", Integer.class,
						"Size in t-dimension for copying an image part if factory method is IMAGE_PART", false, 0.0),
				new MTBOpArgumentDescriptor("sizeC", Integer.class,
						"Size in c-dimension for copying an image part if factory method is IMAGE_PART", false, 0.0),
		};
		  
		MTBOpArgumentDescriptor[] supplementalDescriptors = null;
	
			addDescriptors( parameterDescriptors, inputDescriptors, 
					outputDescriptors, supplementalDescriptors);
	}
	
	  
	public MTBImageFactory()  throws MTBOperatorException {
		
	}
	
	/**
	 * Duplicate constructor
	 * @param inImg
	 * @throws MTBOperatorException 
	 */
  	public MTBImageFactory(MTBImage inImg) throws MTBOperatorException {
  		this.setInImg(inImg);
  		this.setFactoryMethod(FactoryMethod.DUPLICATE);
  	}
  	

	/**
  	 * Convert constructor
  	 * @param inImg
  	 * @param type
  	 * @param scaleDown
  	 * @throws MTBOperatorException
  	 */
  	public MTBImageFactory(MTBImage inImg, MTBImageType type, boolean scaleDown) throws MTBOperatorException {
  		this.setInImg(inImg);
  		this.setFactoryMethod(FactoryMethod.CONVERT);
  		this.setTargetImageType(type);
  		this.setScaleDown(scaleDown);
  	}
  	
  	/**
  	 * Get slice constructor
  	 * @param inImg
  	 * @param z
  	 * @param t
  	 * @param c
  	 * @throws MTBOperatorException
  	 */
  	public MTBImageFactory(MTBImage inImg, int z, int t, int c) throws MTBOperatorException {
  		this.setInImg(inImg);
  		this.setFactoryMethod(FactoryMethod.SLICE);
  		this.setZ(z);
  		this.setT(t);
  		this.setC(c);
  	}
  	
  	/**
  	 * Create a new image from the specified image part
  	 * @param inImg
  	 * @param x
  	 * @param y
  	 * @param z
  	 * @param t
  	 * @param c
  	 * @param sizeX
  	 * @param sizeY
  	 * @param sizeZ
  	 * @param sizeT
  	 * @param sizeC
  	 * @throws MTBOperatorException
  	 */
  	public MTBImageFactory(MTBImage inImg, int x, int y, int z, int t, int c,
  									int sizeX, int sizeY, int sizeZ, int sizeT, int sizeC) throws MTBOperatorException {
  		this.setInImg(inImg);
  		this.setFactoryMethod(FactoryMethod.IMAGE_PART);
  		this.setX(x);
  		this.setY(y);
  		this.setZ(z);
  		this.setT(t);
  		this.setC(c);
  		this.setSizeX(sizeX);
  		this.setSizeY(sizeY);
  		this.setSizeZ(sizeZ);
  		this.setSizeT(sizeT);
  		this.setSizeC(sizeC);
  	}
  	
//  	public boolean validateCustom() {
//  		boolean valid = true;
//  		
//  		FactoryMethod fm = null;
//  		
//		try {
//			fm = this.getFactoryMethod();
//		} catch (MTBOperatorException e) {
//			e.printStackTrace();
//			return false;
//		}
//  		
//  		if (fm == FactoryMethod.CONVERT) {
//  		
//  			try {
//				valid &= (this.getImageType() != null);
//			} catch (MTBOperatorException e) {
//				e.printStackTrace();
//				valid = false;
//			}
//  		}
//  		else if (fm == FactoryMethod.SLICE) {
//  			try {
//  				MTBImage img = this.getInImg();
//  				
//  				valid = valid && (this.getParameter("z") != null) && this.getZ() >= 0 && this.getZ() < img.getSizeZ();
//  				valid = valid && (this.getParameter("t") != null) && this.getT() >= 0 && this.getT() < img.getSizeT();
//  				valid = valid && (this.getParameter("c") != null) && this.getC() >= 0 && this.getC() < img.getSizeC();	
//  				
//  				
//  			} catch (MTBOperatorException e) {
//				e.printStackTrace();
//				valid = false;
//			}
//  		}
//  		else if (fm == FactoryMethod.IMAGE_PART) {
//  			try {
//  				
//  				// TODO: check size of copied part !!!!
//  				
//  				valid = valid && (this.getParameter("x") != null);
//  				valid = valid && (this.getParameter("y") != null);
//  				valid = valid && (this.getParameter("z") != null);
//  				valid = valid && (this.getParameter("t") != null);
//  				valid = valid && (this.getParameter("c") != null);	
//  				valid = valid && (this.getParameter("sizeX") != null);
//  				valid = valid && (this.getParameter("sizeY") != null);
//  				valid = valid && (this.getParameter("sizeZ") != null);
//  				valid = valid && (this.getParameter("sizeT") != null);
//  				valid = valid && (this.getParameter("sizeC") != null);
//		
//  				
//  			} catch (MTBOperatorException e) {
//				e.printStackTrace();
//				valid = false;
//			} 			
//  				
//  		}
//  		
//  		
//  		return valid;
//  	}
  	
  	@Override
  	protected void operate() throws MTBOperatorException {
  		
  		FactoryMethod fm = this.getFactoryMethod();
  		
		MTBImage resultImg = null;
  		
  		if (fm == FactoryMethod.DUPLICATE) {

  			// duplicate image
  			resultImg = this.duplicate(this.getInImg());
  			
  			if (resultImg != null) {
  				try {
  					resultImg.adoptSliceLabels(this.getInImg());
				} catch (IllegalArgumentException e) {
  					e.printStackTrace();
  					throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, "MTBImageFactory.operate(): DUPLICATE failed: " + e.getMessage());
  				}
				
  				this.setResultImg(resultImg);
  			}
  			else 
  				throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, "MTBImageFactory.operate(): DUPLICATE failed.");
  		}
  		else if (fm == FactoryMethod.CONVERT) {

  			// convert image to another type
  			resultImg = this.convertType(this.getInImg(), this.getTargetImageType(), this.getScaleDown());
  			
  			if (resultImg != null) {
  				try {
  					resultImg.adoptSliceLabels(this.getInImg());
  				} catch (IllegalArgumentException e) {
  					e.printStackTrace();
  					throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, "MTBImageFactory.operate(): CONVERT failed: " + e.getMessage());
  				}
  				
  				this.setResultImg(resultImg);
  			}
  			else 
  				throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, "MTBImageFactory.operate(): CONVERT failed.");
  		}
  		else if (fm == FactoryMethod.SLICE) {
  			
  			// get the copy of a slice of the image (in fact runs getImagePart)
  			resultImg = this.getSlice(this.getInImg(), this.getZ(), this.getT(), this.getC());
  				
  			if (resultImg != null) {
  				resultImg.setSliceLabel(this.getInImg().getSliceLabel(this.getZ(), this.getT(), this.getC()), 0, 0, 0);
  				this.setResultImg(resultImg);
  			}
  			else 
  				throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, "MTBImageFactory.operate(): SLICE copying failed.");
  		}
  		else if (fm == FactoryMethod.IMAGE_PART) {
  			
  			// get the copy of a specified image part
  			resultImg = this.getImagePart(this.getInImg(), this.getX(), this.getY(), this.getZ(), this.getT(), this.getC(),
  											this.getSizeX(), this.getSizeY(), this.getSizeZ(), this.getSizeT(), this.getSizeC());
				
  			if (resultImg != null) {
  				MTBImage inImg = this.getInImg();
  				int z0 = this.getZ(); int t0 = this.getT(); int c0 = this.getC();
  				int sz = this.getSizeZ(); int st = this.getSizeT(); int sc = this.getSizeC();
  				
  				for (int t = 0; t < st; t++) {
  					for (int z = 0; z < sz; z++) {
  						for (int c = 0; c < sc; c++) {
  							resultImg.setSliceLabel(inImg.getSliceLabel(z0 + z, t0 + t, c0 + c), z, t, c);
  						}
  					}
  				}
			
  				this.setResultImg(resultImg);
  			}
  			else 
  				throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, "MTBImageFactory.operate(): IMAGE_PART copying failed.");
  		}
  		else {
  			throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, "MTBImageFactory.operate() failed: Unknown creation method " + fm.toString());
  		}

  	}
  	
  	public MTBImage duplicate(MTBImage img) {
  		
  		MTBImage newImg = null;
  		
  		if (img.getType() == MTBImageType.MTB_INT
  				|| img.getType() == MTBImageType.MTB_DOUBLE
  				|| img.getType() == MTBImageType.MTB_RGB) {
  			
  			// duplicate MTBImages which are not based on ImageJ's ImagePlus
  			newImg = this.duplicateNonImageJType(img);
  			
  		}
  		else {
  			
  		// duplicate MTBImages which are based on ImagePlus
  			newImg = this.duplicateImageJType(img);
  		}
  		
  		return newImg;
  	}
  	
  	
  	private MTBImage duplicateImageJType(MTBImage img) {
  		
		// create new image stack
		ImageStack newStack = new ImageStack(img.m_sizeX, img.m_sizeY);
		
		// duplicate slices
		for (int i = 1; i <= img.m_sizeStack; i++) {
			newStack.addSlice(img.m_imgStack.getSliceLabel(i), img.m_imgStack.getProcessor(i).duplicate());
		}
		
		// create new ImagePlus
		ImagePlus newImP = img.m_img.createImagePlus();
	
		newImP.setDimensions(img.m_img.getNChannels(), img.m_img.getNSlices(), img.m_img.getNFrames());
		newImP.setStack(img.m_img.getTitle(), newStack);
		newImP.setOpenAsHyperStack(img.m_img.isHyperStack() || img.m_img.getOpenAsHyperStack());
		
		newImP.setPosition(img.m_img.getCurrentSlice());
		newImP.setDisplayRange(img.m_img.getDisplayRangeMin(), img.m_img.getDisplayRangeMax());
		
		newImP.setTitle(img.m_title);
		
		MTBImage newImg = MTBImage.createMTBImage(newImP);
		
		// copy actual slice index
		newImg.m_actSliceIdx = img.m_actSliceIdx;
		
		// set stepsizes
		newImg.setStepsizeX(img.m_stepX);
		newImg.setStepsizeY(img.m_stepY);
		newImg.setStepsizeZ(img.m_stepZ);
		newImg.setStepsizeT(img.m_stepT);
		
		// set units
		newImg.setUnitX(img.m_unitX);
		newImg.setUnitY(img.m_unitY);
		newImg.setUnitZ(img.m_unitZ);
		newImg.setUnitT(img.m_unitT);	
		
		// return new image wrapper
		return newImg;
  	}
  	
  	private MTBImage duplicateNonImageJType(MTBImage img) {
  		
  		MTBImage newImg = null;
  		
  		if (img.getType() == MTBImageType.MTB_DOUBLE) {
			newImg = new MTBImageDouble(img.m_sizeX, img.m_sizeY, img.m_sizeZ, img.m_sizeT, img.m_sizeC);	
			
			for (int i = 0; i < m_sizeStack; i++) {
				System.arraycopy(((MTBImageDouble)img).m_data[i], 0, ((MTBImageDouble)newImg).m_data[i], 0, img.m_sizeX*img.m_sizeY);
			}
  		}  		
  		else if (img.getType() == MTBImageType.MTB_INT) {
			newImg = new MTBImageInt(img.m_sizeX, img.m_sizeY, img.m_sizeZ, img.m_sizeT, img.m_sizeC);	
			
			for (int i = 0; i < m_sizeStack; i++) {
				System.arraycopy(((MTBImageInt)img).m_data[i], 0, ((MTBImageInt)newImg).m_data[i], 0, img.m_sizeX*img.m_sizeY);
			}
  		}
  		else if (img.getType() == MTBImageType.MTB_RGB) {
			newImg = new MTBImageRGB(img.m_sizeX, img.m_sizeY, img.m_sizeZ, img.m_sizeT, img.m_sizeC);	
			
			for (int i = 0; i < m_sizeStack; i++) {
				System.arraycopy(((MTBImageRGB)img).m_dataR[i], 0, ((MTBImageRGB)newImg).m_dataR[i], 0, img.m_sizeX*img.m_sizeY);
				System.arraycopy(((MTBImageRGB)img).m_dataG[i], 0, ((MTBImageRGB)newImg).m_dataG[i], 0, img.m_sizeX*img.m_sizeY);
				System.arraycopy(((MTBImageRGB)img).m_dataB[i], 0, ((MTBImageRGB)newImg).m_dataB[i], 0, img.m_sizeX*img.m_sizeY);
			}
  		}
	
		// set title
		newImg.m_title = img.m_title;
		
		// copy actual slice index
		newImg.m_actSliceIdx = img.m_actSliceIdx;
		
		// set stepsizes
		newImg.setStepsizeX(img.m_stepX);
		newImg.setStepsizeY(img.m_stepY);
		newImg.setStepsizeZ(img.m_stepZ);
		newImg.setStepsizeT(img.m_stepT);
		
		// set units
		newImg.setUnitX(img.m_unitX);
		newImg.setUnitY(img.m_unitY);
		newImg.setUnitZ(img.m_unitZ);
		newImg.setUnitT(img.m_unitT);	
		
		// return new image wrapper
		return newImg;
	}
  	
  	
    /**
     * Create an image of given type from this image's values. Information is lost
     * if a type of less precision is used!
     * 
     * @param type
     *          new image's type (see static final fields for types)
     * @param scaleDown
     *          If true, the data is scaled down to fit in the range of the new
     *          image type values, if the new image type is of less precision. If
     *          false, the values are simply casted.
     * @return new image of given type
     */
    protected MTBImage convertType(MTBImage img, MTBImageType type, boolean scaleDown) {

      // create new image of given type
      MTBImage newImg = MTBImage.createMTBImage(img.m_sizeX, img.m_sizeY, img.m_sizeZ,
    		  img.m_sizeT, img.m_sizeC, type);

      // set title
      newImg.setTitle(img.m_title);

      // set stepsizes
      newImg.setStepsizeX(img.m_stepX);
      newImg.setStepsizeY(img.m_stepY);
      newImg.setStepsizeZ(img.m_stepZ);
      newImg.setStepsizeT(img.m_stepT);

      // set units
      newImg.setUnitX(img.m_unitX);
      newImg.setUnitY(img.m_unitY);
      newImg.setUnitZ(img.m_unitZ);
      newImg.setUnitT(img.m_unitT);

      // get actual slice index
      int idx = img.getActualSliceIndex();

      if (img.m_type == MTBImageType.MTB_RGB && type != MTBImageType.MTB_RGB) {
        // Conversion from RGB to gray value type

        int[] rgbValue;

        for (int i = 0; i < img.m_sizeStack; i++) {

          img.setActualSliceIndex(i);
          newImg.setActualSliceIndex(i);

          for (int y = 0; y < img.m_sizeY; y++) {
            for (int x = 0; x < img.m_sizeX; x++) {
              rgbValue = ((MTBImageRGB) img).getValue(x, y);

              newImg.putValueDouble(x, y, 0.299 * (double) rgbValue[0] + 0.587
                  * (double) rgbValue[1] + 0.114 * (double) rgbValue[2]);
            }
          }
        }
      } 
      else if (img.m_type != MTBImageType.MTB_RGB && type == MTBImageType.MTB_RGB) {
        // Conversion from gray value type to RGB

        double val;

        double[] minmax = null;
        if (scaleDown) {
          minmax = img.getMinMaxDouble();
        }

        for (int i = 0; i < img.m_sizeStack; i++) {

          img.setActualSliceIndex(i);
          newImg.setActualSliceIndex(i);

          for (int y = 0; y < img.m_sizeY; y++) {
            for (int x = 0; x < img.m_sizeX; x++) {
              val = img.getValueDouble(x, y);

              if (scaleDown) {
                val = Math.round((val - minmax[0]) / (minmax[1] - minmax[0])
                    * 255.0);
              }

              ((MTBImageRGB) newImg).putValue(x, y, (int) val, (int) val,
                  (int) val);
            }
          }
        }
      } 
      else {
        // Conversion of gray value types or conversion from RGB to RGB

        double val;

        double[] minmax = null;

        if (type.ordinal() < img.m_type.ordinal() && scaleDown) {
          minmax = img.getMinMaxDouble();
        }

        for (int i = 0; i < img.m_sizeStack; i++) {

          img.setActualSliceIndex(i);
          newImg.setActualSliceIndex(i);

          for (int y = 0; y < img.m_sizeY; y++) {
            for (int x = 0; x < img.m_sizeX; x++) {
              val = img.getValueDouble(x, y);

              if (type.ordinal() < img.m_type.ordinal() && scaleDown) {
                val = Math.round((val - minmax[0]) / (minmax[1] - minmax[0])
                    * (newImg.getTypeMax() - newImg.getTypeMin())
                    + newImg.getTypeMin());
              }

              newImg.putValueDouble(x, y, val);
            }
          }
        }
      }

      // restore actual slice index
      img.setActualSliceIndex(idx);
      newImg.setActualSliceIndex(0);

      return newImg;
    }
  		
  	
    /**
     * Get a copy of a part of this image as new MTBImage
     * 
     * @param x
     *          x-coordinate where the first value is copied from
     * @param y
     *          y-coordinate where the first value is copied from
     * @param z
     *          z-coordinate where the first value is copied from
     * @param t
     *          t-coordinate where the first value is copied from
     * @param c
     *          c-coordinate where the first value is copied from
     * @param sizeX
     *          size of the copied part in x-dimension
     * @param sizeY
     *          size of the copied part in y-dimension
     * @param sizeZ
     *          size of the copied part in z-dimension
     * @param sizeT
     *          size of the copied part in t-dimension
     * @param sizeC
     *          size of the copied part in c-dimension
     * @return new MTBImage with values equal to the specified part
     * @throws IllegalArgumentException
     *           if image boundaries are exceeded in any way
     */
    protected MTBImage getImagePart(MTBImage img, 
    								int x, int y, int z, int t, int c, int sizeX,
    								int sizeY, int sizeZ, int sizeT, int sizeC)
        							throws IllegalArgumentException {

      if (x + sizeX > img.m_sizeX || y + sizeY > img.m_sizeY
          || z + sizeZ > img.m_sizeZ || t + sizeT > img.m_sizeT
          || c + sizeC > img.m_sizeC) {
        throw new IllegalArgumentException(
            "MTBImage.getImagePart(..): Specified image part exceeds the image boundaries.");
      } else if (x < 0 || x >= img.m_sizeX || y < 0 || y >= img.m_sizeY
          || z < 0 || z >= img.m_sizeZ || t < 0 || t >= img.m_sizeT || c < 0
          || c >= img.m_sizeC) {
        throw new IllegalArgumentException(
            "MTBImage.getImagePart(..): Specified coordinate exceeds the image boundaries.");
      } else {

        MTBImage imgPart = MTBImage.createMTBImage(sizeX, sizeY, sizeZ, sizeT,
            sizeC, img.m_type);

        for (int cc = 0; cc < sizeC; cc++) {
          for (int tt = 0; tt < sizeT; tt++) {
            for (int zz = 0; zz < sizeZ; zz++) {
              for (int yy = 0; yy < sizeY; yy++) {
                for (int xx = 0; xx < sizeX; xx++) {
                  imgPart.putValueDouble(xx, yy, zz, tt, cc, img.getValueDouble(
                      x + xx, y + yy, z + zz, t + tt, c + cc));
                }
              }
            }
          }
        }

        imgPart.m_title = img.m_title;

        // set stepsizes
        imgPart.setStepsizeX(img.m_stepX);
        imgPart.setStepsizeY(img.m_stepY);
        imgPart.setStepsizeZ(img.m_stepZ);
        imgPart.setStepsizeT(img.m_stepT);

        // set units
        imgPart.setUnitX(img.m_unitX);
        imgPart.setUnitY(img.m_unitY);
        imgPart.setUnitZ(img.m_unitZ);
        imgPart.setUnitT(img.m_unitT);

        return imgPart;
      }
    }
    
    /**
     * Get a copy of the specified slice
     * 
     * @param z
     * @param t
     * @param c
     * @return
     * @throws IllegalArgumentException
     */
    protected MTBImage getSlice(MTBImage img, int z, int t, int c) throws IllegalArgumentException {
        return this.getImagePart(img, 0, 0, z, t, c, img.m_sizeX, img.m_sizeY, 1, 1, 1);
    }
    
  	
  	protected MTBImage getInImg() throws MTBOperatorException {
  		return (MTBImage) this.getInput("inImg");
  	}
  	
  	protected void setInImg(MTBImage inImg) throws MTBOperatorException {
  		this.setInput("inImg", inImg);
  	}
  	
  	public MTBImage getResultImg() throws MTBOperatorException {
  		return (MTBImage)this.getOutput("resultImg");
  	}
  	
  	public void setResultImg(MTBImage resultImg) throws MTBOperatorException {
  		this.setOutput("resultImg", resultImg);
  	}  	
  	
  	protected FactoryMethod getFactoryMethod() throws MTBOperatorException {
  		return (FactoryMethod)this.getParameter("factoryMethod");
  	}
  	
  	protected void setFactoryMethod(FactoryMethod fm) throws MTBOperatorException {
  		this.setParameter("factoryMethod", fm);
  	}  	
  	
  	protected MTBImageType getTargetImageType() throws MTBOperatorException {
  		return (MTBImageType)this.getParameter("targetImageType");
  	}
  	
  	protected void setTargetImageType(MTBImageType type) throws MTBOperatorException {
  		this.setParameter("targetImageType", type);
  	}
  	
  	protected boolean getScaleDown() throws MTBOperatorException {
  		return (Boolean)this.getParameter("scaleDown");
  	}
  	
  	protected void setScaleDown(boolean scaleDown) throws MTBOperatorException {
  		this.setParameter("scaleDown", scaleDown);
  	}
  	
  	protected int getX() throws MTBOperatorException {
  		return (Integer)this.getParameter("x");
  	}
  	
  	protected void setX(int x) throws MTBOperatorException {
  		this.setParameter("x", x);
  	}
  	
  	protected int getY() throws MTBOperatorException {
  		return (Integer)this.getParameter("y");
  	}
  	
  	protected void setY(int y) throws MTBOperatorException {
  		this.setParameter("y", y);
  	}
  	
  	protected int getZ() throws MTBOperatorException {
  		return (Integer)this.getParameter("z");
  	}
  	
  	protected void setZ(int z) throws MTBOperatorException {
  		this.setParameter("z", z);
  	}
  	
  	protected int getT() throws MTBOperatorException {
  		return (Integer)this.getParameter("t");
  	}
  	
  	protected void setT(int t) throws MTBOperatorException {
  		this.setParameter("t", t);
  	}
  	
  	protected int getC() throws MTBOperatorException {
  		return (Integer)this.getParameter("c");
  	}
  	
  	protected void setC(int c) throws MTBOperatorException {
  		this.setParameter("c", c);
  	}
  	
  	protected int getSizeX() throws MTBOperatorException {
  		return (Integer)this.getParameter("sizeX");
  	}
  	
  	protected void setSizeX(int sizeX) throws MTBOperatorException {
  		this.setParameter("sizeX", sizeX);
  	}
  	
  	protected int getSizeY() throws MTBOperatorException {
  		return (Integer)this.getParameter("sizeY");
  	}
  	
  	protected void setSizeY(int sizeY) throws MTBOperatorException {
  		this.setParameter("sizeY", sizeY);
  	}
  	
  	protected int getSizeZ() throws MTBOperatorException {
  		return (Integer)this.getParameter("sizeZ");
  	}
  	
  	protected void setSizeZ(int sizeZ) throws MTBOperatorException {
  		this.setParameter("sizeZ", sizeZ);
  	}
  	
  	protected int getSizeT() throws MTBOperatorException {
  		return (Integer)this.getParameter("sizeT");
  	}
  	
  	protected void setSizeT(int sizeT) throws MTBOperatorException {
  		this.setParameter("sizeT", sizeT);
  	}
  	
  	protected int getSizeC() throws MTBOperatorException {
  		return (Integer)this.getParameter("sizeC");
  	}
  	
  	protected void setSizeC(int sizeC) throws MTBOperatorException {
  		this.setParameter("sizeC", sizeC);
  	}
  	
  }
  
  
  
}
