/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.drawing;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.util.Random;
import java.util.Vector;

import de.unihalle.informatik.MiToBo.datatypes.MTBPoint3D;
import de.unihalle.informatik.MiToBo.datatypes.MTBRegion3DSet;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.drawing.DrawRegion2DSet.DrawType;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.exceptions.MTBProcessingDAGException;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException.OperatorExceptionType;
import de.unihalle.informatik.MiToBo.operator.MTBOpArgumentDescriptor;
import de.unihalle.informatik.MiToBo.operator.MTBOperator;

/**
 * A class to visualize 3D regions. 
 * Background is always 0, regions can be drawn in different ways.
 * Four kinds of region images can be created or drawn to an existing image:
 * 
 * LABEL_IMAGE:  draw gray value image where regions are labeled by their index in the set + 1
 * ID_IMAGE:     draw gray value image where regions are labeled by their region id (see MTBRegion3D.id)
 * MASK_IMAGE:   draw a mask image with equal gray values for all regions
 * COLOR_IMAGE:  draw a mask with colored regions
 *
 * Constructors take the kind of image and the regions to be drawn as well as an eventual target image.
 * Default configurations are set by the constructors. Use the different get/set methods to specify
 * nondefault parameters like color, grayvalue etc.
 * 
 * Be sure to set the xMin, xMax, yMin, yMax, zMin, zMax values of the input MTBRegion3DSet, because these values are
 * used to determine the image size!!
 * 
 * @author Oliver Gress
 *
 */
public class DrawRegion3DSet extends MTBOperator {

	protected void collectArguments() throws MTBOperatorException {
        super.collectArguments();

		MTBOpArgumentDescriptor[] inputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("InputRegions", MTBRegion3DSet.class, 
						"Input regions", true, null),
				new MTBOpArgumentDescriptor("TargetImage", MTBImage.class, 
						"Target image", false, null),		
		};
	
		MTBOpArgumentDescriptor[] outputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("ResultImage", MTBImage.class, 
						"Result image", true, null)
		};
	
		MTBOpArgumentDescriptor[] parameterDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("DrawType", DrawType.class,
						"the type of image to be drawn", true, null),
				new MTBOpArgumentDescriptor("ImageType", MTBImageType.class,
						"the datatype of the output image", true, null),		
				new MTBOpArgumentDescriptor("GrayValue", Double.class,
						"the gray value used to paint all regions", false, null),		
				new MTBOpArgumentDescriptor("Color", Color.class,
						"the color used to paint all regions", false, null),
				new MTBOpArgumentDescriptor("Random", Random.class,
						"a random generator generate a random color for each region", false, null)
		};
	
		MTBOpArgumentDescriptor[] supplementalDescriptors = null;
	
			addDescriptors( parameterDescriptors, inputDescriptors, 
					outputDescriptors, supplementalDescriptors);
	}

	/**
	 * Type of image to be drawn:
	 * LABEL_IMAGE:  draw gray value image where regions are labeled by their index in the set + 1
	 * ID_IMAGE:     draw gray value image where regions are labeled by their region id (see MTBRegion3D.id)
	 * MASK_IMAGE:   draw a mask image with equal gray values for all regions
	 * COLOR_IMAGE:  draw a mask with colored regions
	 * TRANSPARENT_IMAGE: draw a mask with colored regions, but the regions are transparent to show the underlying intensity structure of the given target image
	 * 
	 * @author Oliver Gress
	 *
	 */
	public enum DrawType {
		LABEL_IMAGE, ID_IMAGE, MASK_IMAGE, COLOR_IMAGE, TRANSPARENT_IMAGE
	}
	
	/**
	 * Default constructor where NO parameters are set. Don't use this for region image creation.
	 */
	public DrawRegion3DSet() throws MTBOperatorException {	
	}
	
	/**
	 * Simple constructor to create the most common region image types:
	 * see DrawType 
	 * dtype = COLOR_IMAGE: Mask image with randomly colored regions
	 * dtype = MASK_IMAGE:  Mask image of type MTB_BYTE, where regions have value 255
	 * dtype = LABEL_IMAGE: Label image of type MTB_INT
	 * dtype = ID_IMAGE:    ID image of type MTB_INT
	 * dtype = TRANSPARENT_IMAGE: Image with transparent randomly colored regions, needs a RGB target image to work
	 * 
	 * @param dtype drawing type
	 * @param regions a set of regions (set xMin,xMax,yMin,yMax which are used to determine image size)
	 * @throws MTBOperatorException
	 */
	public DrawRegion3DSet(DrawType dtype, MTBRegion3DSet regions) throws MTBOperatorException {
		this.setInputRegions(regions);
		this.setDrawType(dtype);
		
		
		if (dtype == DrawType.COLOR_IMAGE) {
			this.setImageType(MTBImageType.MTB_RGB);
			this.setRandom(new Random(0));
		}
		else if (dtype == DrawType.MASK_IMAGE) {
			this.setImageType(MTBImageType.MTB_BYTE);
			this.setGrayValue(255.0);
		}
		else {
			this.setImageType(MTBImageType.MTB_INT);
		}
	}
	
	/**
	 * Simple constructor to draw the most common region image types to a given image: 
	 * dtype = COLOR_IMAGE: Mask image with randomly colored regions
	 * dtype = MASK_IMAGE:  Mask image of type MTB_BYTE, where regions have value 255
	 * dtype = LABEL_IMAGE: Label image of type MTB_INT
	 * dtype = ID_IMAGE:    ID image of type MTB_INT
	 * dtype = TRANSPARENT_IMAGE: Image with transparent randomly colored regions, needs a RGB target image to work
	 * 
	 * @param dtype drawing type
	 * @param regions a set of regions
	 * @param targetimage target image to draw the regions to
	 * @throws MTBOperatorException
	 */
	public DrawRegion3DSet(DrawType dtype, MTBRegion3DSet regions, MTBImage targetimage) throws MTBOperatorException {
		this(dtype, regions);
		this.setTargetImage(targetimage);
	}


	@Override
	public void validateCustom() throws MTBOperatorException {
		
		if (this.getDrawType() == DrawType.LABEL_IMAGE) {
			if (this.getImageType() == MTBImageType.MTB_RGB) {
				throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED,
								"DrawRegion3DSet.validateCustom(): Label image cannot be drawn to a RGB image.");
			}
			
		}
		else if (this.getDrawType() == DrawType.ID_IMAGE) {
			if (this.getImageType() == MTBImageType.MTB_RGB) {
				throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED,
								"DrawRegion3DSet.validateCustom(): ID image cannot be drawn to a RGB image.");
			}
			else if (this.getImageType() == MTBImageType.MTB_BYTE
					|| this.getImageType() == MTBImageType.MTB_SHORT) {
				System.out.println("Warning: DrawRegion3DSet.validateCustom(): ID image of regions is drawn to " +
						" a MTB_BYTE or MTB_SHORT image. IDs are integer values, possible loss of information.");
			}	
		}
		else if (this.getDrawType() == DrawType.COLOR_IMAGE) {
			if (this.getImageType() == MTBImageType.MTB_RGB) {

				if (!(this.getColor() == null ^ this.getRandom() == null))
					throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED,
								"DrawRegion3DSet.validateCustom(): For color region images, a color or a" +
								" a random number generator (random colors) must be specified (exclusively).");
			}
			else {
				throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED,
						"DrawRegion3DSet.validateCustom(): For color region images, a RGB target image must be specified.");
			}	
		}
		else if (this.getDrawType() == DrawType.TRANSPARENT_IMAGE) {
			if (this.getTargetImage() == null) {
				throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED,
					"DrawRegion3DSet.validateCustom(): For transparent color region image, you must specify a RGB target image");	
			}
			
			if (this.getImageType() == MTBImageType.MTB_RGB) {

				if (!(this.getColor() == null ^ this.getRandom() == null))
					throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED,
								"DrawRegion3DSet.validateCustom(): For transparent color region images, a color or a" +
								" a random number generator (random colors) must be specified (exclusively).");
			}
			else {
				throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED,
						"DrawRegion3DSet.validateCustom(): For transparent color region image, the output type (probably the target) image must be MTB_RGB.");
			}	
		}
		else if (this.getDrawType() == DrawType.MASK_IMAGE) {
			if (this.getGrayValue() == null) {
				throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED,
						"DrawRegion3DSet.validateCustom(): A gray value must be specified to draw a mask image.");
			}
		}
		else {
			throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED,
					"DrawRegion3DSet.validateCustom(): Unknown DrawType: " + this.getDrawType());
		}
		
	}
	
	/**
	 * Set the input regions to be drawn
	 * @param regions
	 * @throws MTBOperatorException
	 */
	public void setInputRegions(MTBRegion3DSet regions) throws MTBOperatorException {
		this.setInput("InputRegions", regions);
	}
	
	/**
	 * Get the regions that have to be drawn
	 * @return set of regions
	 * @throws MTBOperatorException
	 */
	public MTBRegion3DSet getInputRegions() throws MTBOperatorException {
		return (MTBRegion3DSet)this.getInput("InputRegions");
	}	
	
	/**
	 * Specify a target image to which the regions are drawn, specify 'null' to create a new image
	 * @param targetimage
	 * @throws MTBOperatorException
	 */
	public void setTargetImage(MTBImage targetimage) throws MTBOperatorException {
		this.setInput("TargetImage", targetimage);
		this.setImageType(targetimage.getType());
	}
	
	/**
	 * Get the target image to which the regions are drawn, if one was specified. 
	 * @return target image or null, if none was specified and the operator creates a new image
	 * @throws MTBOperatorException
	 */
	public MTBImage getTargetImage() throws MTBOperatorException {
		return (MTBImage)this.getInput("TargetImage");
	}
	
	/**
	 * Set the kind of region image to be drawn
	 * @param dtype see DrawType
	 * @throws MTBOperatorException
	 */
	public void setDrawType(DrawType dtype) throws MTBOperatorException {
		this.setParameter("DrawType", dtype);
	}
	
	/**
	 * Get the kind of region image drawn by the operator
	 * @return
	 * @throws MTBOperatorException
	 */
	public DrawType getDrawType() throws MTBOperatorException {
		return (DrawType)this.getParameter("DrawType");
	}	
	
	/**
	 * Set the resulting image's datatype. If a target image was specified and the
	 * given type differs from the target image's type, the target image is set to null.
	 * In this case a new image is created.
	 * @param type result image type
	 * @throws MTBOperatorException
	 */
	public void setImageType(MTBImageType type) throws MTBOperatorException {
		if (this.getTargetImage() != null && this.getTargetImage().getType() != type) {
			this.setTargetImage(null);
		}
		this.setParameter("ImageType", type);
	}
	
	/**
	 * Get the datatype of the resulting image
	 * @return
	 * @throws MTBOperatorException
	 */
	public MTBImageType getImageType() throws MTBOperatorException {
		return (MTBImageType)this.getParameter("ImageType");
	}	
	
	/**
	 * Set the gray value of the regions in a mask image (MASK_IMAGE only). Be aware of the
	 * resulting image's datatype.
	 * @param value
	 * @throws MTBOperatorException
	 */
	public void setGrayValue(double value) throws MTBOperatorException {
		this.setParameter("GrayValue", value);
	}
	
	/**
	 * Get the gray value of the regions in a mask image (MASK_IMAGE only).
	 * @return Gray value or null, if none was specified
	 * @throws MTBOperatorException
	 */
	public Double getGrayValue() throws MTBOperatorException {
		return (Double)this.getParameter("GrayValue");
	}
	
	/**
	 * Set the uniform color of the regions in a color image (COLOR_IMAGE only).  
	 * @param c
	 * @throws MTBOperatorException
	 */
	public void setColor(Color c) throws MTBOperatorException {
		this.setParameter("Color", c);
		this.setParameter("Random", null);
	}
	
	/**
	 * Get the uniform color of the regions in a color image (COLOR_IMAGE only).
	 * @return Color or null, if none was specified.
	 * @throws MTBOperatorException
	 */
	public Color getColor() throws MTBOperatorException {
		return (Color)this.getParameter("Color");
	}
	
	/**
	 * Set a random number generator to draw each region in a random color (COLOR_IMAGE only).
	 * @param r
	 * @throws MTBOperatorException
	 */
	public void setRandom(Random r) throws MTBOperatorException {
		this.setParameter("Random", r);
		this.setParameter("Color", null);
	}
	
	/**
	 * Get the random number generator which is responsable to draw each region in a random color (COLOR_IMAGE only).
	 * @return Random number generator or null, if none was specified
	 * @throws MTBOperatorException
	 */
	public Random getRandom() throws MTBOperatorException {
		return (Random)this.getParameter("Random");
	}
	
	/**
	 * Set the resulting image.
	 * @param image
	 * @throws MTBOperatorException
	 */
	protected void setResultImage(MTBImage image) throws MTBOperatorException {
		this.setOutput("ResultImage", image);
	}
	
	/**
	 * Get the resulting region image. If a target image (InputImage) was specified, this is the same object.
	 * @return
	 * @throws MTBOperatorException
	 */
	public MTBImage getResultImage() throws MTBOperatorException {
		return (MTBImage)this.getOutput("ResultImage");
	}
	
	@Override
	protected void operate() throws MTBOperatorException,
			MTBProcessingDAGException {

		MTBRegion3DSet regs = this.getInputRegions();
		
		DrawType dtype = this.getDrawType();
		MTBImage outImg = null;
		int x,y,z;
		int sizeX, sizeY, sizeZ;
		
		if (this.getTargetImage() != null) {
			sizeX = this.getTargetImage().getSizeX();
			sizeY = this.getTargetImage().getSizeY();
			sizeZ = this.getTargetImage().getSizeZ();
			outImg = this.getTargetImage();
		}
		else {
			sizeX = (int)Math.ceil(regs.getXmax() - regs.getXmin()) + 1;
			sizeY = (int)Math.ceil(regs.getYmax() - regs.getYmin()) + 1;
			sizeZ = (int)Math.ceil(regs.getZmax() - regs.getZmin()) + 1;
			
			outImg = MTBImage.createMTBImage(sizeX, 
												sizeY, 
												sizeZ, 
												1, 
												1,
												this.getImageType());
		}
		
		
		
		if (dtype == DrawType.COLOR_IMAGE) {
		
			Color color = this.getColor();
			Random rand = this.getRandom();

			int c = 255;
			// if output has one color for all regions
			if (color != null)
				c = this.color(color);
				
			for (int i = 0; i < regs.size(); i++) {
				Vector<MTBPoint3D> pts = regs.get(i).getPoints();
				
				// if output has random colors for the different regions
				if (rand != null) {
					c = this.randomColor(rand);
				}
				
				for (int j = 0; j < pts.size(); j++) {
					x = (int)pts.get(j).getX();
					y = (int)pts.get(j).getY();
					z = (int)pts.get(j).getZ();
					
					if (x >= 0 && x < sizeX && y >= 0 && y < sizeY && z >= 0 && z < sizeZ) {
						outImg.putValueInt(x, y, z, c);
					}	
					else {
						System.err.println("Warning: Cannot draw region pixel outside of image. Coordinate (" + x + "," + y + "," + z + "), " +
								"image size " + sizeX + "x" + sizeY + "x" + sizeZ);
					}
				}	
			}
		}
		else if (dtype == DrawType.TRANSPARENT_IMAGE) {
			
			Color color = this.getColor();
			Random rand = this.getRandom();
			Color color2;
			
			int c = 255;
			double val;
			// if output has one color for all regions
				
			for (int i = 0; i < regs.size(); i++) {
				Vector<MTBPoint3D> pts = regs.get(i).getPoints();
				
				// if output has random colors for the different regions
				if (rand != null) {
					color = this.int2Color(this.randomColor(rand));
				}
				
				for (int j = 0; j < pts.size(); j++) {
					x = (int)pts.get(j).getX();
					y = (int)pts.get(j).getY();
					z = (int)pts.get(j).getZ();
					
					if (x >= 0 && x < sizeX && y >= 0 && y < sizeY && z >= 0 && z < sizeZ) {
						color2 = this.int2Color(outImg.getValueInt(x, y, z));
						val = (color2.getRed() + color2.getGreen() + color2.getBlue())/(3.0*255.0);
						c = this.color(new Color((int)Math.round(color.getRed()*val),
													(int)Math.round(color.getGreen()*val),
													(int)Math.round(color.getBlue()*val)));
						
						outImg.putValueInt(x, y, z, c);
					}
					else {
						System.err.println("Warning: Cannot draw region pixel outside of image. Coordinate (" + x + "," + y + "," + z + "), " +
								"image size " + sizeX + "x" + sizeY + "x" + sizeZ);	
					}
				}	
			}
		}
		else {

			double v = 0.0;
			// if output has one color for all regions
			if (this.getGrayValue() != null)
				v = this.getGrayValue();
			
			for (int i = 0; i < regs.size(); i++) {
				Vector<MTBPoint3D> pts = regs.get(i).getPoints();
				
				// if output has random colors for the different regions
				if (dtype == DrawType.LABEL_IMAGE) {
					v += 1.0;
					if (v > outImg.getTypeMax()) {
						v = 1.0;
						System.err.println("Warning: Region label exceeds maximum supported pixel value. Going on resetting label to 1.");
					}
				}
				else if (dtype == DrawType.ID_IMAGE) {
					v = regs.get(i).getID();
					if (v == 0.0) {
						System.err.println("Warning: Drawing region with ID 0 as ID_IMAGE. Same pixel value as default background.");
					}
				}
				
				for (int j = 0; j < pts.size(); j++) {
					x = (int)pts.get(j).getX();
					y = (int)pts.get(j).getY();
					z = (int)pts.get(j).getZ();
					
					if (x >= 0 && x < sizeX && y >= 0 && y < sizeY && z >= 0 && z < sizeZ) {
						outImg.putValueDouble(x, y, z, v);
					}
					else {
						System.err.println("Warning: Cannot draw region pixel outside of image. Coordinate (" + x + "," + y + "," + z + "), " +
																		"image size " + sizeX + "x" + sizeY + "x" + sizeZ);
					}
				}	
			}
		}
		
		this.setResultImage(outImg);
	}

	protected int randomColor(Random r) {
		return (((((int)(r.nextDouble()*200.0 + 55)) & 0xff) << 16) 
			  + ((((int)(r.nextDouble()*200.0 + 55)) & 0xff) << 8) 
			  + (((int)(r.nextDouble()*200.0 + 55)) & 0xff));
	}
	
	protected int color(Color c) {
		return (((c.getRed() & 0xff) << 16) + ((c.getGreen() & 0xff) << 8) + (c.getBlue() & 0xff));
	}
	
	protected Color int2Color(int color) {
		return new Color(((color & 0xff0000)>>16), ((color & 0xff00)>>8), (color & 0xff));
	}

}

