/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.math;

import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.exceptions.MTBProcessingDAGException;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException.OperatorExceptionType;
import de.unihalle.informatik.MiToBo.operator.MTBOpArgumentDescriptor;
import de.unihalle.informatik.MiToBo.operator.MTBOperator;

/**
 * Image arithmetics class with self-documentation. 
 * Operations with one input image result in images of the same type!!
 * Operations with two input images result in an image of type of the higher input image precision!!
 * If two input images do not have the same size the resulting image is null.
 * Each operation is elementwise.
 * 
 * @author gress
 *
 */
public class MTBImageArithmetics {

	private final int firstOneImageOpOrdinal = 6;
	
	protected enum ArithOp {
		ADD, SUB, MULT, DIV, MIN, MAX, ADD_CONST, MULT_CONST, POW_CONST
	}

	private MTBOperator callingOperator;
	
	/** 
	 *  Constructor. (calling operator is null)
	 */
	public MTBImageArithmetics() throws MTBOperatorException {
		this.callingOperator = null;
	}

	/** 
	 *  Constructor with calling operator. You can specify the MTBOperator that calls the image arithmetic functions. This operator is 
	 *  passed to the internal image arithmetic operator class for history recording
	 */
	public MTBImageArithmetics(MTBOperator callingOperator) throws MTBOperatorException {
		this.callingOperator = callingOperator;
	}
	
	/** 
	 * Raise image values to the power of 'exponent'
	 * @param img input image
	 * @param exponent 
	 * @return result image or null if operation was not successful
	 */
	public MTBImage pow(MTBImage img, double exponent) {
		return runOperation(callingOperator, ArithOp.POW_CONST, img, exponent);
	}
	
	/**
	 * Add a constant to each image value
	 * @param img input image
	 * @param constant
	 * @return result image or null if operation was not successful
	 */
	public MTBImage add(MTBImage img, double constant) {
		return runOperation(callingOperator, ArithOp.ADD_CONST, img, constant);
	}
	
	/** 
	 * Multiply each image value by a constant
	 * @param img input image
	 * @param constant
	 * @return result image or null if operation was not successful
	 */
	public MTBImage mult(MTBImage img, double constant) {
		return runOperation(callingOperator, ArithOp.MULT_CONST, img, constant);
	}
	
	/**
	 * Add the elements two images
	 * @param img1
	 * @param img2
	 * @return result image or null if operation was not successful
	 */
	public MTBImage add(MTBImage img1, MTBImage img2) {
		return runOperation(callingOperator, ArithOp.ADD, img1, img2);
	}
	
	/**
	 * Subtract the elements two images (img1 - img2)
	 * @param img1
	 * @param img2
	 * @return result image or null if operation was not successful
	 */
	public MTBImage sub(MTBImage img1, MTBImage img2) {
		return runOperation(callingOperator, ArithOp.SUB, img1, img2);
	}	
	
	/**
	 * Multiply the elements of two images
	 * @param img1
	 * @param img2
	 * @return result image or null if operation was not successful
	 */
	public MTBImage mult(MTBImage img1, MTBImage img2) {
		return runOperation(callingOperator, ArithOp.MULT, img1, img2);
	}	
	
	/**
	 * Divide the elements of two images (img1/img2)
	 * @param img1
	 * @param img2
	 * @return result image or null if operation was not successful
	 */
	public MTBImage div(MTBImage img1, MTBImage img2) {
		return runOperation(callingOperator, ArithOp.DIV, img1, img2);
	}	
	
	/**
	 * Pixelwise minimum
	 * @param img1
	 * @param img2
	 * @return result image or null if operation was not successful
	 */
	public MTBImage min(MTBImage img1, MTBImage img2) {
		return runOperation(callingOperator, ArithOp.MIN, img1, img2);
	}	
	
	/**
	 * Pixelwise maximum
	 * @param img1
	 * @param img2
	 * @return result image or null if operation was not successful
	 */
	public MTBImage max(MTBImage img1, MTBImage img2) {
		return runOperation(callingOperator, ArithOp.MAX, img1, img2);
	}
	
	/**
	 * Run an operation with one input image and one input constant
	 * @param op
	 * @param img
	 * @param constant
	 * @return result image or null if operation was not successful
	 */
	private MTBImage runOperation(MTBOperator callingOperator, ArithOp op, MTBImage img, double constant) {
		MTBImage resultImg = null;

		ImageArithmetics ia;
		try {
			
			ia = new ImageArithmetics(op, img, constant);
			ia.runOp(callingOperator);
			resultImg = ia.getResultImg();
			
		} catch (MTBOperatorException e) {
			e.printStackTrace();
		} catch (MTBProcessingDAGException e) {
			e.printStackTrace();
		}
		
		return resultImg;
	}
	
	/**
	 * Run an operation with two input images
	 * @param callingOperator
	 * @param op
	 * @param img
	 * @param constant
	 * @return result image or null if operation was not successful
	 */
	private MTBImage runOperation(MTBOperator callingOperator, ArithOp op, MTBImage img1, MTBImage img2) {
		MTBImage resultImg = null;

		ImageArithmetics ia;
		try {
			
			ia = new ImageArithmetics(op, img1, img2);
			ia.runOp(callingOperator);
			resultImg = ia.getResultImg();
			
		} catch (MTBOperatorException e) {
			e.printStackTrace();
		} catch (MTBProcessingDAGException e) {
			e.printStackTrace();
		}
		
		return resultImg;
	}
	
	/**
	 * Self documenting image arithmetics class. Here the computations are done.
	 * @author gress
	 *
	 */
	private class ImageArithmetics extends MTBOperator {
		
  		protected void collectArguments() throws MTBOperatorException {
        		super.collectArguments();

			MTBOpArgumentDescriptor[] inputDescriptors = 
				new MTBOpArgumentDescriptor[] {
					new MTBOpArgumentDescriptor("inImg1", MTBImage.class, 
							"Input image 1", true, null),
					new MTBOpArgumentDescriptor("inImg2", MTBImage.class, 
							"Input image 2", false, null)
			};
	
			MTBOpArgumentDescriptor[] outputDescriptors = 
				new MTBOpArgumentDescriptor[] {
					new MTBOpArgumentDescriptor("resultImg", MTBImage.class, 
							"Result image", true, null)
			};
	
			MTBOpArgumentDescriptor[] parameterDescriptors = 
				new MTBOpArgumentDescriptor[] {
					new MTBOpArgumentDescriptor("arithmeticOp", ArithOp.class,
							"Arithmetic operation", true, null),
					new MTBOpArgumentDescriptor("inConst", Double.class,
							"Input constant", false, null)
			};
	
			MTBOpArgumentDescriptor[] supplementalDescriptors = null;
	
			
				addDescriptors( parameterDescriptors, inputDescriptors, 
						outputDescriptors, supplementalDescriptors);
		}


		@SuppressWarnings("unused")
		public ImageArithmetics() throws MTBOperatorException {
			
		}
		
		protected ImageArithmetics(ArithOp op, MTBImage inputImage1, MTBImage inputImage2) throws MTBOperatorException {
			this.setArithmeticOperation(op);
			this.setInImg1(inputImage1);
			this.setInImg2(inputImage2);
		}
		
		protected ImageArithmetics(ArithOp op, MTBImage inputImage, double inputConstant) throws MTBOperatorException {
			this.setArithmeticOperation(op);
			this.setInImg1(inputImage);
			this.setInConst(inputConstant);
		}
		
		@Override
		public void validateCustom() throws MTBOperatorException {
			
			boolean valid = false;
			

				if (this.getArithmeticOperation().ordinal() < firstOneImageOpOrdinal) {
					// operations with two images
					valid = (this.getInImg1() != null && this.getInImg2() != null && this.getInImg1().equalSize(this.getInImg2()));
					
					if (!valid)
						throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED, "MTBImageArithmetics.ImageArithmetics.validateCustom(): " +
									     													"Operation needs two images of equal size.");
				}
				else {
					// operations with one image and one constant
					valid = (this.getInImg1() != null && this.getInConst() != null);
					
					if (!valid)
						throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED, "GaussFilter.validateCustom(): " +
									     													"Operation needs an image and a constant.");
				}
		}
		
		

		@Override
		protected void operate() throws MTBOperatorException {
			MTBImage resultImg = null;
			
			ArithOp op = this.getArithmeticOperation();
			
			
			if (op.ordinal() < firstOneImageOpOrdinal) {
				// Operations with two images	
				
				MTBImage inImg1 = this.getInImg1();
				MTBImage inImg2 = this.getInImg2();
			
				if (op == ArithOp.ADD)
					resultImg = this.add(inImg1, inImg2);
				else if (op == ArithOp.SUB)
					resultImg = this.sub(inImg1, inImg2);				
				else if (op == ArithOp.MULT)
					resultImg = this.mult(inImg1, inImg2);
				else if (op == ArithOp.DIV)
					resultImg = this.div(inImg1, inImg2);
				else if (op == ArithOp.MIN)
					resultImg = this.min(inImg1, inImg2);
				else if (op == ArithOp.MAX)
					resultImg = this.max(inImg1, inImg2);
				
				if (resultImg != null) {
					this.setResultImg(resultImg);
				}		
			}
			else {
			// Operations with one image and one constant	
				
				MTBImage inImg = this.getInImg1();
				double inConst = this.getInConst();
			
				if (op == ArithOp.ADD_CONST)
					resultImg = this.add(inImg, inConst);
				else if (op == ArithOp.MULT_CONST)
					resultImg = this.mult(inImg, inConst);
				else if (op == ArithOp.POW_CONST)
					resultImg = this.pow(inImg, inConst);
				
				if (resultImg != null) {
					this.setResultImg(resultImg);
				}
			}			
					
			if (resultImg == null) 
				throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, "MTBImageArithmetics.ImageArithmetics.operate() failed: Result image is 'null'");
		}
		
		
		protected MTBImage getInImg1() throws MTBOperatorException {
			return (MTBImage)this.getInput("inImg1");
		}
		
		protected void setInImg1(MTBImage inImg1) throws MTBOperatorException {
			this.setInput("inImg1", inImg1);
		}
			
		protected MTBImage getInImg2() throws MTBOperatorException {
			return (MTBImage)this.getInput("inImg2");
		}
		
		protected void setInImg2(MTBImage inImg2) throws MTBOperatorException {
			this.setInput("inImg2", inImg2);
		}		
		
		protected Double getInConst() throws MTBOperatorException {
			return (Double)this.getParameter("inConst");
		}
		
		protected void setInConst(double inConst) throws MTBOperatorException {
			this.setParameter("inConst", inConst);
		}			
		
		protected ArithOp getArithmeticOperation() throws MTBOperatorException {
			return (ArithOp)this.getParameter("arithmeticOp");
		}
		
		protected void setArithmeticOperation(ArithOp op) throws MTBOperatorException {
			this.setParameter("arithmeticOp", op);
		}	
		
		protected MTBImage getResultImg() throws MTBOperatorException {
			return (MTBImage)this.getOutput("resultImg");
		}
		
		protected void setResultImg(MTBImage op) throws MTBOperatorException {
			this.setOutput("resultImg", op);
		}		
		
		/**
		 * Add a constant to each image value
		 * @param img input image
		 * @param constant
		 * @return
		 */
		protected MTBImage add(MTBImage img, double constant) {
			int sizeStack = img.getSizeStack();
			int sizeY = img.getSizeY();
			int sizeX = img.getSizeX();
			
			int idx = img.getActualSliceIndex();
			
			MTBImage newImg = MTBImage.createMTBImage(sizeX, sizeY, img.getSizeZ(), img.getSizeT(), img.getSizeC(), img.getType());

			newImg.setTitle(img.getTitle());
			newImg.setStepsizeX(img.getStepsizeX());
			newImg.setStepsizeY(img.getStepsizeY());
			newImg.setStepsizeZ(img.getStepsizeZ());
			newImg.setStepsizeT(img.getStepsizeT());
			newImg.setUnitX(img.getUnitX());
			newImg.setUnitY(img.getUnitY());
			newImg.setUnitZ(img.getUnitZ());
			newImg.setUnitT(img.getUnitT());
			
			for (int i = 0; i < sizeStack; i++) {
				img.setActualSliceIndex(i);
				newImg.setActualSliceIndex(i);
				
				for (int y = 0; y < sizeY; y++) {
					for (int x = 0; x < sizeX; x++) {
						newImg.putValueDouble(x, y, img.getValueDouble(x,y) + constant);
					}
				}
			}
			
			// restore actual slice index
			img.setActualSliceIndex(idx);
			newImg.setActualSliceIndex(0);
			
			return newImg;
		}
		
		/**
		 * Multiply each image value by a constant
		 * @param img input image
		 * @param constant
		 * @return
		 */
		protected MTBImage mult(MTBImage img, double constant) {
			int sizeStack = img.getSizeStack();
			int sizeY = img.getSizeY();
			int sizeX = img.getSizeX();
			
			int idx = img.getActualSliceIndex();
			
			MTBImage newImg = MTBImage.createMTBImage(sizeX, sizeY, img.getSizeZ(), img.getSizeT(), img.getSizeC(), img.getType());

			newImg.setTitle(img.getTitle());
			newImg.setStepsizeX(img.getStepsizeX());
			newImg.setStepsizeY(img.getStepsizeY());
			newImg.setStepsizeZ(img.getStepsizeZ());
			newImg.setStepsizeT(img.getStepsizeT());
			newImg.setUnitX(img.getUnitX());
			newImg.setUnitY(img.getUnitY());
			newImg.setUnitZ(img.getUnitZ());
			newImg.setUnitT(img.getUnitT());
			
			for (int i = 0; i < sizeStack; i++) {
				img.setActualSliceIndex(i);
				newImg.setActualSliceIndex(i);
				
				for (int y = 0; y < sizeY; y++) {
					for (int x = 0; x < sizeX; x++) {
						newImg.putValueDouble(x, y, img.getValueDouble(x,y) * constant);
					}
				}
			}
			
			// restore actual slice index
			img.setActualSliceIndex(idx);
			newImg.setActualSliceIndex(0);
			
			return newImg;
		}	
		
		/**
		 * Raise image values to the power of 'constant'
		 * @param img
		 * @param constant
		 * @return
		 */
		protected MTBImage pow(MTBImage img, double constant) {
			int sizeStack = img.getSizeStack();
			int sizeY = img.getSizeY();
			int sizeX = img.getSizeX();
			
			int idx = img.getActualSliceIndex();
			
			MTBImage newImg = MTBImage.createMTBImage(sizeX, sizeY, img.getSizeZ(), img.getSizeT(), img.getSizeC(), img.getType());

			newImg.setTitle(img.getTitle());
			newImg.setStepsizeX(img.getStepsizeX());
			newImg.setStepsizeY(img.getStepsizeY());
			newImg.setStepsizeZ(img.getStepsizeZ());
			newImg.setStepsizeT(img.getStepsizeT());
			newImg.setUnitX(img.getUnitX());
			newImg.setUnitY(img.getUnitY());
			newImg.setUnitZ(img.getUnitZ());
			newImg.setUnitT(img.getUnitT());
			
			for (int i = 0; i < sizeStack; i++) {
				img.setActualSliceIndex(i);
				newImg.setActualSliceIndex(i);
				
				for (int y = 0; y < sizeY; y++) {
					for (int x = 0; x < sizeX; x++) {
						newImg.putValueDouble(x, y, Math.pow(img.getValueDouble(x,y), constant));
					}
				}
			}
			
			// restore actual slice index
			img.setActualSliceIndex(idx);
			newImg.setActualSliceIndex(0);
			
			return newImg;
		}		
		
		/**
		 * Add the elements of two images
		 * @param img1
		 * @param img2
		 * @return
		 */
		protected MTBImage add(MTBImage img1, MTBImage img2) {
			int sizeStack = img1.getSizeStack();
			int sizeY = img1.getSizeY();
			int sizeX = img1.getSizeX();
			
			int idx1 = img1.getActualSliceIndex();
			int idx2 = img2.getActualSliceIndex();
			
			MTBImage newImg = MTBImage.createMTBImage(sizeX, sizeY, img1.getSizeZ(), img1.getSizeT(), img1.getSizeC(), 
					(img1.getType().ordinal() >= img2.getType().ordinal() ? img1.getType() : img2.getType()));

			newImg.setTitle(img1.getTitle());
			newImg.setStepsizeX(img1.getStepsizeX());
			newImg.setStepsizeY(img1.getStepsizeY());
			newImg.setStepsizeZ(img1.getStepsizeZ());
			newImg.setStepsizeT(img1.getStepsizeT());
			newImg.setUnitX(img1.getUnitX());
			newImg.setUnitY(img1.getUnitY());
			newImg.setUnitZ(img1.getUnitZ());
			newImg.setUnitT(img1.getUnitT());
			
			for (int i = 0; i < sizeStack; i++) {
				img1.setActualSliceIndex(i);
				img2.setActualSliceIndex(i);
				newImg.setActualSliceIndex(i);
				
				for (int y = 0; y < sizeY; y++) {
					for (int x = 0; x < sizeX; x++) {
						newImg.putValueDouble(x, y, img1.getValueDouble(x,y) + img2.getValueDouble(x,y));
					}
				}
			}
			
			// restore actual slice index
			img1.setActualSliceIndex(idx1);
			img2.setActualSliceIndex(idx2);
			newImg.setActualSliceIndex(0);
			
			return newImg;
		}
		
		/**
		 * Subtract the elements of two images (img1 - img2)
		 * @param img1
		 * @param img2
		 * @return
		 */
		protected MTBImage sub(MTBImage img1, MTBImage img2) {
			int sizeStack = img1.getSizeStack();
			int sizeY = img1.getSizeY();
			int sizeX = img1.getSizeX();
			
			int idx1 = img1.getActualSliceIndex();
			int idx2 = img2.getActualSliceIndex();
			
			MTBImage newImg = MTBImage.createMTBImage(sizeX, sizeY, img1.getSizeZ(), img1.getSizeT(), img1.getSizeC(), 
										(img1.getType().ordinal() >= img2.getType().ordinal() ? img1.getType() : img2.getType()));
			
			newImg.setTitle(img1.getTitle());
			newImg.setStepsizeX(img1.getStepsizeX());
			newImg.setStepsizeY(img1.getStepsizeY());
			newImg.setStepsizeZ(img1.getStepsizeZ());
			newImg.setStepsizeT(img1.getStepsizeT());
			newImg.setUnitX(img1.getUnitX());
			newImg.setUnitY(img1.getUnitY());
			newImg.setUnitZ(img1.getUnitZ());
			newImg.setUnitT(img1.getUnitT());

			
			for (int i = 0; i < sizeStack; i++) {
				img1.setActualSliceIndex(i);
				img2.setActualSliceIndex(i);
				newImg.setActualSliceIndex(i);
				
				for (int y = 0; y < sizeY; y++) {
					for (int x = 0; x < sizeX; x++) {
						newImg.putValueDouble(x, y, img1.getValueDouble(x,y) - img2.getValueDouble(x,y));
					}
				}
			}
			
			// restore actual slice index
			img1.setActualSliceIndex(idx1);
			img2.setActualSliceIndex(idx2);
			newImg.setActualSliceIndex(0);
			
			return newImg;
		}
		
		/**
		 * Multiply the elements of two images
		 * @param img1
		 * @param img2
		 * @return
		 */
		protected MTBImage mult(MTBImage img1, MTBImage img2) {
			int sizeStack = img1.getSizeStack();
			int sizeY = img1.getSizeY();
			int sizeX = img1.getSizeX();
			
			int idx1 = img1.getActualSliceIndex();
			int idx2 = img2.getActualSliceIndex();
			
			MTBImage newImg = MTBImage.createMTBImage(sizeX, sizeY, img1.getSizeZ(), img1.getSizeT(), img1.getSizeC(), 
					(img1.getType().ordinal() >= img2.getType().ordinal() ? img1.getType() : img2.getType()));

			newImg.setTitle(img1.getTitle());
			newImg.setStepsizeX(img1.getStepsizeX());
			newImg.setStepsizeY(img1.getStepsizeY());
			newImg.setStepsizeZ(img1.getStepsizeZ());
			newImg.setStepsizeT(img1.getStepsizeT());
			newImg.setUnitX(img1.getUnitX());
			newImg.setUnitY(img1.getUnitY());
			newImg.setUnitZ(img1.getUnitZ());
			newImg.setUnitT(img1.getUnitT());
			
			for (int i = 0; i < sizeStack; i++) {
				img1.setActualSliceIndex(i);
				img2.setActualSliceIndex(i);
				newImg.setActualSliceIndex(i);
				
				for (int y = 0; y < sizeY; y++) {
					for (int x = 0; x < sizeX; x++) {
						newImg.putValueDouble(x, y, img1.getValueDouble(x,y) * img2.getValueDouble(x,y));
					}
				}
			}
			
			// restore actual slice index
			img1.setActualSliceIndex(idx1);
			img2.setActualSliceIndex(idx2);
			newImg.setActualSliceIndex(0);
			
			return newImg;
		}
		
		/**
		 * Divide the elements of two images (img1/img2)
		 * @param img1
		 * @param img2
		 * @return
		 */
		protected MTBImage div(MTBImage img1, MTBImage img2) {
			int sizeStack = img1.getSizeStack();
			int sizeY = img1.getSizeY();
			int sizeX = img1.getSizeX();
			
			int idx1 = img1.getActualSliceIndex();
			int idx2 = img2.getActualSliceIndex();
			
			MTBImage newImg = MTBImage.createMTBImage(sizeX, sizeY, img1.getSizeZ(), img1.getSizeT(), img1.getSizeC(), 
					(img1.getType().ordinal() >= img2.getType().ordinal() ? img1.getType() : img2.getType()));

			newImg.setTitle(img1.getTitle());
			newImg.setStepsizeX(img1.getStepsizeX());
			newImg.setStepsizeY(img1.getStepsizeY());
			newImg.setStepsizeZ(img1.getStepsizeZ());
			newImg.setStepsizeT(img1.getStepsizeT());
			newImg.setUnitX(img1.getUnitX());
			newImg.setUnitY(img1.getUnitY());
			newImg.setUnitZ(img1.getUnitZ());
			newImg.setUnitT(img1.getUnitT());
			
			for (int i = 0; i < sizeStack; i++) {
				img1.setActualSliceIndex(i);
				img2.setActualSliceIndex(i);
				newImg.setActualSliceIndex(i);
				
				for (int y = 0; y < sizeY; y++) {
					for (int x = 0; x < sizeX; x++) {
						newImg.putValueDouble(x, y, img1.getValueDouble(x,y) / img2.getValueDouble(x,y));
					}
				}
			}
			
			// restore actual slice index
			img1.setActualSliceIndex(idx1);
			img2.setActualSliceIndex(idx2);
			newImg.setActualSliceIndex(0);
			
			return newImg;
		}
		
		/**
		 * Pixelwise minimum
		 * @param img1
		 * @param img2
		 * @return
		 */
		protected MTBImage min(MTBImage img1, MTBImage img2) {
			int sizeStack = img1.getSizeStack();
			int sizeY = img1.getSizeY();
			int sizeX = img1.getSizeX();
			
			int idx1 = img1.getActualSliceIndex();
			int idx2 = img2.getActualSliceIndex();
			
			double v1, v2;
			
			
			MTBImage newImg = MTBImage.createMTBImage(sizeX, sizeY, img1.getSizeZ(), img1.getSizeT(), img1.getSizeC(), 
					(img1.getType().ordinal() >= img2.getType().ordinal() ? img1.getType() : img2.getType()));

			newImg.setTitle(img1.getTitle());
			newImg.setStepsizeX(img1.getStepsizeX());
			newImg.setStepsizeY(img1.getStepsizeY());
			newImg.setStepsizeZ(img1.getStepsizeZ());
			newImg.setStepsizeT(img1.getStepsizeT());
			newImg.setUnitX(img1.getUnitX());
			newImg.setUnitY(img1.getUnitY());
			newImg.setUnitZ(img1.getUnitZ());
			newImg.setUnitT(img1.getUnitT());
			
			for (int i = 0; i < sizeStack; i++) {
				img1.setActualSliceIndex(i);
				img2.setActualSliceIndex(i);
				newImg.setActualSliceIndex(i);
				
				for (int y = 0; y < sizeY; y++) {
					for (int x = 0; x < sizeX; x++) {
						v1 = img1.getValueDouble(x,y);
						v2 = img2.getValueDouble(x,y);
						newImg.putValueDouble(x, y, v1 <= v2 ? v1 : v2);
					}
				}
			}
			
			// restore actual slice index
			img1.setActualSliceIndex(idx1);
			img2.setActualSliceIndex(idx2);
			newImg.setActualSliceIndex(0);
			
			return newImg;
		}
		
		/**
		 * Pixelwise minimum
		 * @param img1
		 * @param img2
		 * @return
		 */
		protected MTBImage max(MTBImage img1, MTBImage img2) {
			int sizeStack = img1.getSizeStack();
			int sizeY = img1.getSizeY();
			int sizeX = img1.getSizeX();
			
			int idx1 = img1.getActualSliceIndex();
			int idx2 = img2.getActualSliceIndex();
			
			double v1, v2;
			
			
			MTBImage newImg = MTBImage.createMTBImage(sizeX, sizeY, img1.getSizeZ(), img1.getSizeT(), img1.getSizeC(), 
					(img1.getType().ordinal() >= img2.getType().ordinal() ? img1.getType() : img2.getType()));

			newImg.setTitle(img1.getTitle());
			newImg.setStepsizeX(img1.getStepsizeX());
			newImg.setStepsizeY(img1.getStepsizeY());
			newImg.setStepsizeZ(img1.getStepsizeZ());
			newImg.setStepsizeT(img1.getStepsizeT());
			newImg.setUnitX(img1.getUnitX());
			newImg.setUnitY(img1.getUnitY());
			newImg.setUnitZ(img1.getUnitZ());
			newImg.setUnitT(img1.getUnitT());
			
			for (int i = 0; i < sizeStack; i++) {
				img1.setActualSliceIndex(i);
				img2.setActualSliceIndex(i);
				newImg.setActualSliceIndex(i);
				
				for (int y = 0; y < sizeY; y++) {
					for (int x = 0; x < sizeX; x++) {
						v1 = img1.getValueDouble(x,y);
						v2 = img2.getValueDouble(x,y);
						newImg.putValueDouble(x, y, v1 >= v2 ? v1 : v2);
					}
				}
			}
			
			// restore actual slice index
			img1.setActualSliceIndex(idx1);
			img2.setActualSliceIndex(idx2);
			newImg.setActualSliceIndex(0);
			
			return newImg;
		}

	}
	
}
