/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.morphology;

import de.unihalle.informatik.MiToBo.datatypes.images.*;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.exceptions.*;
import de.unihalle.informatik.MiToBo.operator.*;

/**
 * This class implements morphological operations on grayscale images.
 * <p>
 * The class applies squared masks to the given image and performs
 * the chosen operation. The output image is of the same type as
 * the input image. For binary images the result should be the same
 * as for binary morphology operators.
 * 
 *  Attention: if masksize is even, errors may result due 
 *             to lack of operator symmetry
 *
 * @author moeller
 */
public class GrayscaleMorphology extends MTBOperator {

	protected void collectArguments() throws MTBOperatorException {
        super.collectArguments();

		MTBOpArgumentDescriptor[] inputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("inImg", MTBImage.class, 
											"Input image", true, null)
			};
	
		MTBOpArgumentDescriptor[] outputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("resultImg", MTBImage.class, 
						"Result image", true, null)
		};
	
		MTBOpArgumentDescriptor[] parameterDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("processMode", gwMorphOp.class,
						"Processing Mode", true, null),			
				new MTBOpArgumentDescriptor("masksize", java.lang.Integer.class,
						"Masksize", true, null)
		};
	
		MTBOpArgumentDescriptor[] supplementalDescriptors = null;
	
		
		addDescriptors( parameterDescriptors, inputDescriptors, 
				outputDescriptors, supplementalDescriptors);
	}

	/**
	 * Available operating modes.
	 * 
	 * @author moeller
	 */
	public static enum gwMorphOp {
		/**
		 * Grayscale erosion.
		 */
		GWM_ERODE,
		/**
		 * Grayscale dilation.
		 */
		GWM_DILATE,
		/**
		 * Grayscale opening.
		 */
		GWM_OPEN,
		/**
		 * Grayscale closing.
		 */
		GWM_CLOSE,
		/**
		 * Top hat operator for bright 'mountains' on dark ground.
		 */
		GWM_TOPHAT_CONVEX,
		/**
		 * Top hat operator for dark 'valleys' in bright ground.
		 */
		GWM_TOPHAT_CONCAVE
	}
	
	public GrayscaleMorphology() throws MTBOperatorException {
		// nothing to do here
	}

	/**
	 * Default constructor, an empty meta data object is initialized. 
	 * 
	 * @param ip	Image to work on. 
	 * @param op	Operator mode.
	 * @param masksize	Size of squared mask.
	 * @throws MTBProcessingDAGException 
	 */
	public GrayscaleMorphology(MTBImage inImg, gwMorphOp op, int msize) 
		throws MTBOperatorException {
			this.setInput("inImg", inImg);
			this.setParameter("processMode",op);
			this.setParameter("masksize",new Integer(msize));
	}

	/**
	 * Get reference to the current input image.
	 * 
	 * @return	Input image to work on.
	 * @throws MTBProcessingDAGException 
	 */
	public MTBImage getInputImage() throws MTBOperatorException {
		return (MTBImage)(this.getInput("inImg"));
	}
	/**
	 * Get the desired operator mode.
	 * 
	 * @return	Modus of operator.
	 * @throws MTBProcessingDAGException 
	 */
	public gwMorphOp getMode() throws MTBOperatorException {
		return (gwMorphOp)(this.getParameter("processMode"));
	}
	/**
	 * Get current masksize.
	 * 
	 * @return	Masksize of dilation mask.
	 * @throws MTBProcessingDAGException 
	 */
	public int getMasksize() throws MTBOperatorException {
		return ((Integer)(this.getParameter("masksize"))).intValue();
	}
	/**
	 * Get the result image after applying the operator.
	 * Attention, reference might be null.
	 * 
	 * @return	Reference to result image.
	 * @throws MTBProcessingDAGException 
	 */
	public MTBImageByte getResultImage() throws MTBOperatorException {
	    return (MTBImageByte)(this.getOutput("resultImg"));
	}
	
	/**
	 * Set the result image.
	 * 
	 * @param rimage	Result image.
	 * @throws MTBProcessingDAGException 
	 */
	public void setResultImage(MTBImage rimage) throws MTBOperatorException {
		this.setOutput("resultImg", rimage);
	}

	/**
	 * MTB input image to work on.
	 */
	MTBImage mtbImg;
	
	@Override
	protected void operate() throws MTBOperatorException {
		
		// convert the parameters
		this.mtbImg= this.getInputImage();

		// prepare result image
		MTBImageByte result= null;

		// get the desired operator mode
		switch(this.getMode())
		{
		case GWM_ERODE:
			result= this.erodeImage(this.mtbImg, this.getMasksize());
			break;
		case GWM_DILATE:
			result= this.dilateImage(this.mtbImg, this.getMasksize());
			break;
		case GWM_OPEN:
			result= this.openImage(this.mtbImg, this.getMasksize());
			break;
		case GWM_CLOSE:
			result= this.closeImage(this.mtbImg, this.getMasksize());
			break;
		case GWM_TOPHAT_CONVEX:
			result= this.tophat_convex(this.mtbImg, this.getMasksize());
			break;
		case GWM_TOPHAT_CONCAVE:
			result= this.tophat_concave(this.mtbImg, this.getMasksize());
			break;
		}
		// set result image
		this.setResultImage(result);
	}
	
	/**
	 * Apply a grayscale erosion to the image.
	 * <p>
	 * The erosion is performed assuming a squared structural 
	 * element with specified mask size and grayscale entries of 0.
	 * 
	 * @param inImg	Input image.
	 * @param msize	Size of mask.
	 * @return	Eroded image.
	 */
	private MTBImageByte erodeImage(MTBImage inImg, int msize) {

		int width= inImg.getSizeX();
		int height= inImg.getSizeY();
		
		int masksize_2= (int)(msize/2.0);
		
		MTBImageByte result= 
			(MTBImageByte)MTBImage.createMTBImage(width, height, 
											1, 1, 1, MTBImageType.MTB_BYTE); 

		// erode the image
		for (int y=0; y<height;++y) {
			for (int x=0; x<width; ++x) {
				int min= Integer.MAX_VALUE;
				for (int dy= -masksize_2; dy<=masksize_2; ++dy) {
					if (y+dy < 0 || y+dy >= height)
						continue;
					for (int dx=-masksize_2; dx<=masksize_2; ++dx) {
						if (x+dx < 0 || x+dx >= width)
							continue;
						if (inImg.getValueInt(x+dx,y+dy) < min)
							min= inImg.getValueInt(x+dx,y+dy);
					}
				}
				result.putValueInt(x, y, min);
			}
		}
		return result;
	}

	/**
	 * Apply a grayscale dilation to the image.
	 * <p>
	 * The dilation is performed assuming a squared structural 
	 * element with specified mask size and grayscale entries of 0.
	 * 
	 * @param inImg	Input image.
	 * @param msize	Size of mask.
	 * @return	Dilated image.
	 */
	private MTBImageByte dilateImage(MTBImage inImg, int msize) {

		int width= inImg.getSizeX();
		int height= inImg.getSizeY();
		
		int masksize_2= (int)(msize/2.0);
		
		MTBImageByte result= 
			(MTBImageByte)MTBImage.createMTBImage(width, height, 
											1, 1, 1, MTBImageType.MTB_BYTE); 

		// erode the image
		for (int y=0; y<height;++y) {
			for (int x=0; x<width; ++x) {
				int max= 0;
				for (int dy= -masksize_2; dy<=masksize_2; ++dy) {
					if (y+dy < 0 || y+dy >= height)
						continue;
					for (int dx=-masksize_2; dx<=masksize_2; ++dx) {
						if (x+dx < 0 || x+dx >= width)
							continue;
						if (inImg.getValueInt(x+dx,y+dy) > max)
							max= inImg.getValueInt(x+dx,y+dy);
					}
				}
				result.putValueInt(x, y, max);
			}
		}
		return result;
	}

	/**
	 * Apply a grayscale opening to the image.
	 * <p>
	 * The opening is performed assuming a squared structural 
	 * element with specified mask size and grayscale entries of 0.
	 * 
	 * @param inImg	Input image.
	 * @param msize	Size of mask.
	 * @return	Opened image.
	 */
	private MTBImageByte openImage(MTBImage inImg, int msize) {
		MTBImageByte erodedImg= this.erodeImage(inImg, msize);
		MTBImageByte openedImg= this.dilateImage(erodedImg, msize);
		return openedImg;
	}

	/**
	 * Apply a grayscale closing to the image.
	 * <p>
	 * The closing is performed assuming a squared structural 
	 * element with specified mask size and grayscale entries of 0.
	 * 
	 * @param inImg	Input image.
	 * @param msize	Size of mask.
	 * @return	Closed image.
	 */
	private MTBImageByte closeImage(MTBImage inImg, int msize) {
		MTBImageByte dilatedImg= this.dilateImage(inImg, msize);
		MTBImageByte closedImg= this.erodeImage(dilatedImg, msize);
		return closedImg;
	}

	/**
	 * Apply a top hat operator to the image for convexity detection.
	 * <p>
	 * The operation is performed assuming a squared structural 
	 * element with specified mask size and grayscale entries of 0.
	 * 
	 * @param inImg	Input image.
	 * @param msize	Size of mask.
	 * @return	Detection result (positiv answers = white).
	 */
	private MTBImageByte tophat_convex(MTBImage inImg, int msize) {
		MTBImageByte openedImg= this.openImage(inImg, msize);
		MTBImageByte diffImg= 
			(MTBImageByte)MTBImage.createMTBImage(inImg.getSizeX(), 
							inImg.getSizeY(), 1, 1, 1, MTBImageType.MTB_BYTE); 
		for (int y=0;y<inImg.getSizeY();++y) {
			for (int x=0;x<inImg.getSizeX();++x) {
				diffImg.putValueInt(x, y, 
					inImg.getValueInt(x, y)- openedImg.getValueInt(x, y));
			}
		}
		return diffImg;
	}

	/**
	 * Apply a top hat operator to the image for concavity detection.
	 * <p>
	 * The operation is performed assuming a squared structural 
	 * element with specified mask size and grayscale entries of 0.
	 * 
	 * @param inImg	Input image.
	 * @param msize	Size of mask.
	 * @return	Detection result (positiv answers = white).
	 */
	private MTBImageByte tophat_concave(MTBImage inImg, int msize) {
		MTBImageByte closedImg= this.closeImage(inImg, msize);
		MTBImageByte diffImg= 
			(MTBImageByte)MTBImage.createMTBImage(inImg.getSizeX(), 
							inImg.getSizeY(), 1, 1, 1, MTBImageType.MTB_BYTE); 
		for (int y=0;y<inImg.getSizeY();++y) {
			for (int x=0;x<inImg.getSizeX();++x) {
				diffImg.putValueInt(x, y, 
					closedImg.getValueInt(x, y) - inImg.getValueInt(x, y));
			}
		}
		return diffImg;
	}
}
