/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.morphology;

import de.unihalle.informatik.MiToBo.datatypes.images.*;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.exceptions.*;
import de.unihalle.informatik.MiToBo.operator.*;

/**
 * This class implements a morphological dilation on binary images.
 * It assumes that foreground pixels have a value larger than 0,
 * and that background pixels have value 0. Structuring element is
 * a square matrix of size 'masksize' x 'masksize', with reference
 * pixel in the center of the matrix.
 * 
 * Attention: if masksize is even, errors may result due 
 *            to lack of operator symmetry
 *
 * @author moeller
 */
public class ImgDilate extends MTBOperator {

	protected void collectArguments() throws MTBOperatorException {
        super.collectArguments();

	MTBOpArgumentDescriptor[] inputDescriptors = 
		new MTBOpArgumentDescriptor[] {
			new MTBOpArgumentDescriptor("inImg", MTBImage.class, 
																	"Input image", true, null)
		};

		MTBOpArgumentDescriptor[] outputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("resultImg", MTBImageByte.class, 
																		"Result image", true, null)
		};

		MTBOpArgumentDescriptor[] parameterDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("masksize", java.lang.Integer.class,
																		"Masksize", true, null)
		 	};
		
		MTBOpArgumentDescriptor[] supplementalDescriptors = null;

			addDescriptors( parameterDescriptors, inputDescriptors, 
					outputDescriptors, supplementalDescriptors);
		}

		public ImgDilate() throws MTBOperatorException {
			// nothing to do here
		}

		public ImgDilate(MTBImage inimg, int masksize) throws MTBOperatorException {
	      this.setInput("inImg", inimg);
				this.setParameter("masksize", new Integer(masksize));
		}

		public MTBImage getInputImage() throws MTBOperatorException {
	      return ((MTBImage)this.getInput("inImg"));
		}

		public int getMasksize() throws MTBOperatorException {
			return ((Integer)this.getParameter("masksize")).intValue();
		}
		
		public MTBImageByte getResultImage() throws MTBOperatorException {
			return (MTBImageByte)this.getOutput("resultImg");
		}
		
		private void setResultImage(MTBImageByte result) throws MTBOperatorException {
			this.setOutput("resultImg", result);
		}
		
		/**
		 * This method does the actual work. 
		 * @throws MTBProcessingDAGException 
		 */
		@Override
		protected void operate() throws MTBOperatorException {
			MTBImageByte result= 
				this.applyMask(this.getInputImage(), this.getMasksize());
			this.setResultImage(result);
		}
	
	/**
	 * Applies a dilation mask to the given intensity image. 
	 * 
	 * Pixels with intensities larger than zero are assumed to 
	 * belong to the foreground while pixels with intensity 
	 * zero are handled as background pixels.
	 * 
	 * @param ip	Input image for dilation.
	 * @param msize	Size of squared dilation mask.
	 * @return	Dilated result image.
	 */
	private MTBImageByte applyMask(MTBImage inImg, int msize) {

		int width= inImg.getSizeX();
		int height= inImg.getSizeY();
		
		int masksize_2= (int)(msize/2.0);
		
		MTBImageByte result= 
			(MTBImageByte)MTBImage.createMTBImage(width,height,1,1,1,
					                                             MTBImageType.MTB_BYTE); 

		// dilate the image
		for (int y=0; y<height;++y) {
			for (int x=0; x<width; ++x) {
				result.putValueInt(x, y, 0);
				for (int dy= -masksize_2; dy<=masksize_2; ++dy) {
					if (y+dy < 0 || y+dy >= height)
						continue;
					for (int dx=-masksize_2; dx<=masksize_2; ++dx) {
						if (x+dx < 0 || x+dx >= width)
							continue;
						if (inImg.getValueInt(x+dx,y+dy) > 0) {
							result.putValueInt(x, y, 255);
							dx= masksize_2+1;
							dy= masksize_2+1;
						}
					}
				}
			}
		}
		return result;
	}
}
