/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.operator;

import java.util.*;

import org.apache.xmlbeans.*;

public class MTBOpNode {

	private boolean verbose = true;

	private MTBInputPort[] inputPorts;  // the input ports of this opNode
	private MTBOutputPort[] outputPorts; // the output ports of this opNode
	//private MTBOperator  op;	   // give the type of operator of this opNode
	private String name;			// name of the operator
	private String version;			// version of the operator
	private Class operatorClass;			// class of the operator

	private Hashtable<String, String> parameterHash;
	private XmlObject parameterHashAsXml; // parameteHash as an XmlObject

	private MTBOpNode parent; // parent of this opNode, will be set during back tracing
	private Vector<MTBOpNode> children; // children of this opNode, will be set during back tracing
	private Vector<MTBOpNode> directlyRegisteredChildren; // children of this opNode, set when calling operators by runOp
	private int depth; // depth in calling stack of Operators, set when tracing back
	private Vector<MTBDataPort> includedData; // MTBDataPorts created within this opNode

	public MTBOpNode( MTBOperator op) {
		this.depth = Integer.MAX_VALUE;

		this.name = op.getName();
		this.operatorClass = op.getClass();
		this.version = op.getVersion();

		try { 
			inputPorts = new MTBInputPort[ op.getNumInputs()];
			String[] inputNames = op.getInputNames();
			for ( int i=0 ; i < op.getNumInputs() ; i++ )
				inputPorts[i] = new MTBInputPort( this, i, inputNames[i]);
	
			outputPorts = new MTBOutputPort[ op.getNumOutputs()];
			String[] outputNames = op.getOutputNames();
			for ( int i=0 ; i < op.getNumOutputs() ; i++ )
				outputPorts[i] = new MTBOutputPort( this, i, outputNames[i]);
	
			parameterHash = new Hashtable<String, String>();
	
			children = new Vector<MTBOpNode>();
			directlyRegisteredChildren = new Vector<MTBOpNode>();
			includedData = new Vector<MTBDataPort>();

			// copy the parameters of the descriptor to a private hash
			// this is necessary, as the descriptor may be reused lateron with different
			// values of its parameters (or the values changes for another purpose)
			String[] parameterNames = op.getParameterNames();
			for ( int n = 0 ; n < op.getNumParameters() ; n++ ) {
				if ( op.getParameter( parameterNames[n]) != null ) 
					parameterHash.put( op.getParameterDescriptor( parameterNames[n]).name, 
								op.getParameter( parameterNames[n]).toString());
			}
			//parameterHashAsXml = op.parametersToXmlObject();
		} catch (Exception e) {
			e.printStackTrace();
		} 
	}

	public String getName() {
		return name;
	}

	public String getVersion() {
		return version;
	}

	public Class getOperatorClass() {
		return operatorClass;
	}

	void setDepth( int depth) {
		this.depth = depth;
	}

	public int getDepth() {
		return depth;
	}

	void addChild( MTBOpNode opNode) {
		children.add( opNode);
	}

	void addDirectChild( MTBOpNode opNode) {
		directlyRegisteredChildren.add( opNode);
	}

	void addData( MTBDataPort data) {
		includedData.add( data);
	}

	public MTBPort getInputPort( int i) {
		return inputPorts[i];
	}

	public MTBOutputPort getOutputPort( int i) {
		return outputPorts[i];
	}

	/** Set the origin of the input port with index i.
      * This sets the origin of the accociated input port as well
      * as its canoncial classname. The explanation is usually to
      * be copied from the argument descriptor of this input port.
      *
      * @param	i index of the input port
      * @param	data MTBData object to assiciate with the port
      * @param	explanation explnatory string of port
     **/
	void setInOrigin( int i, MTBData data, String explanation) {
		if ( data != null ) {
			inputPorts[i].setOrigin( data.getDataport());
			if ( data.getClass().getPackage() == null )
				inputPorts[i].setClassname( data.getClass().getName());
			else
				inputPorts[i].setClassname( data.getClass().getPackage()+"."+data.getClass().getName());
		}
		inputPorts[i].setExplanation( explanation);

	}

	/** Set the origin of the output port with index i.
      * This sets the origin of the accociated output port as well
      * as its canoncial classname. The explanation is usually to
      * be copied from the argument descriptor of this output port.
      *
      * @param	i index of the input port
      * @param	data MTBData object to assiciate with the port
      * @param	explanation explnatory string of port
     **/
	void setOutOrigin( int i, MTBData data, String explanation) {
		if ( data != null ) {
			outputPorts[i].setOrigin( data.getDataport());
			if ( data.getClass().getPackage() == null ) {
				outputPorts[i].setClassname( data.getClass().getName());
			} else {
				outputPorts[i].setClassname( data.getClass().getPackage()+"."+data.getClass().getName());
			}
		}
		outputPorts[i].setExplanation( explanation);
	}

	public void setParent( MTBOpNode parent) {
		this.parent = parent;
	}

	public MTBOpNode getParent() {
		return parent;
	}

	protected Vector<MTBOpNode> getChildren() {
		return children;
	}

	protected Vector<MTBOpNode> getDirectlyRegisteredChildern() {
		return directlyRegisteredChildren;
	}

	protected Vector<MTBDataPort> getIncludedData() {
		return includedData;
	}

	protected MTBInputPort[] getInputPorts() {
		return inputPorts;
	}

	protected MTBOutputPort[] getOutputPorts() {
		return outputPorts;
	}

	public Enumeration<String> getParameterKeys() {
		return this.parameterHash.keys();
	}

	public String getParameter( String key) {
		return this.parameterHash.get( key);
	}

	public XmlObject getParameterHashAsXml() {
		return this.parameterHashAsXml;
	}

	public void print() {
		System.out.println( "MTBOpNode ( with " + inputPorts.length +
					"/" + outputPorts.length + " ports for operator " + name +
					" at depth " + depth);
		System.out.println( ">>>>>   children: ");
		Iterator<MTBOpNode> oItr = children.iterator();
		while ( oItr.hasNext() ) {
			oItr.next().getName();
		}
		System.out.println( "<<<<<   children: ");
		System.out.println( ">>>>>   direct children: ");
		oItr = directlyRegisteredChildren.iterator();
		while ( oItr.hasNext() ) {
			oItr.next().getName();
		}
		System.out.println( "<<<<<   direct children: ");
	}

}
