/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.operator;

import de.unihalle.informatik.MiToBo.exceptions.*;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException.OperatorExceptionType;

import com.thoughtworks.xstream.*;
import com.thoughtworks.xstream.io.xml.DomDriver;

import org.apache.xmlbeans.*;

import java.io.*;
import java.util.*;

/** This is the abstract super class for all MITOBO operators, defining the interface of 
  * the operator and automatically handling the logging of the processing history.
  * <p>
  * Each derived class is expected to override the method <code>collectArguments</code> to 
  * define the descriptors for all parameters, inputs, outputs,
  * and supplemental values.
  * These constitute the interface of the operator. Values of parameters and inputs
  * have to be set prior to invoking the operator via runOp() and the resulting outputs may be retrieved
  * from the operator after return from runOp().
  * In addition further convenvience constructors may be available which additionally
  * set parameters, inputs, and/or supplementals.
  * In addition to parameters, inputs, and outputs an operator may use
  * supplemental (values) e.g. to pass variables to control output or debugging information
  * as well as return intermediate results. The outputs of the operator are expected to
  * be independent of the value of these supplemental values which are not stored in the processing history.
  * <p>
  * The method operate() implements the processing of this operator. All information
  * passed into and back from the operator are passed via the member variables of the operator
  * which may be set and retreived with the appropriate set and get methods.
  * This method should be no means be used to invoke processing of the operator directly,
  * rather  runOp() is to be used.
  * <p> 
  * The final method runOp() is called by the user of an operator to invoke processing.
  * runOp first sets all required but unset parameters and inputs to their default values
  * and subsequently check the validity of parameters and inputs.
  * <p>
  * Refer to the documentation for more details and examples.
  */

abstract public class MTBOperator {
	// realy only debugging
	private boolean debug = false;

	/** A stack of opNodes for the opNodes associated with each (potentially
	 *  recursive calls to this instance of the operator.
	 *  Whenever runOp() is called, an opNode is created which is associated with
	 *  this call. It is pushed onto the stack and popped at the end of ruOp().
	 */
	Stack<MTBOpNode> currentOpNodeStack;

	public final String name;
	public final String version; // svn version control

	// descriptors collected from this operator and all its super classes
	private Hashtable<String,MTBOpArgumentDescriptor> parameterDescriptorsAll;
	private Hashtable<String,MTBOpArgumentDescriptor> inputDescriptorsAll;
	private Hashtable<String,MTBOpArgumentDescriptor> outputDescriptorsAll;
	private Hashtable<String,MTBOpArgumentDescriptor> supplementalDescriptorsAll;

	// hashes for the name,value pairs for all types of values
	private Hashtable<String, Object> parameterHash;
	private Hashtable<String, Object> inputHash;
	private Hashtable<String, Object> outputHash;
	private Hashtable<String, Object> supplementalHash;

	/** This constructor initializes an operator.
	 *  Name and version are retrieved and hashes initialized.
	 */
	protected MTBOperator()  throws MTBOperatorException {
		this.name = this.getClass().getSimpleName();
		this.version = VersionInfo.getCvsTag();

		this.currentOpNodeStack = new Stack<MTBOpNode>(); 

		this.parameterHash = new Hashtable<String, Object>();
		this.inputHash = new Hashtable<String, Object>();
		this.outputHash = new Hashtable<String, Object>();
		this.supplementalHash = new Hashtable<String, Object>();

		this.parameterDescriptorsAll = new Hashtable<String,MTBOpArgumentDescriptor>();
		this.inputDescriptorsAll = new Hashtable<String,MTBOpArgumentDescriptor>();
		this.outputDescriptorsAll = new Hashtable<String,MTBOpArgumentDescriptor>();
		this.supplementalDescriptorsAll = new Hashtable<String,MTBOpArgumentDescriptor>();

		collectArguments();
	}

	/** Get the name of this operator 
	 * @return name of the operator
	 */
	public final String getName() {
		return name;
	}

	/** Get the version in SVN of this operator 
	 * @return version of the operator
	 */
	public final String getVersion() {
		return version;
	}

	// -------------------------------------------------------
	// access functions to parameters

	/** Get the number of parameters
	 * @return number of parameters
	 */
	public final int getNumParameters() {
		return parameterDescriptorsAll.size();
	}

	/** Get the names of parameters
	 * @return array of parameter names
	 */
	public final String[] getParameterNames() {
		String[] keyArray = new String[getNumParameters()];
		Enumeration<String> keys = parameterDescriptorsAll.keys();
		int i=0;
		while ( keys.hasMoreElements() ) {
			keyArray[i] = keys.nextElement();
			i++;
		}
		return keyArray;
	}

	/** Get the parameter descriptor for given name.
	 * @param	name	Name of the parameter to get the new value for
	 * @return	descriptor value 
	 */
	public final MTBOpArgumentDescriptor getParameterDescriptor( String name) 
		throws MTBOperatorException {
		if ( ! parameterDescriptorsAll.containsKey( name) )
			throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		return parameterDescriptorsAll.get( name);
	}

	/** Get the value of a parameter specified by name. 
	 * If a parameter is of type Boolean and is unset, than False is returned,
     * for all other types null is return for unset parameters.
	 *
	 * @param	name	Name of the parameter to get the new value for
	 * @return	If the parameter is set return its value. If the parameter is unset return False for parameter of type Boolean,
	 *				null otherwise.
	 */
	public final Object getParameter(String name) 
		throws MTBOperatorException {
		if ( ! parameterDescriptorsAll.containsKey( name)  )
			throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		if ( parameterHash.get( name) == null ) {
			if ( getParameterDescriptor( name).myclass == java.lang.Boolean.class )
				return Boolean.FALSE;
			else
				return null;
		} else {
			return parameterHash.get( name);
		}
	}

	/** Set the value of a parameter specified by name. 
	 * Type checking on the type of the value is performed.
     * Check is successful if value is an instance of a sub class of the specified parameter or
     * the standard widening for primitive data types apply to the numeric wrapper class.
	 *
	 * @param	name	Name of the parameter to set a new value for
	 * @param	value	new value
	 */
	public final void setParameter(String name, Object value) 
		throws MTBOperatorException {
		Object checkedValue;
		if ( value != null ) {
			if ( ! parameterDescriptorsAll.containsKey( name)  )
				throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
			 
			if ( (checkedValue = checkParamClass( value, parameterDescriptorsAll.get( name).myclass)) == null ) 
				throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.INVALID_CLASS,name);

			parameterHash.put( name, checkedValue);
		}
	}

	/** Returns a clone of the parameter hash
	 *
	 * @return clone of the parameter hash
	 */
	@SuppressWarnings("unchecked")
	public final Hashtable<String,Object> cloneParameters() {
		return (Hashtable<String,Object>)(this.parameterHash.clone());
	}

	// -------------------------------------------------------
	// access functions to inputs

	/** Get the number of inputs
	 * @return number of inputs
	 */
	public final int getNumInputs() {
		return inputDescriptorsAll.size();
	}

	/** Get the names of inputs
	 * @return array of input names
	 */
	public final String[] getInputNames() {
		String[] keyArray = new String[getNumInputs()];
		Enumeration<String> keys = inputDescriptorsAll.keys();
		int i=0;
		while ( keys.hasMoreElements() ) {
			keyArray[i] = keys.nextElement();
			i++;
		}
		return keyArray;
	}

	/** Get the input descriptor for given name.
	 * @param	name	Name of the input to get the new value for
	 * @return	descriptor value 
	 */
	public final MTBOpArgumentDescriptor getInputDescriptor( String name) 
		throws MTBOperatorException {
		if ( ! inputDescriptorsAll.containsKey( name) )
			throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		return inputDescriptorsAll.get( name);
	}

	/** Get the value of an input specified by name. 
	 *
	 * @param	name	Name of the input to get the new value for
	 * @return	input value or null unset
	 */
	public final MTBData getInput(String name) 
		throws MTBOperatorException {
		if ( ! inputDescriptorsAll.containsKey( name)  )
			throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		return (MTBData)(inputHash.get( name));
	}

	/** Set the value of an input specified by name. 
	 * Type checking on the type of the value is performed.
     * Check is successful if value is an instance of the input specified.
	 *
	 * @param	name	Name of the input to set a new value for
	 * @param	value	new value
	 */
	public final void setInput(String name, Object value) 
		throws MTBOperatorException {
		Object checkedValue;
		if ( value != null ) {
			if ( ! inputDescriptorsAll.containsKey( name)  )
				throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
			 
			if ( (checkedValue = checkParamClass( value, inputDescriptorsAll.get( name).myclass)) == null ) 
				throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.INVALID_CLASS,name);

			inputHash.put( name, checkedValue);
		}
	}

	// -------------------------------------------------------
	// access functions to outputs

	/** Get the number of outputs
	 * @return number of outputparameters
	 */
	public final int getNumOutputs() {
		return outputDescriptorsAll.size();
	}

	/** Get the names of outputs
	 * @return array of output names
	 */
	public final String[] getOutputNames() {
		String[] keyArray = new String[getNumOutputs()];
		Enumeration<String> keys = outputDescriptorsAll.keys();
		int i=0;
		while ( keys.hasMoreElements() ) {
			keyArray[i] = keys.nextElement();
			i++;
		}
		return keyArray;
	}

	/** Get the output descriptor for given name.
	 * @param	name	Name of the output to get the new value for
	 * @return	descriptor value 
	 */
	public final MTBOpArgumentDescriptor getOutputDescriptor( String name) 
		throws MTBOperatorException {
		if ( ! outputDescriptorsAll.containsKey( name) )
			throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		return outputDescriptorsAll.get( name);
	}

	/** Get the value of a output specified by name. 
	 *
	 * @param	name	Name of the output to get the new value for
	 * @return	output descriptor or null if unset
	 */
	public final MTBData getOutput(String name) 
		throws MTBOperatorException {
		if ( ! outputDescriptorsAll.containsKey( name)  )
			throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		return (MTBData)(outputHash.get( name));
	}

	/** Set the value of an output specified by name. 
	 * Type checking on the type of the value is performed.
     * Check is successful if value is an instance of the output specified.
	 *
	 * @param	name	Name of the output to set a new value for
	 * @param	value	new value
	 */
	public final void setOutput(String name, Object value) 
		throws MTBOperatorException {
		Object checkedValue;
		if ( value != null ) {
			if ( ! outputDescriptorsAll.containsKey( name)  )
				throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
			 
			if ( (checkedValue = checkParamClass( value, outputDescriptorsAll.get( name).myclass)) == null ) 
				throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.INVALID_CLASS,name);

			outputHash.put( name, checkedValue);
		}
	}

	// -------------------------------------------------------
	// access functions to supplemental values

	/** Get the number of supplementals
	 * @return number of supplementals
	 */
	public final int getNumSupplementals() {
		return supplementalDescriptorsAll.size();
	}

	/** Get the names of supplementals
	 * @return array of supplemental names
	 */
	public final String[] getSupplementalNames() {
		String[] keyArray = new String[getNumSupplementals()];
		Enumeration<String> keys = supplementalDescriptorsAll.keys();
		int i=0;
		while ( keys.hasMoreElements() ) {
			keyArray[i] = keys.nextElement();
			i++;
		}
		return keyArray;
	}

	/** Get the supplemental descriptor for given name.
	 * @param	name	Name of the supplemental to get the new value for
	 * @return	descriptor value 
	 */
	public final MTBOpArgumentDescriptor getSupplementalDescriptor( String name) 
		throws MTBOperatorException {
		if ( ! supplementalDescriptorsAll.containsKey( name) )
			throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		return supplementalDescriptorsAll.get( name);
	}

	/** Get the value of a supplemental specified by name. 
	 *  If a parameter is of type Boolean and is unset, than False is returned,
	 *  for all other types null is return for unset parameters.
	 *
	 * @param	name	Name of the supplemental to get the new value for
	 * @return	If the supplemental is set return its value. If the supplemental is unset return False for supplemental of type Boolean,
	 *			null otherwise.
	 */
	public final Object getSupplemental(String name) 
		throws MTBOperatorException {
		if ( ! supplementalDescriptorsAll.containsKey( name)  )
			throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		if ( supplementalHash.get( name) == null ) {
			if ( getSupplementalDescriptor( name).myclass == java.lang.Boolean.class )
				return Boolean.FALSE;
			else
				return null;
		} else {
			return supplementalHash.get( name);
		}
	}

	/** Set the value of a supplemental specified by name. 
	 * Type checking on the type of the value is performed.
     * Check is successful if value is an instance of a sub class of the specified supplemental or
     * the standard widening for primitive data types apply to the numeric wrapper class.
	 *
	 * @param	name	Name of the supplemental to set a new value for
	 * @param	value	new value
	 * @return	true if supplemental could be set,
	 */
	public final void setSupplemental(String name, Object value) 
		throws MTBOperatorException {
		Object checkedValue;
		if ( value != null ) {
			if ( ! supplementalDescriptorsAll.containsKey( name)  )
				throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
			 
			if ( (checkedValue = checkParamClass( value, supplementalDescriptorsAll.get( name).myclass)) == null ) 
				throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.INVALID_CLASS,name);

			supplementalHash.put( name, checkedValue);
		}
	}

	//==========================

	/** This is the only legal method to invoke the operator and handles everything 
	 * necessary to protocol the processing history.
	 * Prior to calling <code>runOp</code>, input data and parameters have to be set.
	 * When called <code>runOp</code> sets all required but unset parameters and inputs to 
	 * their default values as define for this operator.
	 * Note, if an argument is null it is considered as unset.
	 * Subsequently the validity of parameters and inputs are checked.
	 * Validity requires for all operators, that all required parameters and inputs
	 * have non null values.
	 * In addition the implementation of an operator may impose further constrains
	 * defined by the method <code>validateCustom</code>
	 * which e.g. may restrict the interval of numerical parameters.
	 * After successful validation the <code>operate</code> method is called to
	 * do the actuall processing of this oeprator.
	 * Upon return from <code>runOp</code>, resulting output and supplemental data
	 * may be retrieved from the operator object.
	 * <br>
	 * The argument <code>callingOperator</code> should be a reference to the operator
	 * which calls this operator as a nested operation.
	 * If <code>callingOperator</code> is null the history mechanism will find the
	 * operator called by its <code>runOp</code> only if one of its input or output data
	 * is referenced by the enclosing operator(s). Otherwise the operator called
	 * with a null <code>callingOperator</code> will not be included into the processing
	 * history.
	 *
	 * @param   callingOperator    The operator object which is calling the current operator (may be null)
	 */

	public final void runOp( MTBOperator callingOperator) throws MTBOperatorException,MTBProcessingDAGException {

        if ( getVerbose() ) {
            System.out.println( "");
            System.out.println( "Start of runOp");
            this.print();
        }
 
		// set defaults and validate
		setDefaults();
		validate();

		// create an opNode for this operation in the DAG
		MTBOpNode opNode = new MTBOpNode( this);
        if ( callingOperator != null ) {
			if ( getVerbose() ) {
				System.out.println( "MTBOperator::runOp adding direct registered child " + this + "(" + opNode + ")\n" +
							"                   to operator " + callingOperator + "(" + callingOperator.getCurrentOpNode() + ")");
			}

            callingOperator.getCurrentOpNode().addDirectChild( opNode);
        }

		// push AFTER we added the child to handle recursive calls to an operator instance
		currentOpNodeStack.push( opNode);


		// handle inputs
		MTBPort origPorts[] = new MTBPort[this.getNumInputs()];
		String[] inputNames = getInputNames();

		// first we store the original dataport of inputs
		for ( int i = 0; i < this.getNumInputs() ; i++ ) {
			MTBData input = this.getInput( inputNames[i]);

			if ( input != null ) {
				// connect each input port of this opNode to the port
				// of the input data the operator was supplied with
				opNode.setInOrigin( i, input, getInputDescriptor( inputNames[i]).explanation);

				origPorts[i] = input.getDataport(); 
			} else {
				opNode.setInOrigin( i, null, getInputDescriptor( inputNames[i]).explanation);
				origPorts[i] = null;
			}
		}

		// and now we can set the dataports of the data to the input ports
		// (we have to do this in two for-loops to handle multiple inputs with
		// indentical MTBData objects correctly)
		for ( int i = 0; i < this.getNumInputs() ; i++ ) {
			MTBData input = this.getInput( inputNames[i]);

			if ( input != null ) {
				// set the output port of the input to the input port of this opNode
				input.setDataport( opNode.getInputPort( i));
			}
		}

        if ( getVerbose() != null && getVerbose() ) {
            System.out.println( "");
            System.out.println( "  before operate");
            for ( int i = 0; i < this.getNumInputs() ; i++ ) {
                System.out.println( "    port of " + i + "-th input " + opNode.getInputPort( i) +
                    " with origin " + opNode.getInputPort( i).getOrigin() );
            }
        }
		// do the work
		operate();

		// handle output ports
		String[] outputNames = getOutputNames();
		for ( int i = 0; i < this.getNumOutputs() ; i++ ) {
			MTBData output = this.getOutput( outputNames[i]);

			if ( output != null ) {
				// connect each output port of this opNode to the output port
				// its results (i.e. oututs) originate from
				opNode.setOutOrigin( i, output, getOutputDescriptor( outputNames[i]).explanation);

				// store properties of output at the output port
				Enumeration<String> keys = output.getPropertyKeys();
				while ( keys.hasMoreElements() ) {
					String key = keys.nextElement();
					opNode.getOutputPort( i).setProperty( key, output.getProperty( key));
				}
			} else {
				opNode.setOutOrigin( i, null, getOutputDescriptor( outputNames[i]).explanation);
			}
		}
		// restore port of inputs
		for ( int i = 0; i < this.getNumInputs() ; i++ ) {
			if ( origPorts[i] != null ) 
				this.getInput( inputNames[i]).setDataport( origPorts[i]);
		}

		// handle outputs
		for ( int i = 0; i < this.getNumOutputs() ; i++ ) {
			MTBData output = this.getOutput( outputNames[i]);

			if ( output != null ) {
				// set the output port of the output (i.e. result) of the operator
				// to the output port of the opNode
				output.setDataport( opNode.getOutputPort( i)); 
			}
		}

		if ( getVerbose() != null && getVerbose() ) {
			System.out.println( "  End of runOp");
			this.print();
			for ( int i = 0; i < this.getNumOutputs() ; i++ ) {
				System.out.println( "    port for " + i + "-th result " + opNode.getOutputPort( i) +
					" with origin " + opNode.getOutputPort( i).getOrigin() );
				if ( opNode.getOutputPort( i).getOrigin() != null )
					opNode.getOutputPort( i).getOrigin().print();
			}
		}

		// pop opNode from stack
		currentOpNodeStack.pop();
	}

	// ===================================================================

	/** Validates the parameters and inputs of this operator.
	 *  The generic part of this validation checks that each required parameter
	 *  and input is set to a non-null value or supplies a default value.
	 *  The custom part of validation is operator depended.
	 *  An MTBOperatorException is thrown if either validation fails.
 	 */
	public final void validate() throws MTBOperatorException {
		validateGeneric();
		validateCustom();
	}

	/** Generic validation of the parameters and inputs of this operator.
	 *  Validation succeeds if each required parameter
	 *  and input is set to a non-null value or supplies a default value.
	 *  An MTBOperatorException is thrown if validation fails.
 	 */
	public final void validateGeneric() throws MTBOperatorException {
		
		String[] inputNames = getInputNames();
		for ( int i = 0; i < this.getNumInputs() ; i++ ) {
			if ( inputDescriptorsAll.get( inputNames[i]).required &&
					getInput( inputNames[i]) == null &&
					inputDescriptorsAll.get( inputNames[i]).defaultValue  == null ) {
				throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED, 
												"MTBOperator.validateGeneric(): Problems with input \"" + inputNames[i] + "\"");  
			}            
		}

		String[] parameterNames = getParameterNames();
		for ( int i = 0; i < this.getNumParameters() ; i++ ) {
			if ( parameterDescriptorsAll.get( parameterNames[i]).required &&
					getParameter( parameterNames[i]) == null &&
					parameterDescriptorsAll.get( parameterNames[i]).defaultValue  == null ) {
				throw new MTBOperatorException(OperatorExceptionType.VALIDATION_FAILED,
											"MTBOperator.validateGeneric(): Problems with parameter \"" + parameterNames[i] + "\"");  
			}
		} 
	}

	/**  The validation of parameters and inputs.
 	 */
	public void validateCustom() throws MTBOperatorException {
		return;
	}

	/** Set default values for all required inputs and parameters which are
	 *  unset or set to null.
	 *  Note, this guarantees not all required values to set to non null as
	 *  defaults may be null also.
 	 */
	public final void setDefaults() {
		String[] inputNames = getInputNames();
		try {
        	for ( int i = 0; i < this.getNumInputs() ; i++ ) {
            	if ( inputDescriptorsAll.get( inputNames[i]).required &&
                 	getInput( inputNames[i]) == null )
					setInput( inputNames[i], inputDescriptorsAll.get( inputNames[i]).defaultValue);
			}
	
			String[] parameterNames = getParameterNames();
        	for ( int i = 0; i < this.getNumParameters() ; i++ ) {
				// NOTE: we may not use getParameter( parameterNames[i]) here, as getParameter
				//       returns false for an unset Boolean
            	if ( parameterDescriptorsAll.get( parameterNames[i]).required &&
					parameterHash.get( parameterNames[i]) == null ) {
				 	setParameter( parameterNames[i], parameterDescriptorsAll.get( parameterNames[i]).defaultValue);
				}
			}
		} catch (MTBOperatorException e) {
			e.printStackTrace();
		}
	}

	// ===================================================================
	/** Set the verbose state of this opertor.
	 * @param	verbose 	New verbose state
	 */
	public void setVerbose( Boolean verbose) throws MTBOperatorException {
		this.setSupplemental( "verbose", verbose);
	}

	/** Return the verbose state of this opertor.
	 * @return  Current verbose state
	 */
	public Boolean getVerbose() throws MTBOperatorException {
		return (Boolean)(this.getSupplemental( "verbose"));
	}

	/** Return the opNode associated with the current call to the runOp-method
	 *  of the operator.
	 * @return  Current opNode
	 */
	public MTBOpNode getCurrentOpNode() {
		return currentOpNodeStack.peek();
	}

	// ===================================================================

	/** This method does the actual work and needs to be implemented by every subclass.
	 */

	protected abstract void operate() throws MTBOperatorException,MTBProcessingDAGException;

	// ===================================================================

	/** Write the parameter values to an xml file
	 *
	 * @param	filename	filename of the xml file to write to
	 */

	public void writeParametersToXml( String filename ) {
		try {
			PrintStream out = new PrintStream( filename);

			XStream xstream = new XStream(new DomDriver());
			//String xml = xstream.toXML( parameterHash);
			String xml = xstream.toXML( new MTBParameterWrapper(this));
	
			out.println( xml);
		} catch ( Exception e) {
			System.err.println(e);
			e.printStackTrace();
		}

	}

	/** parse the parameter values to an XmlObject
	 *
	 */

	public XmlObject parametersToXmlObject() 
		throws XmlException {

		XStream xstream = new XStream(new DomDriver());
		String xml = xstream.toXML( new MTBParameterWrapper(this));
		XmlObject res = XmlObject.Factory.parse( xml);
	
		return res;
	}

	/** Set the parameter values as read from an Xml file.
	 * The class and package name found in the Xml are compared to
	 * the operator to set the parameters to and are required to match.
	 *
	 * @param	filename	filename of the xml file to read from
	 * @return	version string found in Xml file of null if Xml file could not be properly read
	 *				or class or package name check failed
	 */

	public String setParametersFromXml( String filename ) {
		return setParametersFromXml( filename, getClass().getSimpleName(), getClass().getPackage().getName());
	}
		
	/** Set the parameter values as read from an Xml file.
	 * The class and package name found in the Xml are compared to
	 * the <code>className</code> and <code>packageName</code> respectively.
	 * If one of these names is <code>null</code> any name in the Xml file
	 * is accepted.
	 *
	 * @param	filename	filename of the xml file to read from
	 * @param	className	class name expected in the Xml file or null
	 * @param	packageName	package name expected in the Xml file or null
	 * @return	version string found in Xml file of null if Xml file could not be properly read
	 *				or class or package name check failed
	 */

	public String setParametersFromXml( String filename, String className, String packageName ) {
		
		try {
			FileInputStream in = new FileInputStream( filename);

			XStream xstream = new XStream(new DomDriver());

			MTBParameterWrapper pw = (MTBParameterWrapper)(xstream.fromXML( in));

			if ( className != null && ! className.equals( pw.getClassName()) ) {
				System.err.println( "MTBOperator::setParametersFromXml class name does not match, got " + 
					pw.getClassName() + " but expected " + className);
				return null;
			}

			if ( packageName != null && ! packageName.equals( pw.getPackageName()) ) {
				System.err.println( "MTBOperator::setParametersFromXml package name does not match, got " + 
					pw.getPackageName() + " but expected " + packageName);
				return null;
			}

			parameterHash = (Hashtable<String, Object>)(pw.getParameteres());
			return pw.getVersion();
			
		} catch ( FileNotFoundException e) {
			System.err.println( "MTBOperator::setParametersFromXml cannot open " + filename);
			return null;
		}

	}

	// ===================================================================

	/** Print some information of the current state this operator to System.out.
	 */

	public void print() {
		print( System.out);
	}
		
	/** Print some information the current state of this operator to <code>outfile</code>.
	 * @param	outfile	Stream to print on
	 */

	public void print( PrintStream outfile) {
		try {
			outfile.println( "  MTBOperator: " + name + " -> " + this); 

			String[] parameterNames = getParameterNames();
			for ( int n=0 ; n < getNumParameters() ; n++ )
				outfile.println( "    Parameter " + parameterDescriptorsAll.get( parameterNames[n]).name + " = " + getParameter( parameterNames[n]));

			String[] inputNames = getInputNames();
			for ( int n=0 ; n < getNumInputs() ; n++ )
				outfile.println( "    Input " + inputDescriptorsAll.get( inputNames[n]).name + " = " + getInput( inputNames[n]));

			String[] outputNames = getOutputNames();
			for ( int n=0 ; n < getNumOutputs() ; n++ )
				outfile.println( "    Output " + outputDescriptorsAll.get( outputNames[n]).name + " = " + getOutput( outputNames[n]));

			String[] supplementalNames = getSupplementalNames();
			for ( int n=0 ; n < getNumSupplementals() ; n++ )
				outfile.println( "    Supplemental " + supplementalDescriptorsAll.get( supplementalNames[n]).name + " = " + getSupplemental( supplementalNames[n]));

		} catch (MTBOperatorException e) {
			e.printStackTrace();
		}
	}

	/** Print information of the interface of this operator to System.out.
	 */

	public void printInterface() {
		printInterface( System.out);
	}
		
	/** Print information of the interface of this operator to <code>outfile</code>.
	 * @param	outfile	Stream to print on
	 */

	public void printInterface( PrintStream outfile) {
		outfile.println( "Interface of MTBOperator: " + name);

		String[] parameterNames = getParameterNames();
		for ( int n=0 ; n < getNumParameters() ; n++ )
			outfile.println( "    Param <" + parameterDescriptorsAll.get( parameterNames[n]).name +  "> " +
							(parameterDescriptorsAll.get( parameterNames[n]).required ? "" : " not") +
							" required, type: " + parameterDescriptorsAll.get( parameterNames[n]).myclass.getSimpleName());

		String[] inputNames = getInputNames();
		for ( int n=0 ; n < getNumInputs() ; n++ )
			outfile.println( "    Input <" + inputDescriptorsAll.get( inputNames[n]).name + "> " +
							(inputDescriptorsAll.get( inputNames[n]).required ? "" : " not") +
							" required, type: " + inputDescriptorsAll.get( inputNames[n]).myclass.getSimpleName());

		String[] outputNames = getOutputNames();
		for ( int n=0 ; n < getNumOutputs() ; n++ )
			outfile.println( "    Output <" + outputDescriptorsAll.get( outputNames[n]).name + "> " +
							(outputDescriptorsAll.get( outputNames[n]).required ? "" : " not") +
							" required, type: " + outputDescriptorsAll.get( outputNames[n]).myclass.getSimpleName());

		String[] supplementalNames = getSupplementalNames();
		for ( int n=0 ; n < getNumSupplementals() ; n++ )
			outfile.println( "    Supplemental <" + supplementalDescriptorsAll.get( supplementalNames[n]).name + "> " +
							(supplementalDescriptorsAll.get( supplementalNames[n]).required ? "" : " not") +
							" required, type: " + supplementalDescriptorsAll.get( supplementalNames[n]).myclass.getSimpleName());
	}

	// ================== initialization

	/** This method is essential to initialize the descriptors of all arguments.
     * Each sub class of <code>MTBOperator</code>
     * is required to override the method <code>collectArguments</code>, otherwise
     * the inheritance of arguments wil lbe broken.
     * As a minimum, <code>collectArguments</code> needs to invoke
     * the <code>collectArguments</code> of its direct super class.
     * Subsequently further arguments may be added invoking the 
     * <code>addDescriptors</code> method.
     * The base class of all operators, <code>MTBOperator</code>, defines as its only argument the
     * supplemental verbose flag which is my default set to false.
     * See the documentation for further details and examples.
	 */
	protected void collectArguments()
			throws MTBOperatorException {
		if ( debug )
			System.out.println( "MTBOperator::collectArguments");
		MTBOpArgumentDescriptor[] parameterDescriptors = null;
		MTBOpArgumentDescriptor[] inputDescriptors = null;
		MTBOpArgumentDescriptor[] outputDescriptors = null;
		MTBOpArgumentDescriptor[] supplementalDescriptors = new MTBOpArgumentDescriptor[] {
			new MTBOpArgumentDescriptor( "verbose", Boolean.class, "Verbose flag", false, new Boolean(false))
		};

		addDescriptors( parameterDescriptors, inputDescriptors,
						 outputDescriptors, supplementalDescriptors);

		if ( debug ) 
			System.out.println("init instance of MTBOperator");

	}

    /** Add descriptors for parameters, inputs, outputs, and supplemental values
     * to the descriptors already defined for this operator.
     * This method should be invoked once for each object derived
     * from <code>MTBOperator</code> by its overridden method <code>collectArguments</code>.
     */
	protected final void addDescriptors( MTBOpArgumentDescriptor[] parameterDescriptors, 
									MTBOpArgumentDescriptor[] inputDescriptors,
									MTBOpArgumentDescriptor[] outputDescriptors, 
									MTBOpArgumentDescriptor[] supplementalDescriptors) 
	throws MTBOperatorException {
		if ( debug ) 
			System.out.println( "addDescriptors");

		if ( parameterDescriptors != null ) {
			for ( int i = 0 ; i < parameterDescriptors.length ; i++ ) {
				if ( debug )
					System.out.println( "   initDescriptors:: adding <" + parameterDescriptors[i].name + ">");
				parameterDescriptorsAll.put( parameterDescriptors[i].name, parameterDescriptors[i]);
			}
		}

		if ( inputDescriptors != null ) {
			for ( int i = 0 ; i < inputDescriptors.length ; i++ ) {
				if ( debug )
					System.out.println( "   initDescriptors:: adding <" + inputDescriptors[i].name + ">");

				if ( ! MTBData.class.isAssignableFrom( inputDescriptors[i].myclass) ) {
					throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.MTB_DATA_REQUIRED,
						"MTBOperator::addDescriptors no MTBData class for <" + inputDescriptors[i].name +
                                    ">: " + inputDescriptors[i].myclass);
				}

				inputDescriptorsAll.put( inputDescriptors[i].name, inputDescriptors[i]);
			}
		}

		if ( outputDescriptors != null ) {
			for ( int i = 0 ; i < outputDescriptors.length ; i++ ) {
				if ( debug )
					System.out.println( "   initDescriptors:: adding <" + outputDescriptors[i].name + ">");

				if ( ! MTBData.class.isAssignableFrom( outputDescriptors[i].myclass) ) {
					throw new MTBOperatorException(MTBOperatorException.OperatorExceptionType.MTB_DATA_REQUIRED,
						"MTBOperator::initDescriptors no MTBData class for <" + outputDescriptors[i].name +
                                    ">: " + outputDescriptors[i].myclass);
				}

				outputDescriptorsAll.put( outputDescriptors[i].name, outputDescriptors[i]);
			}
		}

		if ( supplementalDescriptors != null ) {
			for ( int i = 0 ; i < supplementalDescriptors.length ; i++ ) {
				if ( debug )
					System.out.println( "   initDescriptors:: adding <" + supplementalDescriptors[i].name + ">");
				supplementalDescriptorsAll.put( supplementalDescriptors[i].name, supplementalDescriptors[i]);
			}
		}

	}

	// ================== helper function

	///** Lookup the name in a hash of strings,descriptor pairs
      //*/
	//private boolean lookupName( String name, Vector<MTBOpArgumentDescriptor> descriptors) {
		//if ( descriptors != null ) {
			//if ( descriptors.get( name) != null )
				//return true;
		//} else {
			//return false;
		//}
	//}

	/** Check if class of object o is compatible with myclass.
	 * This is the case if value is a sub class of myclass or 
	 * the standard widening for primitive data types apply to the numeric wrapper class.
	 * If the object is a sub class, return this object. If widening applies
	 * return a new object of myclass initialized to the value of <code>o</code>
	 * If compatible check fails return null.
	 *
	 * @param	o	object to check class compatibility
	 * @param	myclass	class to check against
	 * @return	if o is instance of a sub class of myclass, return o, if widening applies
	 *		a new instance of class myclass with value taken from o, null otherwise
	 */
	private Object checkParamClass( Object o, Class myclass) {
		if ( myclass == java.lang.Double.class ) {
			if ( (o.getClass() == java.lang.Double.class) )
				return o;
			else if (o.getClass() == java.lang.Float.class)  
				return ((Float)o).doubleValue();
			else if (o.getClass() == java.lang.Long.class)  
				return ((Long)o).doubleValue();
			else if  (o.getClass() == java.lang.Integer.class)  
				return ((Integer)o).doubleValue();
			else if (o.getClass() == java.lang.Short.class)  
				return ((Short)o).doubleValue();
			else if (o.getClass() == java.lang.Byte.class)  
				return ((Byte)o).doubleValue();
		} else if ( myclass == java.lang.Float.class ) {
			if ( (o.getClass() == java.lang.Float.class)  )
				return o;
			else if ( (o.getClass() == java.lang.Long.class)  )
				return ((Long)o).floatValue();
			else if ( (o.getClass() == java.lang.Integer.class)  )
				return ((Integer)o).floatValue();
			else if ( (o.getClass() == java.lang.Short.class)  )
				return ((Short)o).floatValue();
			else if ( (o.getClass() == java.lang.Byte.class)  )
				return ((Byte)o).floatValue();
		} else if ( myclass == java.lang.Long.class ) {
			if ( (o.getClass() == java.lang.Long.class)  )
				return o;
			else if ( (o.getClass() == java.lang.Integer.class)  )
				return ((Integer)o).longValue();
			else if ( (o.getClass() == java.lang.Short.class)  )
				return ((Short)o).longValue();
			else if ( (o.getClass() == java.lang.Byte.class)  )
				return ((Byte)o).longValue();
		} else if ( myclass == java.lang.Integer.class ) {
			if ( (o.getClass() == java.lang.Integer.class)  )
				return o;
			else if ( (o.getClass() == java.lang.Short.class)  )
				return ((Short)o).intValue();
			else if ( (o.getClass() == java.lang.Byte.class)  )
				return ((Byte)o).intValue();
		} else if ( myclass == java.lang.Short.class ) {
			if ( (o.getClass() == java.lang.Short.class)  )
				return o;
			else if ( (o.getClass() == java.lang.Byte.class) )
				return ((Byte)o).shortValue();
		} else  if ( myclass.isInstance( o) ) {
			return o;
		}
		
		return null;
	}

}

