/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.segmentation.regions.filling;


import java.util.Vector;

//import loci.formats.StatusEvent;
//import loci.formats.StatusListener;
//import loci.formats.StatusReporter;
import de.unihalle.informatik.MiToBo.datatypes.MTBPoint3D;
import de.unihalle.informatik.MiToBo.datatypes.MTBRegion3D;
import de.unihalle.informatik.MiToBo.datatypes.MTBRegion3DSet;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.exceptions.MTBProcessingDAGException;
import de.unihalle.informatik.MiToBo.operator.MTBOpArgumentDescriptor;
import de.unihalle.informatik.MiToBo.operator.MTBOperator;
import de.unihalle.informatik.MiToBo.segmentation.regions.labeling.LabelComponentsSequential3D;

/**
 * Operator to fill holes in 3D connected components
 * @author gress
 *
 */
public class FillHoles3D extends MTBOperator { //implements StatusReporter {
	
  protected void collectArguments() throws MTBOperatorException {
        super.collectArguments();

		MTBOpArgumentDescriptor[] inputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("inImg", MTBImage.class, 
						"Input image", true, null)
		};	
		MTBOpArgumentDescriptor[] outputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("resultImg", MTBImage.class, 
						"Result image", true, null)
		};
	
		MTBOpArgumentDescriptor[] parameterDescriptors = null;
		MTBOpArgumentDescriptor[] supplementalDescriptors = null;
	
		
			addDescriptors( parameterDescriptors, inputDescriptors, 
					outputDescriptors, supplementalDescriptors);
	}
	
	public FillHoles3D() throws MTBOperatorException {
	//	this.m_statusListeners = new Vector<StatusListener>();
	}

	public FillHoles3D(MTBImage image) throws MTBOperatorException {
	//	this.m_statusListeners = new Vector<StatusListener>();
		this.setInputImage(image);
	}
	
	/**
	 * Get reference to the current input image.
	 * 
	 * @return	Input image to work on.
	 * @throws MTBOperatorException 
	 */
	public MTBImage getInputImage() throws MTBOperatorException {
		return (MTBImage)(this.getInput("inImg"));
	}

	/**
	 * Set input image.
	 * 
	 * @param img Input image to work on.
	 * @throws MTBOperatorException 
	 */
	public void setInputImage(MTBImage img) throws MTBOperatorException {
		this.setInput("inImg", img);
	}
	
	/**
	 * Get the resulting filtered image.
	 * Attention, reference might be null.
	 * 
	 * @return	Reference to result image.
	 * @throws MTBOperatorException 
	 */
	public MTBImage getResultImage() throws MTBOperatorException {
		return (MTBImage)(this.getOutput("resultImg"));
	}
	
	/**
	 * Set the result image.
	 * 
	 * @param resultImage	Result image.
	 * @throws MTBOperatorException 
	 */
	public void setResultImage(MTBImage resultImage) throws MTBOperatorException {
		this.setOutput("resultImg", resultImage);
	}
	
	/** vector of installed StatusListeners */
//	private Vector<StatusListener> m_statusListeners;
	
	@Override
	protected void operate() throws MTBOperatorException,MTBProcessingDAGException {
		MTBImage inImg = this.getInputImage();

		LabelComponentsSequential3D lcs = new LabelComponentsSequential3D(inImg, true);
		lcs.runOp(this);
		MTBRegion3DSet regs = lcs.getResultingRegions();
		MTBImage outImg = MTBImage.createMTBImage(inImg.getSizeX(), 
				inImg.getSizeY(), 
				inImg.getSizeZ(), 
				inImg.getSizeT(), 
				inImg.getSizeC(), 
				MTBImageType.MTB_BYTE);
		for (int i = 0; i < regs.size(); i++) {
		//	this.fireStatusEvent(i, regs.size(), "Filling regions of " + inImg.getTitle() + "...");
			this.paintFilledRegion(regs.elementAt(i), outImg);
		}
		//this.fireStatusEvent(regs.size(), regs.size(), "Filling regions of " + inImg.getTitle() + " done");
		this.setResultImage(outImg);
	}

	private void paintFilledRegion(MTBRegion3D reg, MTBImage regImg) throws MTBOperatorException,MTBProcessingDAGException {
		Vector<MTBPoint3D> pts = reg.getPoints();
		MTBPoint3D pt;
		int x, y, z, xmin=-1, xmax=-1, ymin=-1, ymax=-1, zmin=-1, zmax=-1;
		
		for (int i = 0; i < pts.size(); i++) {
			x = (int)pts.get(i).getX();
			y = (int)pts.get(i).getY();
			z = (int)pts.get(i).getZ();
			
			if (i == 0) {
				xmin = x;
				xmax = x;
				ymin = y;
				ymax = y;
				zmin = z;
				zmax = z;
			}
			else {
				if (x < xmin) xmin = x;
				if (x > xmax) xmax = x;
				if (y < ymin) ymin = y;
				if (y > ymax) ymax = y;
				if (z < zmin) zmin = z;
				if (z > zmax) zmax = z;
			}
			
			regImg.putValueInt(x, y, z, 0, 0, 255);		
		}
		
		int wSizeX = xmax-xmin+1;
		int wSizeY = ymax-ymin+1;
		int wSizeZ = zmax-zmin+1;
		
		MTBImage iw = MTBImage.createMTBImage(wSizeX, wSizeY, wSizeZ, 1, 1, MTBImageType.MTB_BYTE);
		
		for (int zw = 0; zw < wSizeZ; zw++) {
			for (int yw = 0; yw < wSizeY; yw++) {
				for (int xw = 0; xw < wSizeX; xw++) {
					iw.putValueInt(xw, yw, zw, 0, 0, 255);
				}
			}
		}
		
		pts = reg.getPoints();						
		
		for (int j = 0; j < pts.size(); j++) {
			pt = pts.get(j);
			iw.putValueInt((int)pt.getX()-xmin, (int)pt.getY()-ymin, (int)pt.getZ()-zmin, 0, 0, 0);
		}
		
		LabelComponentsSequential3D lcs = new LabelComponentsSequential3D(iw, false);
		lcs.runOp(this);
		MTBRegion3DSet regs = lcs.getResultingRegions();
		
		
		boolean bgRegion;
		for (int i = 0; i < regs.size(); i++) {
			
			pts = regs.elementAt(i).getPoints();
			
			bgRegion = false;
			
			for (int j = 0; j < pts.size(); j++) {
				pt = pts.get(j);
				x = (int)pt.getX();
				y = (int)pt.getY();
				z = (int)pt.getZ();
				
				if (x == 0 || x == wSizeX-1 || y == 0 || y == wSizeY-1 || z == 0 || z == wSizeZ-1)
					bgRegion = true;
			}
			
			if (!bgRegion) {
				
				for (int j = 0; j < pts.size(); j++) {
					pt = pts.get(j);
					
					regImg.putValueInt((int)pt.getX()+xmin, 
									   (int)pt.getY()+ymin,
									   (int)pt.getZ()+zmin,
									   0, 0,
									   255);
				}
			}
		}
	}

	// ----- Event triggering
	
//	/**
//	 * Fire the same StatusEvent to all installed StatusListener Objects
//	 * @param progress current progress index
//	 * @param maximum maximum progress index
//	 * @param message status string
//	 */
//	protected void fireStatusEvent(int progress, int maximum, String message) {
//		
//		StatusEvent e = new StatusEvent(progress, maximum, message);
//		 
//		for (int i = 0; i < m_statusListeners.size(); i++) {
//			m_statusListeners.get(i).statusUpdated(e);
//		}
//	}
//	
//	
//	
//	// ----- Function to be implemented for StatusReporter interface
//	
//	/*
//	 * (non-Javadoc)
//	 * @see loci.formats.StatusReporter#addStatusListener(loci.formats.StatusListener)
//	 */
//	public void addStatusListener(StatusListener statuslistener) {	
//		m_statusListeners.add(statuslistener);	
//	}
//
//	/*
//	 * (non-Javadoc)
//	 * @see loci.formats.StatusReporter#getStatusListeners()
//	 */
//	public StatusListener[] getStatusListeners() {
//
//		if (m_statusListeners.size() == 0) {
//			return null;
//		}
//		else {
//			return (StatusListener[])m_statusListeners.toArray();
//		}
//	}
//
//	/*
//	 * (non-Javadoc)
//	 * @see loci.formats.StatusReporter#removeStatusListener(loci.formats.StatusListener)
//	 */
//	public void removeStatusListener(StatusListener statuslistener) {
//		m_statusListeners.remove(statuslistener);
//	}

}
