/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.segmentation.regions.labeling;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Vector;

import de.unihalle.informatik.MiToBo.datatypes.MTBRegion3D;
import de.unihalle.informatik.MiToBo.datatypes.MTBRegion3DSet;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.drawing.DrawRegion3DSet;
import de.unihalle.informatik.MiToBo.drawing.DrawRegion3DSet.DrawType;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.exceptions.MTBProcessingDAGException;
import de.unihalle.informatik.MiToBo.operator.MTBOpArgumentDescriptor;
import de.unihalle.informatik.MiToBo.operator.MTBOperator;

/**
 * 
 * @author gress
 *
 * This class implements a sequential component labeling technique on 
 * binary image volumes. Foreground pixels are assumed to have a value > 0,
 * all pixels with value 0 are thought as background.
 * 
 * The static routine returns a vector of regions.
 *
 */
public class LabelComponentsSequential3D extends MTBOperator {

  protected void collectArguments() throws MTBOperatorException {
        super.collectArguments();

		MTBOpArgumentDescriptor[] inputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("inImg", MTBImage.class, 
						"Input image", true, null)
		};	
		MTBOpArgumentDescriptor[] outputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("resultingRegions", MTBRegion3DSet.class, 
						"Resulting regions", true, null),
				new MTBOpArgumentDescriptor("labelImage", MTBImage.class, 
							"Label image", true, null)
		};
		MTBOpArgumentDescriptor[] parameterDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("diagonalNeighborhood", Boolean.class, 
						"Use diagonal neighborhood", true, new Boolean(true))
		};
		MTBOpArgumentDescriptor[] supplementalDescriptors = null;
	
		
			addDescriptors( parameterDescriptors, inputDescriptors, 
					outputDescriptors, supplementalDescriptors);
	}

	/**
	 * Constructor
	 * @param img input image
	 * @param diagonalNeighbors neighborhood flag. true for 26-neighborhood, false for 6-neighborhood (3D !!)
	 * @throws MTBOperatorException 
	 */
	public LabelComponentsSequential3D(MTBImage img, boolean diagonalNeighbors) 
		throws IllegalArgumentException, MTBOperatorException {
		
		if (img.getSizeT() > 1 || img.getSizeC() > 1) {
			throw new IllegalArgumentException("Input image must have size 1 in t- and c-dimension (true spatial 3D-Stack).");
		}
		this.setInput("inImg", img);
		this.setParameter("diagonalNeighborhood", new Boolean(diagonalNeighbors));
	}

	/**
	 * Get reference to the current input image.
	 * 
	 * @return	Input image to work on.
	 * @throws MTBOperatorException 
	 */
	MTBImage getInputImage() throws MTBOperatorException {
		return (MTBImage)(this.getInput("inImg"));
	}
	
	/**
	 * Get the neighborhood flag.
	 * @return true if 26-neighborhood is used, false if 6-neighborhood is used
	 * @throws MTBOperatorException 
	 */
	boolean getDiagonalNeighborsFlag() throws MTBOperatorException {
		return ((Boolean)(this.getParameter("diagonalNeighborhood"))).booleanValue();
	}
	
	/**
	 * Attach the vector of resulting regions of a certain time and channel to this parameter object.
	 * @param channel
	 * @throws MTBOperatorException 
	 */
	void setResultingRegions(MTBRegion3DSet regions/*, int time, int channel*/) throws MTBOperatorException {
	//	String key = "t" + time + "_c" + channel;
		this.setOutput("resultingRegions", regions);
	}
	
	/**
	 * Add the label image to the parameter object
	 * @param regions
	 * @throws MTBOperatorException 
	 */
	void setLabelImage(MTBImage labelImage) throws MTBOperatorException {
		this.setOutput("labelImage", labelImage);
	}

	/**
	 * Get label image of type MTB_INT
	 * @return
	 * @throws MTBOperatorException 
	 */
	public MTBImage getLabelImage() throws MTBOperatorException {
		return (MTBImage)this.getOutput("labelImage");
	}
	
	/**
	 * Get the vector of resulting regions of a certain time and channel from this parameter object.
	 * @return
	 * @throws MTBOperatorException 
	 */
	public MTBRegion3DSet getResultingRegions() throws MTBOperatorException {
		return (MTBRegion3DSet)this.getOutput("resultingRegions");
	}
	

	private MTBImage m_labelImage;

	private int m_sizeX;
	private int m_sizeY;
	private int m_sizeZ;
	
	@Override
	protected void operate() throws MTBOperatorException, MTBProcessingDAGException {
		MTBImage img = this.getInputImage();
		boolean diagNeighbors = this.getDiagonalNeighborsFlag();
		

		this.setResultingRegions(regionSegmentation(img, diagNeighbors));
		
		// draw label image
		DrawRegion3DSet drawer = new DrawRegion3DSet(DrawType.ID_IMAGE, this.getResultingRegions());
		drawer.runOp(this);

		this.setLabelImage(drawer.getResultImage());
	}
	
	/**
	 * @param binImage binary input image
	 * @param diagonalNeighbors	set true for 26-neighborhood components, set false for 6-neighborhood
	 * @param time specifies the time step index (defaults to 0 if sizeT of the MTBImage is 1)
	 * @param channel specifies the channel index (defaults to 0 if sizeC of the MTBImage is 1)
	 * @return 					returns a vector of regions	
	 */
	protected MTBRegion3DSet regionSegmentation(MTBImage binImage, boolean diagonalNeighbors/*, int time, int channel*/) {
		
		m_sizeX = binImage.getSizeX();
		m_sizeY = binImage.getSizeY();
		m_sizeZ = binImage.getSizeZ();
		
//		m_t = time;
//		m_c = channel;		
		
		MTBRegion3DSet regions= new MTBRegion3DSet(0,0,0,binImage.getSizeX()-1,binImage.getSizeY()-1,binImage.getSizeZ()-1);
		
		m_labelImage = MTBImage.createMTBImage(m_sizeX, m_sizeY, m_sizeZ, 1, 1, MTBImageType.MTB_INT);
        m_labelImage.setTitle(binImage.getTitle() + " labels");
		
        // neighbor label values of the actual pixel
		// l_xyz encodes the neighbor position where 'm' stands for (actual-coord - 1) in the corresponding dimension (xyz),
		// 'p' stands for (actual-coord + 1) and 'o' for (actual-coord)
		int l_mmm; // this is the label of the neighbor (x-1,y-1,z-1)
		int l_omm, l_pmm, l_mom, l_oom, l_pom, l_mpm, l_opm, l_ppm, l_mmo, l_omo, l_pmo, l_moo;
        
        // if only 6-neighborhood is regarded
        // set the diagonal neighors to zero to leave them
        // out of consideration in the following algorithm
        l_mmm = 0;
        l_omm = 0;
        l_pmm = 0;
        l_mom = 0;
        l_pom = 0;
        l_mpm = 0;
        l_opm = 0;
        l_ppm = 0;
        l_mmo = 0;
        l_pmo = 0;
        
        // smallest neigbor-label
        int l_min;
        
        // count of labeled neighbors
        int l_cnt;
        
        // label variable, which is incremented for new labels
		int label = 1;
		
		
		// collision sets for certain labels (vectorindex = label-1)
		Vector< LinkedHashSet<Integer> > collisions = new Vector< LinkedHashSet<Integer> >();
		
		// tmp references to collision sets
		LinkedHashSet<Integer> tmpSet1;
		LinkedHashSet<Integer> tmpSet2;
		// Iterator over the set elements
		Iterator<Integer> setIter;
		Integer labelInt;
		
		// first step: 
		// run line by line through the image
		// and label pixels		
		for (int z = 0; z < m_sizeZ; z++) {
			for (int y = 0; y < m_sizeY; y++) {
				for (int x = 0; x < m_sizeX; x++) {
					
					if (binImage.getValueInt(x,y,z,0,0/*m_t,m_c*/) > 0 && m_labelImage.getValueInt(x,y,z,0,0)==0) {
						
						l_cnt = 0;
						l_min = 0;
						
						if (diagonalNeighbors) {
							
							// get relevant neighboring labels (26-neighborhood)
							l_mmm = getLabelSavely(x-1, y-1, z-1);
							l_omm = getLabelSavely(x  , y-1, z-1);
							l_pmm = getLabelSavely(x+1, y-1, z-1);
							l_mom = getLabelSavely(x-1, y  , z-1);
							l_oom = getLabelSavely(x  , y  , z-1);
							l_pom = getLabelSavely(x+1, y  , z-1);
							l_mpm = getLabelSavely(x-1, y+1, z-1);
							l_opm = getLabelSavely(x  , y+1, z-1);
							l_ppm = getLabelSavely(x+1, y+1, z-1);
							l_mmo = getLabelSavely(x-1, y-1 ,z  );
							l_omo = getLabelSavely(x  , y-1, z  );
							l_pmo = getLabelSavely(x+1, y-1, z  );
							l_moo = getLabelSavely(x-1, y  , z  );
					        
							// determine smallest label and count labeled neighbors
							if (l_mmm > 0) {
								l_cnt++;
								l_min = l_mmm;
							}
							
							if (l_omm > 0) {
								l_cnt++;
								
								if (l_min == 0 || l_omm < l_min) {
									l_min = l_omm;
								}
							}
							
							if (l_pmm > 0) {
								l_cnt++;
								
								if (l_min == 0 || l_pmm < l_min) {
									l_min = l_pmm;
								}
							}
							
							if (l_mom > 0) {
								l_cnt++;
							
								if (l_min == 0 || l_mom < l_min) {
									l_min = l_mom;
								}
							}
							
							
							if (l_oom > 0) {
								l_cnt++;
							
								if (l_min == 0 || l_oom < l_min) {
									l_min = l_oom;
								}
							}
							
							if (l_pom > 0) {
								l_cnt++;
							
								if (l_min == 0 || l_pom < l_min) {
									l_min = l_pom;
								}
							}
							
							if (l_mpm > 0) {
								l_cnt++;
							
								if (l_min == 0 || l_mpm < l_min) {
									l_min = l_mpm;
								}
							}
							
							if (l_opm > 0) {
								l_cnt++;
							
								if (l_min == 0 || l_opm < l_min) {
									l_min = l_opm;
								}
							}
							
							if (l_ppm > 0) {
								l_cnt++;
							
								if (l_min == 0 || l_ppm < l_min) {
									l_min = l_ppm;
								}
							}
							
							if (l_mmo > 0) {
								l_cnt++;
							
								if (l_min == 0 || l_mmo < l_min) {
									l_min = l_mmo;
								}
							}
							
							if (l_omo > 0) {
								l_cnt++;
							
								if (l_min == 0 || l_omo < l_min) {
									l_min = l_omo;
								}
							}
							
							if (l_pmo > 0) {
								l_cnt++;
							
								if (l_min == 0 || l_pmo < l_min) {
									l_min = l_pmo;
								}
							}
							
							if (l_moo > 0) {
								l_cnt++;
							
								if (l_min == 0 || l_moo < l_min) {
									l_min = l_moo;
								}
							}				
											
						}
						else {
							
							// get relevant neighboring labels (6-neighborhood)
							l_oom = getLabelSavely(x  , y  , z-1);
							l_omo = getLabelSavely(x  , y-1, z  );
							l_moo = getLabelSavely(x-1, y  , z  );
						
							// determine smallest label and count labeled neighbors
							if (l_oom > 0) {
								l_cnt++;
								l_min = l_oom;
							}
							
							if (l_omo > 0) {
								l_cnt++;
								
								if (l_min == 0 || l_omo < l_min) {
									l_min = l_omo;
								}
							}	
							
							if (l_moo > 0) {
								l_cnt++;
								
								if (l_min == 0 || l_moo < l_min) {
									l_min = l_moo;
								}
							}
						}
	
						if (l_cnt == 0) {
						// no neighbors of interest are labeled
							
							// give pixel a new label
							m_labelImage.putValueInt(x,y,z,0,0,label);
							
							// create the collision set for that label
							// only containing the new label (-> no collision)
							tmpSet1 = new LinkedHashSet<Integer>();
							tmpSet1.add(label);
							
							// add the new collision set to the vector of collision sets
							collisions.add(tmpSet1);
							
							label++;
						}
						else if (l_cnt == 1) {
						// no label-collision, only one neighbor is labeled, adopt the label
							
							m_labelImage.putValueInt(x,y,z,0,0,l_min);
						}
						else {
						// there might be a label-collision! More than one neighbor is labeled	
							
							
							// assign lowest label to the actual pixel
							m_labelImage.putValueInt(x,y,z,0,0,l_min);
							
							// get the collision set of the label l_min
							tmpSet1 = collisions.get(l_min - 1);
				
							// test for label collisions
							if (l_mmm > 0 && l_mmm != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_mmm - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}
									
							if (l_omm > 0 && l_omm != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_omm - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}						
							
							if (l_pmm > 0 && l_pmm != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_pmm - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}					
																							
							if (l_mom > 0 && l_mom != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_mom - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}						
																							
							if (l_oom > 0 && l_oom != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_oom - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}						
																				
							if (l_pom > 0 && l_pom != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_pom - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}						
																							
							if (l_mpm > 0 && l_mpm != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_mpm - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}						
																							
							if (l_opm > 0 && l_opm != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_opm - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}						
																							
							if (l_ppm > 0 && l_ppm != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_ppm - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}						
																							
							if (l_mmo > 0 && l_mmo != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_mmo - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}								
																							
							if (l_omo > 0 && l_omo != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_omo - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}						
																			
							if (l_pmo > 0 && l_pmo != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_pmo - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}						
																							
							if (l_moo > 0 && l_moo != l_min) {
								
								// get collision set of the colliding label
								tmpSet2 = collisions.get(l_moo - 1);
								
								// if the collision sets are not already the same, join them
								if (tmpSet2 != tmpSet1) {
									setIter = tmpSet2.iterator();
									
									while (setIter.hasNext()) {
										labelInt = setIter.next();
										tmpSet1.add(labelInt);
										collisions.set(labelInt.intValue() - 1, tmpSet1);
									}
								}	
							}									
						}
					}
				}
			}
		}
		
		// run over collision sets and build a simple label mapping array (labelMap):
		// (index of the array - 1) corresponds to the found label
		// the array value corresponds to the value to which the found label is mapped
		int[] labelMapping = new int[label - 1];
		for (int i = 0; i < labelMapping.length; i++) {
			
			// get the collision set for label (i+1)
			tmpSet1 = collisions.get(i);
	
			// set the label mapping to the first label in the set
			setIter = tmpSet1.iterator();
			labelMapping[i] = setIter.next().intValue();		
		}
		
		// clean up
		collisions.clear();
		
		
		// second step: 
		// run line by line through the image and resolve the label collisions,
		// create Region objects and compute its properties
		
		// relates a label (int) to a Region object
		HashMap<Integer,MTBRegion3D> rmap = new HashMap<Integer,MTBRegion3D>();
		Vector<Integer> real_labels = new Vector<Integer>();
		MTBRegion3D act_region;
		int real_label;
		
		for (int z = 0; z  < m_sizeZ; z++) {		
			for (int y = 0; y  < m_sizeY; y++) {
				for (int x = 0; x < m_sizeX; x++) {
					
					label = m_labelImage.getValueInt(x,y,z,0,0);
					
					// if a labeled pixel is found
					if (label != 0) {
						
						real_label = labelMapping[label - 1];
						
						if (rmap.containsKey(real_label)) {
						// update region					
							
							act_region = rmap.get(real_label);
							act_region.addPoint(x,y,z);
							m_labelImage.putValueInt(x, y, z, 0, 0, real_label);
						}
						else {
						// new region
							
							act_region = new MTBRegion3D();
							act_region.addPoint(x,y,z);	
							rmap.put(real_label, act_region);
							
							real_labels.add(real_label);
							m_labelImage.putValueInt(x, y, z, 0, 0, real_label);
						}				
					}
				}
			}
		}
			
		m_labelImage = null;
		
		// build vector of regions
		for (int i = 0; i < real_labels.size(); i++) {
			act_region = rmap.get(real_labels.get(i));
			act_region.setID(i+1);
			regions.add(act_region);
		}
		
		// clean up
		rmap.clear();
		real_labels.clear();
		
		return regions;
	}
	
	
	/** returns the value of m_labelImage at position (x,y,z).
	 * Also checks for image boundaries, returns 0 when (x,y,z) is outside of the image
	 * @param x x-coordinate
	 * @param y y-coordinate
	 * @param z z-coordinate
	 * @return pixel value at (x,y,z) if (x,y,z) inside the image. If outside of the image, returns 0
	 */
	private int getLabelSavely(int x, int y, int z) {
		
		if ((x >= 0) && (x < m_sizeX) && (y >= 0) && (y < m_sizeY) && (z >= 0) && (z < m_sizeZ)) {
			return m_labelImage.getValueInt(x,y,z,0,0);
		}
		else {
			return 0;
		}
	}
}
