/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package de.unihalle.informatik.MiToBo.segmentation.thresholds;

import ij.ImagePlus;
import ij.gui.NewImage;
import ij.process.ImageProcessor;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImageByte;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException.OperatorExceptionType;
import de.unihalle.informatik.MiToBo.operator.MTBOpArgumentDescriptor;
import de.unihalle.informatik.MiToBo.operator.MTBOperator;

/**
 * 
 * Image thresholding class.
 * 
 *
 */
public class ImgThresh extends MTBOperator {

  protected void collectArguments() throws MTBOperatorException {
        super.collectArguments();

		MTBOpArgumentDescriptor[] inputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("InputImage", MTBImage.class, 
																	"Input image", true, null),
				new MTBOpArgumentDescriptor("DestinationImage", MTBImage.class, 
																	"Optional destination image, if ommited a new image is created", false, null)
		};

		MTBOpArgumentDescriptor[] outputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("ResultImage", MTBImage.class, 
																		"Result image", true, null)
		};

		MTBOpArgumentDescriptor[] parameterDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("Threshold", java.lang.Double.class,
																		"Threshold", true, null),
				new MTBOpArgumentDescriptor("FGValue", java.lang.Double.class,
																		"Foreground value", true, 255.0),
				new MTBOpArgumentDescriptor("BGValue", java.lang.Double.class,
																		"Background value", true, 0.0),
				new MTBOpArgumentDescriptor("ActualSliceOnly", java.lang.Boolean.class,
																		"Threshold only current slice ", false, Boolean.FALSE)
		 	};
		
		
		addDescriptors( parameterDescriptors, inputDescriptors, 
				outputDescriptors, null);
	}


	@Override
	protected void operate() throws MTBOperatorException {
	
		
		MTBImage img = getInputImage();
		double thresh = getThreshold();
		boolean actSliceOnly = getActualSliceOnly();
		
		double fgVal, bgVal;
		
		if (isFGOriginalValue()) {
			fgVal = Double.POSITIVE_INFINITY;
		}
		else {
			fgVal = getFGValue();
		}
		
		if (isBGOriginalValue()) {
			bgVal = Double.POSITIVE_INFINITY;
		}
		else {
			bgVal = getBGValue();
		}
		
		
		MTBImage resultImg = getDestinationImage();
		
		if (resultImg == null) {
			resultImg = this.threshold(img, thresh, fgVal, bgVal, actSliceOnly);
		}
		else {
			try {
				this.threshold(img, resultImg, thresh, fgVal, bgVal, actSliceOnly);
			} catch(IllegalArgumentException e) {
				throw new MTBOperatorException(OperatorExceptionType.OPERATE_FAILED, "ImgThresh.operate() failed: " + e.getMessage());
			}
		}
		
		setResultImage(resultImg);
	}
	
	
	/**
	 * Apply threshold to an MTBImage and create a new thresholded MTBImage. Specify values for foreground and background value.
	 * If fgValue/bgValue is Double.POSITIVE_INFINITY the original image values are used for foreground/background pixels.
	 * If fgValue == 255 and bgValue == 0, the result image will be of type MTB_BYTE
	 * @param img source MTBImage
	 * @param thresh Threshold
	 * @param fgValue Foreground pixel value. If set to Double.POSITIVE_INFINITY, pixels higher than/equal to the threshold are set to their original values.
	 * @param bgValue Background pixel value. If set to Double.POSITIVE_INFINITY, pixels lower to the threshold are set to their original values.
	 * @param actualSliceOnly flag for only thresholding actual slice (true) or whole image (false)
	 * @return new thresholded MTBImage object
	 */
	protected MTBImage threshold(MTBImage img, double thresh, double fgValue, double bgValue, boolean actualSliceOnly) {
		int width = img.getSizeX();
		int height = img.getSizeY();
		int stacksize = img.getSizeStack();
		int actSlice = img.getActualSliceIndex();
			
		MTBImage threshImg;
		if (fgValue == 255.0 && bgValue == 0.0) {
			threshImg = MTBImage.createMTBImage(width, height, img.getSizeZ(), img.getSizeT(), img.getSizeC(), MTBImageType.MTB_BYTE);
		}
		else {
			threshImg = MTBImage.createMTBImage(width, height, img.getSizeZ(), img.getSizeT(), img.getSizeC(), img.getType());
		}
		
		threshImg.adoptSliceLabels(img);
		threshImg.copyPhysicalProperties(img);
			
		double val;
		int i = 0;
		
		if (actualSliceOnly) {
			i = actSlice;
			stacksize = i+1;
		}
		
		while (i < stacksize) {
			img.setActualSliceIndex(i);
			threshImg.setActualSliceIndex(i);
			
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					
					val = img.getValueDouble(x, y);
					
					if (val >= thresh) {
						if (Double.isInfinite(fgValue)) {
							threshImg.putValueDouble(x, y, val);
						}
						else {
							threshImg.putValueDouble(x, y, fgValue);
						}
					}
					else {
						if (Double.isInfinite(bgValue)) {
							threshImg.putValueDouble(x, y, val);
						}
						else {
							threshImg.putValueDouble(x, y, bgValue);
						}
					}
				}
			}
			
			i++;
		}
		
		img.setActualSliceIndex(actSlice);
		threshImg.setActualSliceIndex(actSlice);
		
		return threshImg;
	}
	
	/**
	 * Apply threshold to an MTBImage and write results to a destination MTBImage. Specify values for foreground and background value.
	 * If fgValue/bgValue is Double.POSITIVE_INFINITY the original image values are used for foreground/background pixels.
	 * @param img source MTBImage
	 * @param threshImg destination image to write the thresholded result to
	 * @param thresh Threshold
	 * @param fgValue Foreground pixel value. If set to Double.POSITIVE_INFINITY, pixels higher than/equal to the threshold are set to their original values.
	 * @param bgValue Background pixel value. If set to Double.POSITIVE_INFINITY, pixels lower to the threshold are set to their original values.
	 * @param actualSliceOnly flag for only thresholding actual slice (true) or whole image (false)
	 * @return new thresholded MTBImage object
	 */
	protected void threshold(MTBImage img, MTBImage threshImg, double thresh, double fgValue, double bgValue, boolean actualSliceOnly) throws IllegalArgumentException {
		int width = img.getSizeX();
		int height = img.getSizeY();
		int stacksize = img.getSizeStack();
		int actSlice = threshImg.getActualSliceIndex();
		
		
		if (threshImg != null) {
			if (!img.equalSize(threshImg)) {
				throw new IllegalArgumentException("Input image and thresholded destination image must have same size.");
			}
		}
		else {
			throw new IllegalArgumentException("Thresholded destination image may not be null.");
		}
			
		double val;
		int i = 0;
		
		if (actualSliceOnly) {
			i = actSlice;
			stacksize = i+1;
		}
		
		while (i < stacksize) {
			img.setActualSliceIndex(i);
			threshImg.setActualSliceIndex(i);
			
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					
					val = img.getValueDouble(x, y);
					
					if (val >= thresh) {
						if (Double.isInfinite(fgValue)) {
							threshImg.putValueDouble(x, y, val);
						}
						else {
							threshImg.putValueDouble(x, y, fgValue);
						}
					}
					else {
						if (Double.isInfinite(bgValue)) {
							threshImg.putValueDouble(x, y, val);
						}
						else {
							threshImg.putValueDouble(x, y, bgValue);
						}
					}
				}
			}
			
			i++;
		}
		
		img.setActualSliceIndex(actSlice);
		threshImg.setActualSliceIndex(actSlice);
	}
	
	
	/**
	 * Threshold an ImageProcessor. Pixels higher than the threshold are set to 255, pixels lower or equal than the threshold are set to 0.
	 * @param ip Input ImageProcessor
	 * @param thresh Threshold
	 * @return ImagePlus of type 'byte'
	 */
  @Deprecated
	public ImagePlus applyThreshold(ImageProcessor ip, int thresh) {

		int width= ip.getWidth();
		int height= ip.getHeight();
		
		// create result image
		ImagePlus binImg= 
			NewImage.createByteImage("Thresholded Image", width, height, 1, NewImage.FILL_BLACK);
		ImageProcessor binIP= binImg.getProcessor();
		byte[] outPixels= (byte [])binIP.getPixels();
		
		int pos= 0;
		for (int y=0; y<height;++y) {
			for (int x=0; x<width; ++x) {
				pos= y*width+x;
				if (ip.getPixel(x,y) >= thresh) {
					outPixels[pos]= (byte)255;
				}
				else {
					outPixels[pos]= (byte)0;
				}
			}
		}
		return binImg;
	}
	
	/**
   * Threshold an ImageProcessor. Pixels higher than the threshold are set to 255, pixels lower or equal than the threshold are set to 0.
   * @param img Input Image
   * @param thresh Threshold
   * @return MTBImageByte of type 'byte'
   */
  @Deprecated
  public MTBImageByte applyThreshold(MTBImage img, int thresh) {

    int width= img.getSizeX();
    int height= img.getSizeY();
    ImageProcessor ip = img.getImagePlus().getProcessor();
    
    // create result image
    ImagePlus binImg= 
      NewImage.createByteImage("Thresholded Image", width, height, 1, NewImage.FILL_BLACK);
    ImageProcessor binIP= binImg.getProcessor();
    byte[] outPixels= (byte [])binIP.getPixels();
    
    int pos= 0;
    for (int y=0; y<height;++y) {
      for (int x=0; x<width; ++x) {
        pos= y*width+x;
        if (ip.getPixel(x,y) >= thresh) {
          outPixels[pos]= (byte)255;
        }
        else {
          outPixels[pos]= (byte)0;
        }
      }
    }
    return (MTBImageByte) MTBImageByte.createMTBImage(binImg);
  }
	
	
	/**
	 * Apply threshold to an ImageProcessor. Specify values for foreground and background value.
	 * If fgValue/bgValue is -1 the original image values are used for foreground/background pixels.
	 * @param ip Input ImageProcessor
	 * @param thresh Threshold
	 * @param fgValue Foreground pixel value. If set to -1, pixels higher than/equal to the threshold are set to their original values.
	 * @param bgValue Background pixel value. If set to -1, pixels lower to the threshold are set to their original values.
	 * @return new thresholded ImageProcessor object
	 */
  	@Deprecated
	public ImageProcessor applyThreshold(ImageProcessor ip, int thresh, int fgValue, int bgValue) {
		int width= ip.getWidth();
		int height= ip.getHeight();
		
		ImageProcessor result = ip.duplicate();
		int val;
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				val = ip.getPixel(x, y);
				if (val >= thresh) {
					if (fgValue != -1) {
						result.putPixel(x, y, fgValue);
					}
					else {
						result.putPixel(x, y, val);
					}
				}
				else {
					if (bgValue != -1) {
						result.putPixel(x, y, bgValue);
					}
					else {
						result.putPixel(x, y, val);
					}
				}
			}
		}
		return result;
	}
	
	/**
	 * Apply threshold to an MTBImage and create a new thresholded MTBImage. Specify values for foreground and background value.
	 * If fgValue/bgValue is -1 the original image values are used for foreground/background pixels.
	 * @param img source MTBImage
	 * @param thresh Threshold
	 * @param fgValue Foreground pixel value. If set to -1, pixels higher than/equal to the threshold are set to their original values.
	 * @param bgValue Background pixel value. If set to -1, pixels lower to the threshold are set to their original values.
	 * @return new thresholded MTBImage object
	 */
  	@Deprecated
	public MTBImage applyThreshold(MTBImage img, double thresh, double fgValue, double bgValue) {
		int width = img.getSizeX();
		int height = img.getSizeY();
		int stacksize = img.getSizeStack();
		int actSlice = img.getActualSliceIndex();
			
		MTBImage threshImg = img.duplicate();
			
		double val;
		for (int i = 0; i < stacksize; i++) {
			img.setActualSliceIndex(i);
			threshImg.setActualSliceIndex(i);
			
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					
					val = img.getValueDouble(x, y);
					
					if (val >= thresh) {
						if (fgValue != -1) {
							threshImg.putValueDouble(x, y, fgValue);
						}
						else {
							threshImg.putValueDouble(x, y, val);
						}
					}
					else {
						if (bgValue != -1) {
							threshImg.putValueDouble(x, y, bgValue);
						}
						else {
							threshImg.putValueDouble(x, y, val);
						}
					}
				}
			}
		}
		
		img.setActualSliceIndex(actSlice);
		
		return threshImg;
	}
	
	/**
	 * Apply threshold to an MTBImage and write results to an existing MTBImage. Specify values for foreground and background value.
	 * If fgValue/bgValue is -1 the original image values are used for foreground/background pixels.
	 * @param img source MTBImage
	 * @param threshImg thresholding results destination MTBImage
	 * @param thresh Threshold
	 * @param fgValue Foreground pixel value. If set to -1, pixels higher than/equal to the threshold are set to their original values.
	 * @param bgValue Background pixel value. If set to -1, pixels lower to the threshold are set to their original values.
	 * @return new thresholded MTBImage object
	 */
  	@Deprecated
	public void applyThreshold(MTBImage img, MTBImage threshImg, double thresh, double fgValue, double bgValue) {
		int width = img.getSizeX();
		int height = img.getSizeY();
		int stacksize = img.getSizeStack();
		int actSlice = img.getActualSliceIndex();
		int actSliceT = threshImg.getActualSliceIndex();
			
		double val;
		for (int i = 0; i < stacksize; i++) {
			img.setActualSliceIndex(i);
			threshImg.setActualSliceIndex(i);
			
			for (int y = 0; y < height; y++) {
				for (int x = 0; x < width; x++) {
					
					val = img.getValueDouble(x, y);
					
					if (val >= thresh) {
						if (fgValue != -1) {
							threshImg.putValueDouble(x, y, fgValue);
						}
						else {
							threshImg.putValueDouble(x, y, val);
						}
					}
					else {
						if (bgValue != -1) {
							threshImg.putValueDouble(x, y, bgValue);
						}
						else {
							threshImg.putValueDouble(x, y, val);
						}
					}
				}
			}
		}
		
		img.setActualSliceIndex(actSlice);
		threshImg.setActualSliceIndex(actSliceT);
	}
	
	/**
	 * Apply threshold to the slice of a MTBImage and write result to an existing MTBImage. Specify values for foreground and background value.
	 * If fgValue/bgValue is -1 the original image values are used for foreground/background pixels.
	 * @param img source MTBImage
	 * @param threshImg thresholding results destination MTBImage
	 * @param thresh Threshold
	 * @param fgValue Foreground pixel value. If set to -1, pixels higher than/equal to the threshold are set to their original values.
	 * @param bgValue Background pixel value. If set to -1, pixels lower to the threshold are set to their original values.
	 * @param sliceIndex index of the slice that has to be thresholded
	 * @return new thresholded MTBImage object
	 */
  	@Deprecated
	public void applyThresholdToSlice(MTBImage img, MTBImage threshImg, double thresh, double fgValue, double bgValue, int sliceIndex) {
		int width = img.getSizeX();
		int height = img.getSizeY();
		
		int actSlice = img.getActualSliceIndex();
		int actSliceT = threshImg.getActualSliceIndex();
			
		double val;

		img.setActualSliceIndex(sliceIndex);
		threshImg.setActualSliceIndex(sliceIndex);
			
		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				
				val = img.getValueDouble(x, y);
				
				if (val >= thresh) {
					if (fgValue != -1) {
						threshImg.putValueDouble(x, y, fgValue);
					}
					else {
						threshImg.putValueDouble(x, y, val);
					}
				}
				else {
					if (bgValue != -1) {
						threshImg.putValueDouble(x, y, bgValue);
					}
					else {
						threshImg.putValueDouble(x, y, val);
					}
				}
			}
		}
		
		img.setActualSliceIndex(actSlice);
		threshImg.setActualSliceIndex(actSliceT);
	}
	
	
		/**
		 * Constructor for thresholding using 255 as foreground and 0 as background value.
		 * The result image will be of type MTB_BYTE
		 * @param img input image
		 * @param threshold values < threshold are set to 0, value >= threshold are set to 255
		 */
		public ImgThresh(MTBImage img, double threshold) throws MTBOperatorException {

			this.setInput("InputImage", img);
			this.setParameter("Threshold", new Double(threshold));
			this.setParameter("FGValue", new Double(255));
			this.setParameter("BGValue", new Double(0));
			this.setParameter("ActualSliceOnly", new Boolean(false));
		}
		
		public ImgThresh() throws MTBOperatorException {
			// nothing to be done here
		}

		/**
		 * 
		 * Constructor for thresholding using 'fgValue' as foreground and 'bgValue' as background value.
		 * The result image will have the same type as the input image.
		 * @param img input image
		 * @param threshold values < threshold are set to 'fgValue', value >= threshold are set to 'bgValue'
		 * @param fgValue foreground value (POSITIVE_INFINITY is interpreted as use original value)
		 * @param bgValue background value (POSITIVE_INFINITY is interpreted as use original value)
		 */
		public ImgThresh(MTBImage img, double threshold, double fgValue, double bgValue) throws MTBOperatorException {
			this.setInput("InputImage", img);
			this.setParameter("Threshold", new Double(threshold));
			this.setParameter("FGValue", new Double(fgValue));
			this.setParameter("BGValue", new Double(bgValue));
		}
				
		/**
		 * Get input image
		 * @return input image
		 */
		public MTBImage getInputImage() throws MTBOperatorException {
			return (MTBImage)this.getInput("InputImage");
		}
		
		/**
		 * Get threshold
		 * @return threshold
		 */
		public double getThreshold() throws MTBOperatorException {
			return (Double)this.getParameter("Threshold");
		}
		
		/**
		 * Get foreground value 
		 * @return foreground value
		 */
		public Double getFGValue() throws MTBOperatorException {
			return (Double)(this.getParameter("FGValue"));
		}
		
		/**
		 * Get background value 
		 * @return background value
		 */
		public Double getBGValue() throws MTBOperatorException {
			return (Double)(this.getParameter("BGValue"));
		}
		
		/**
		 * Test if foreground pixels keep their original pixel values
		 * @return true if foreground pixels keep their original pixel values
		 */
		public boolean isFGOriginalValue() throws MTBOperatorException {
			return getFGValue() == Double.POSITIVE_INFINITY;
		}
		
		/**
		 * Test if background pixels keep their original pixel values
		 * @return true if background pixels keep their original pixel values
		 */
		public boolean isBGOriginalValue() throws MTBOperatorException {
			return getBGValue() == Double.POSITIVE_INFINITY;
		}
		
		/**
		 * Tell the thresholder to use the original pixel value for foreground pixels
		 */
		public void setFGOriginalValue() throws MTBOperatorException {
			this.setFGValue( Double.POSITIVE_INFINITY);
		}
		
		/**
		 * Tell the thresholder to use the original pixel value for background pixels
		 */
		public void setBGOriginalValue() throws MTBOperatorException {
			this.setBGValue( Double.POSITIVE_INFINITY);
		}
		
		/**
		 * Get flag for thresholding only the actual slice (true) or the whole image (false) 
		 * @return
		 */
		public boolean getActualSliceOnly() throws MTBOperatorException {
			return (Boolean)this.getParameter("ActualSliceOnly");
		}
		
		/**
		 * Only actual slice is thresholded
		 */
		public void setActualSliceOnly() throws MTBOperatorException {
			this.setParameter("ActualSliceOnly", true);
		}
		
		/**
		 * The whole image is thresholded (default)
		 */
		public void unsetActualSliceOnly() throws MTBOperatorException {
			this.setParameter("ActualSliceOnly", false);
		}
		
		/**
		 * Get the result image
		 * @return result image
		 */
		public MTBImage getResultImage() throws MTBOperatorException {
			return (MTBImage)this.getOutput("ResultImage");
		}
		
		/**
		 * Set result image. You may specify a destination image to which the thresholding results are written. Otherwise the operator sets a new resulting image.
		 * @param img result image
		 */
		public void setResultImage(MTBImage img) throws MTBOperatorException {
			this.setOutput("ResultImage", img);
		}
	
	/** Set value of Parameter argument FGValue.
	  * @param value New value for FGValue
	  */
	public void setFGValue( java.lang.Double value ) throws MTBOperatorException {
         	this.setParameter("FGValue", value);
	}

	/** Set value of Parameter argument BGValue.
	  * @param value New value for BGValue
	  */
	public void setBGValue( java.lang.Double value ) throws MTBOperatorException {
         this.setParameter("BGValue", value);
	}
 /** Get value of Input argument DestinationImage.
  * @return value of DestinationImage
  */
public de.unihalle.informatik.MiToBo.datatypes.images.MTBImage getDestinationImage() throws MTBOperatorException {
	return (de.unihalle.informatik.MiToBo.datatypes.images.MTBImage)(this.getInput("DestinationImage"));
}
/** Set value of Input argument DestinationImage.
  * @param value New value for DestinationImage
  */
public void setDestinationImage( de.unihalle.informatik.MiToBo.datatypes.images.MTBImage value ) throws MTBOperatorException {
	 this.setInput("DestinationImage", value);
}

}

