/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package mtb_morphology;

import ij.*;
import ij.process.ImageProcessor;
import ij.plugin.filter.PlugInFilter;

import java.io.*;

import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.exceptions.*;
import de.unihalle.informatik.MiToBo.morphology.*;
import de.unihalle.informatik.MiToBo.operator.*;
import de.unihalle.informatik.MiToBo.io.files.*;

/**
 * Morphological erosion on binary images.
 * 
 * This plugin applies a squared morphological mask to a given 
 * binary image. Pixels with values larger than 0 are assumed
 * as foreground, while pixels with value 0 are assumed as 
 * background. 
 * 
 * @author moeller
 */
public class Erode_Image implements PlugInFilter {

    /** input image */
    private MTBImage m_img;

	@Override
	public int setup(String arg, ImagePlus imp) {
        try {
            m_img = MTBImage.createMTBImage(imp);
        } catch(IllegalArgumentException e) {
            // do nothing because the below returned types are all legal.
            // ImageJ will forbid an illegal type
        }

		return DOES_8G + DOES_16;
	}

	@Override
  	public void run(ImageProcessor arg0) {
		// get mask size from user
		int masksize= -1;
		while (masksize < 0) {
			masksize= (int)IJ.getNumber("Enter masksize (3, 5, 7, ...): ", 3);       
			if (masksize < 0)
				IJ.error("Please enter a positive number!");
		}
		// set inputs and parameters
		try {
			ImgErode imEro= new ImgErode( m_img, masksize);
			imEro.runOp( null);
			((MTBImage)imEro.getOutput("resultImg")).show();
    	} catch (MTBException e) {
	    	// TODO Auto-generated catch block
	    	e.printStackTrace();
    	}
  	}

    public static void main(String [] args) {
        try {

            if ( args.length != 3 ) {
                System.err.println( "usage: Erode_Image inimage masksize resultimg");
                System.exit(-1);
            }

			ImageOpen IO = new ImageOpen( args[0], null, true, 0);
			IO.runOp(null);
            MTBImage inImg = IO.getImage();

			ImgErode imEro= new ImgErode( inImg, new Integer(args[1]));
			imEro.runOp( null);

			MTBImage resultImg = (MTBImage)imEro.getOutput( "resultImg");
			ImageSave writer= new ImageSave( args[2], resultImg, true, true);
			writer.runOp( null);
        } catch ( Exception e) {
            e.printStackTrace();
        }
    }

}
