/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package mtb_regions;

import ij.IJ;
import ij.ImagePlus;
import ij.gui.GenericDialog;
import ij.plugin.filter.PlugInFilter;
import ij.process.ImageProcessor;
import de.unihalle.informatik.MiToBo.datatypes.MTBRegion2DSet;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImageByte;
import de.unihalle.informatik.MiToBo.drawing.DrawRegion2DSet;
import de.unihalle.informatik.MiToBo.drawing.DrawRegion2DSet.DrawType;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.io.files.ImageOpen;
import de.unihalle.informatik.MiToBo.io.files.ImageSave;
import de.unihalle.informatik.MiToBo.morphology.ComponentPostprocess;
import de.unihalle.informatik.MiToBo.segmentation.regions.labeling.LabelComponentsSequential;

/**
 * 
 * @author moeller
 *
 * This class implements a sequential component labeling technique on 
 * binary images. Foreground pixels are assumed to have a value > 0,
 * all pixels with value 0 are thought as background.
 * 
 * Called as plugin the class displays a labeled image where 
 * extracted regions are colored in pseudo colors.
 * 
 * The static routine returns a vector of regions.
 *
 */
public class Label_Components_Sequential implements PlugInFilter {
	
	/** original image */
	private MTBImage m_img;
	
	private int returnImgIdx;
	
	
	public int setup(String arg, ImagePlus img) {
        try {
        	m_img = MTBImage.createMTBImage(img);
        } catch(IllegalArgumentException e) {
        	// do nothing because the below returned types are all legal.
        	// ImageJ will forbid an illegal type
        }       

		return DOES_8G+DOES_16+DOES_32;
	}
	
	/**
	 *  Main plugin method.
	 */
	public void run(ImageProcessor ip) {

		try { 
			LabelComponentsSequential lcs = parameterDialog();
			
			if (lcs == null) {
				return;
			}
			
			lcs.runOp(null);
				
			MTBRegion2DSet regions = lcs.getResultingRegions();
			
			if (returnImgIdx == 0) {
			
				// draw random color region image
				DrawRegion2DSet drawer = new DrawRegion2DSet(DrawType.COLOR_IMAGE, regions);
				drawer.runOp(null);
				
				MTBImage regionImg = drawer.getResultImage();
				regionImg.setTitle(m_img.getTitle() + " random color");
				
				regionImg.show();
			}
			else {
				
				lcs.getLabelImage().show();
			}
			
		} catch (Exception e) {
			e.printStackTrace();
			IJ.showStatus("MiToBo LabelComponentsSequential failed");
		}
	}
	
	
	
	// ---- Plugin helper functions ----
	
	/** Dialog window for parameter input 
	 * @return new LabelComponentsSequential3DParams object
	 */
	protected LabelComponentsSequential parameterDialog() throws MTBOperatorException {
		
		// Dialog for parameter choice
		GenericDialog gd = new GenericDialog("Component labeling parameters");
		
		String[] neighOptions = {"4-Neighborhood", "8-Neighborhood"};
		gd.addChoice("Choose Neighborhood:", neighOptions, neighOptions[1]);	
		
		String[] returnImg = {"Random Color Image", "Label Image"};
		gd.addChoice("Return:", returnImg, returnImg[0]);		
		
		gd.showDialog();
		
		
		if (gd.wasOKed()) {
			
			// get neighborhood
			boolean diagonalNeighbors = (gd.getNextChoiceIndex() == 0) ? false : true;	
			
			// get returned image type
			returnImgIdx = gd.getNextChoiceIndex();
			
			LabelComponentsSequential lcs = new LabelComponentsSequential(m_img, diagonalNeighbors);

			return lcs;
			
		}
		else {
			return null;
		}
	}

    public static void main(String [] args) {
        try {

            if ( args.length < 2 ) {
                System.err.println( "usage: Label_Components_Sequential inimage resultimg [minRegionSize]");
                System.exit(-1);
            }

            ImageOpen IO = new ImageOpen( args[0], null, true, 0);
            IO.runOp(null);
            MTBImage inImg = IO.getImage();

            LabelComponentsSequential op= new LabelComponentsSequential( inImg, false);
            op.runOp( null);

	    	if ( args.length >= 3 ) {
				int minSize = new Integer( args[2]);
System.out.println( "erase small regions less then " + minSize);
				ComponentPostprocess cpp= new ComponentPostprocess( 
					(MTBImageByte)(op.getLabelImage().convertType(MTBImage.MTBImageType.MTB_BYTE, false)),
                	ComponentPostprocess.ProcessMode.ERASE_SMALL_COMPS,minSize,-1,-1);
				
				cpp.setDiagonalNeighbors(false);
				cpp.runOp( null);
            	MTBImage resultImg = (MTBImage)cpp.getResultImage();
            	ImageSave writer= new ImageSave( args[1], resultImg, true, true);
            	writer.runOp(null);

	    	} else {
            	MTBImage resultImg = (MTBImage)op.getOutput( "LabelImage");
            	ImageSave writer= new ImageSave( args[1], resultImg, true, true);
            	writer.runOp(null);
			}
        } catch ( Exception e) {
            e.printStackTrace();
        }
    }



}
