/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package mtb_regions;

import ij.IJ;
import ij.ImagePlus;
import ij.gui.GenericDialog;
import ij.plugin.filter.PlugInFilter;
import ij.process.ImageProcessor;
//import loci.formats.StatusEvent;
//import loci.formats.StatusListener;
import de.unihalle.informatik.MiToBo.datatypes.MTBRegion3DSet;
import de.unihalle.informatik.MiToBo.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.drawing.DrawRegion3DSet;
import de.unihalle.informatik.MiToBo.drawing.DrawRegion3DSet.DrawType;
import de.unihalle.informatik.MiToBo.exceptions.MTBOperatorException;
import de.unihalle.informatik.MiToBo.segmentation.regions.labeling.LabelComponentsSequential3D;

/**
 * 
 * @author gress
 *
 * This class implements a sequential component labeling technique on 
 * binary images. Foreground pixels are assumed to have a value > 0,
 * all pixels with value 0 are thought as background.
 * 
 * Called as plugin the class displays a labeled image where 
 * extracted regions are colored in pseudo colors.
 *
 */
public class Label_Components_Sequential_3D implements PlugInFilter {//, StatusListener {
	
	/** original image */
	private MTBImage m_img;

	private int returnImgIdx;

	
	public int setup(String arg, ImagePlus img) {
		
		try {
			m_img = MTBImage.createMTBImage(img);
		} catch(IllegalArgumentException e) {
			// do nothing
			// illegal types are caught with the return statement
		}
		
		return DOES_8G+DOES_16+DOES_32;
	}
	
	/**
	 *  Main plugin method.
	 */
	public void run(ImageProcessor ip) {
		
		try { 
			LabelComponentsSequential3D lcs = this.createOperatorDialog();
			
			if (lcs == null) {
				return;
			}
			
			lcs.runOp(null);
				
			MTBRegion3DSet regions = lcs.getResultingRegions();
			
			if (returnImgIdx == 0) {
			
				// draw random color region image
				DrawRegion3DSet drawer = new DrawRegion3DSet(DrawType.COLOR_IMAGE, regions);
				drawer.runOp(null);
				
				MTBImage regionImg = drawer.getResultImage();
				regionImg.setTitle(m_img.getTitle() + " random color");
				
				regionImg.show();
			}
			else {
				
				lcs.getLabelImage().show();
			}
			
		} catch (Exception e) {
			e.printStackTrace();
			IJ.showStatus("MiToBo LabelComponentsSequential3D failed");
		}

	}
	
	
	// ---- Plugin helper functions ----
	
	/** create a LabelComponentsSequential3D operator using a dialog for parameter input 
	 * @return configured LabelComponentsSequential3D operator
	 * @throws MTBOperatorException 
	 * @throws IllegalArgumentException 
	 */
	protected LabelComponentsSequential3D createOperatorDialog() throws MTBOperatorException {
		
		// Dialog for parameter choice
		GenericDialog gd = new GenericDialog("Component labeling parameters");
		
		String[] neighOptions = {"6-Neighborhood", "26-Neighborhood"};
		gd.addChoice("Choose Neighborhood:", neighOptions, "6-Neighborhood");	
		
		String[] returnImg = {"Random Color Image", "Label Image"};
		gd.addChoice("Return:", returnImg, returnImg[0]);	
		
		gd.showDialog();
		
		
		if (gd.wasOKed()) {
			
			// get neighborhood
			boolean diagonalNeighbors = (gd.getNextChoiceIndex() == 0) ? false : true;
			
			// get returned image type
			returnImgIdx = gd.getNextChoiceIndex();
			
			LabelComponentsSequential3D lcs = new LabelComponentsSequential3D(m_img, diagonalNeighbors);

			return lcs; 
		}
		else {
			return null;
		}
	}


	
	// ----- functions to be implemented for StatusListener interface
	
//	/*
//	 * (non-Javadoc)
//	 * @see loci.formats.StatusListener#statusUpdated(loci.formats.StatusEvent)
//	 */
//	public void statusUpdated(StatusEvent e) {
//		IJ.showStatus(e.getStatusMessage());
//		IJ.showProgress(e.getProgressValue(), e.getProgressMaximum());
//	}
	
}
