/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package mtb_segmentation;

import java.util.LinkedList;
import java.io.*;

import de.unihalle.informatik.MiToBo.datatypes.images.*;
import de.unihalle.informatik.MiToBo.operator.*;
import de.unihalle.informatik.MiToBo.exceptions.*;
import de.unihalle.informatik.MiToBo.datatypes.*;
import de.unihalle.informatik.MiToBo.io.files.*;
import de.unihalle.informatik.MiToBo.segmentation.thresholds.*;

import ij.*;
import ij.process.*;
import ij.plugin.filter.*;

/**
 * Calculation of Otsu's threshold on grayscale images.
 * 
 * This plugin provides methods to calculate thresholds on 
 * grayscale images using Otsu's method. Assuming a bimodal
 * histogram Otsu thresholding seeks to seperate the intensity
 * values of the image into two more or less equally sized
 * classes that are maximally separated from each other in 
 * terms of their average intensity values.
 * 
 * @see de.unihalle.informatik.MiToBo.segmentation.thresholds.CalcGlobalThreshOtsu
 * 
 * @author moeller
 */
public class CalcOtsuThreshold_Image implements PlugInFilter {

	/** input image */
    private MTBImage m_img;

	@Override
	public int setup(String arg, ImagePlus imp) {
        try {
            m_img = MTBImage.createMTBImage(imp);
        } catch(IllegalArgumentException e) {
            // do nothing because the below returned types are all legal.
            // ImageJ will forbid an illegal type
        }
  
		return DOES_8G+DOES_16;
	}

	@Override
	public void run(ImageProcessor ip) {
		try {
			CalcGlobalThreshOtsu otsuCalc= new CalcGlobalThreshOtsu( m_img);
			otsuCalc.runOp( null);

			Double value = otsuCalc.getOtsuThreshold().getValue();
			IJ.showMessage("Calculated Threshold: " + value);
		} catch (MTBException e) {
        	// TODO Auto-generated catch block
        	e.printStackTrace();
        }

	}

	public static void main(String [] args) {
		try {

			if ( args.length != 1 ) {
				System.err.println( "usage: CalcOtsuThreshold_Image inimage");
				System.exit(-1);
			}

			ImageOpen IO = new ImageOpen( args[0], null, true, 0);
			IO.runOp(null);
			MTBImage inImg =    IO.getImage();

			CalcGlobalThreshOtsu otsuOp = new CalcGlobalThreshOtsu( inImg);
			otsuOp.runOp( null);
			Double value = otsuOp.getOtsuThreshold().getValue();
			System.out.println( "Otsu threshold " + value);
			
		} catch ( Exception e) {
			e.printStackTrace();
		}
	}
}
