/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

package mtb_segmentation;

import java.awt.AWTEvent;
import java.awt.Choice;
import java.util.Vector;

import ij.*;
import ij.gui.DialogListener;
import ij.gui.GenericDialog;
import ij.process.*;

import de.unihalle.informatik.MiToBo.datatypes.images.*;
import de.unihalle.informatik.MiToBo.operator.*;
import de.unihalle.informatik.MiToBo.exceptions.*;
import de.unihalle.informatik.MiToBo.segmentation.thresholds.*;

/**
 * Plugin for thresholding plain images, stacks and hyperstacks of 8 or 16 Bit grayvalue type.
 * @author gress
 *
 */
public class Threshold_Image extends  MTBOperatorPlugInFilter implements DialogListener {
	
	@Override
  protected void collectArguments() throws MTBOperatorException {
        super.collectArguments();

		MTBOpArgumentDescriptor[] inputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("DestinationImage", MTBImage.class, 
																	"Optional destination image, if ommited a new image is created", false, null)
		};

		MTBOpArgumentDescriptor[] outputDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("ResultImage", MTBImage.class, 
																		"Result image", true, null)
		};

		MTBOpArgumentDescriptor[] parameterDescriptors = 
			new MTBOpArgumentDescriptor[] {
				new MTBOpArgumentDescriptor("Threshold", java.lang.Double.class,
																		"Threshold", true, null),
				new MTBOpArgumentDescriptor("FGValue", java.lang.Double.class,
																		"Foreground value", true, null),
				new MTBOpArgumentDescriptor("BGValue", java.lang.Double.class,
																		"Background value", true, null),
				new MTBOpArgumentDescriptor("ActualSliceOnly", java.lang.Boolean.class,
																		"Threshold only current slice ", false, Boolean.TRUE)
		 	};
		
			addDescriptors( parameterDescriptors, inputDescriptors, 
					outputDescriptors, null);
		}
	
	protected Threshold_Image() throws MTBOperatorException {
		// nothing to be done here
	}

	/** image to threshold */
	private MTBImage m_img;
	
	/** thresholded image */
	private MTBImage m_threshImg;
	
	/** original ImagePlus */
	private ImagePlus m_imgPlus;
	
	/** threshold */
	private int m_thresh;
	
	/** foreground pixel value */
	private int m_fg;
	
	/** background pixel value */
	private int m_bg;
	
	/** maximum value in original image */
	private double m_imgMax;
	
	// ----- functions to be implemented for PlugInFilter interface
	
	/*
	 * (non-Javadoc)
	 * @see ij.plugin.filter.PlugInFilter#setup(java.lang.String, ij.ImagePlus)
	 */
	@Override
  public int setup(String arg, ImagePlus imp) {
		super.setup(arg,imp);
		this.m_imgPlus = imp;
		return DOES_16 + DOES_8G;
	}
	
    @Override
    public void operate() throws MTBOperatorException, MTBProcessingDAGException {

        ImgThresh thresOp = new ImgThresh((MTBImage)this.getInput("inImg"),
                                       (Double)this.getParameter("Threshold"),
                                       (Double)this.getParameter("FGValue"),
                                       (Double)this.getParameter("BGValue"));
		thresOp.setParameter( "ActualSliceOnly", this.getParameter( "ActualSliceOnly"));
		thresOp.setInput( "DestinationImage", this.getInput( "DestinationImage"));

        thresOp.runOp(this);

        this.setOutput( "ResultImage", thresOp.getResultImage());
    }

	/*
	 * (non-Javadoc)
	 * @see ij.plugin.filter.PlugInFilter#run(ij.process.ImageProcessor)
	 */
	@Override
  public void run(ImageProcessor ip) {
		this.m_img = this.mtbInput;
		this.m_img.setActualSliceIndex(this.inputImage.getCurrentSlice()-1);
		
		// get min and max values of the image
		double[] minmax;
		minmax = this.m_img.getMinMaxDouble();
			
		this.m_imgMax = minmax[1];
		
		// create thresholded image object
		this.m_threshImg = this.m_img.duplicate();
		
		// set ImagePlus parameters
		this.m_threshImg.getImagePlus().setDisplayRange(minmax[0], minmax[1]);
		this.m_threshImg.getImagePlus().setPosition(this.inputImage.getCurrentSlice());

		this.m_threshImg.show();
		
		// thresholding
		if (thresholdSetupDialog()) {

			try {
				ImgThresh m_th = new ImgThresh( this.m_img, new Double(this.m_thresh), this.m_fg, this.m_bg);
				m_th.unsetActualSliceOnly();
				m_th.setDestinationImage( this.m_threshImg);
			
				m_th.runOp(null);

				// this is probably superflous
				this.m_threshImg = m_th.getResultImage();
			} catch ( Exception e) {
				e.printStackTrace();
			}

			this.m_threshImg.updateAndRepaintWindow();
		}
		else {
			this.m_threshImg.close();
		}

	}
	
	
	
	// ----- Plugin helper functions
	
	/**
	 * Dialog for choosing thresholding parameters
	 * @return true if the dialog was terminated by OK-button, false in any other case
	 */
	private boolean thresholdSetupDialog() {
		
		GenericDialog gd = new GenericDialog("Thresholding configuration");
		
		// threshold parameter
		gd.addMessage("Choose a threshold:");
		gd.addSlider("Threshold", 0.0, this.m_imgMax, this.m_imgMax/2.0);
		
		// fg/bg pixel values
		gd.addMessage("Pixel values:");
		String[] choices = {"White", "Black", "Original pixel value"};
		gd.addChoice("Foreground", choices, "White");
		gd.addChoice("Background", choices, "Black");
		
		gd.addDialogListener(this);
		
		gd.showDialog();
		
		if (gd.wasOKed()) {	
			return true;
		}
		else {
			return false;
		}
	}
	
	
	// ----- functions to be implemented for DialogListener interface
	
	/*
	 * (non-Javadoc)
	 * @see ij.gui.DialogListener#dialogItemChanged(ij.gui.GenericDialog,java.awt.AWTEvent)
	 */
	@Override
  public boolean dialogItemChanged(GenericDialog gd, AWTEvent e) {
	
		// get threshold
		this.m_thresh = (int)Math.round(gd.getNextNumber());
		
		// get fg/bg pixel values
		Vector vc = gd.getChoices();
		Choice c = (Choice)vc.get(0);
		int ci = c.getSelectedIndex();
				
		
		// fg
		if (ci == 0) 
			this.m_fg = (int)this.m_imgPlus.getDisplayRangeMax();
		else if (ci == 1)
			this.m_fg = 0;
		else
			this.m_fg = -1;
		
		c = (Choice)vc.get(1);
		ci = c.getSelectedIndex();
		
		// bg
		if (ci == 0) 
			this.m_bg = (int)this.m_imgPlus.getDisplayRangeMax();
		else if (ci == 1)
			this.m_bg = 0;
		else
			this.m_bg = -1;		
		
		
		try {
			ImgThresh thresOp = new ImgThresh(this.m_img, (double)this.m_thresh, (double)this.m_fg, (double)this.m_bg);
			thresOp.setInput( "DestinationImage", this.m_threshImg);
		
			if (this.m_fg == -1)
				thresOp.setFGOriginalValue();
			if (this.m_bg == -1)
				thresOp.setBGOriginalValue();
		
			thresOp.setActualSliceOnly();
			
			thresOp.runOp(null);
			thresOp.getResultImage().updateAndRepaintWindow();

		} catch ( Exception ee) {
			ee.printStackTrace();
		}
		
		
		return true;
	}
}
